   Module EigenValue
   include 'defs.fi'
   Private
   !Integer, Parameter :: float = selected_real_kind(15)
   Public EigenVal       ! DGEEV  - nonsymmetric matrix
   Public EigenValGen    ! DGGEV  - general problem, nonsymmetric
   Public EigenValSym    ! DSYGV  - general problem, symmetric
   Public EigenValTriSym ! DSTEQR - symmetric tridiagonal problem
   public sort_idx
   integer, parameter :: Debug = 0
!-------------------------------------------------------------------------------
   Contains
!-------------------------------------------------------------------------------
   Function EigenVal(A,VRx,VLx) result(w)
      Real(float) :: A(:,:),w(size(A,1),2)
      Real(float), optional :: VLx(:,:),VRx(:,:)
      Real(float) :: VL(size(A,1),size(A,1)),VR(size(A,1),size(A,1))
      Real(float) :: Work(8*size(A,1))
      Integer :: n,Nwork,info,j,isort(size(A,1))
      character :: CHR*1,CHL*1
!-------------------------------------------------------------------------------
!  EigenVal computes the eigenvalues and optionally the left & right eigenvectors
!  of matrix A.
!  w(:,2) - eigenvectors returned, real & imaginary parts
!  A  - the matrix
!  VR - right eigenvectors (optional)
!  VL - left eigenvectors (optional)
!-------------------------------------------------------------------------------
!  LaPack routine DGEEV computes for an N-by-N real nonsymmetric matrix A, the
!  eigenvalues and, optionally, the left and/or right eigenvectors.     
!                                                                       
!  The right eigenvector v(j) of A satisfies                            
!                   A * v(j) = lambda(j) * v(j)                         
!  where lambda(j) is its eigenvalue.                                   
!  The left eigenvector u(j) of A satisfies                             
!                u(j)**H * A = lambda(j) * u(j)**H                      
!  where u(j)**H denotes the conjugate transpose of u(j).               
!                                                                       
!  The computed eigenvectors are normalized to have Euclidean norm      
!  equal to 1 and largest component real.                               
!-------------------------------------------------------------------------------
      n = size(A,1)
      Nwork = 8*N
      if(Debug > 0)write(*,*)' EigenVal: n =',n
      Nwork = 8*N;  CHR = 'N';  CHL = 'N'
      if(Present(VLx))CHL = 'V'
      if(Present(VRx))CHR = 'V'
      Call DGEEV(CHL,CHR,N,A,N,W(1,1),W(1,2),VL,N,VR,N,WORK,Nwork,Info)
      isort = sort_idx(abs(w(:,1)))
      if(Debug > 0 .and. info /= 0)write(*,*)' EigenVal: n, Info = ',n,Info
      w(:,1) = w(isort,1);      w(:,2) = w(isort,2);
      if(Present(VLx))then
         do j = 1,n
            VLx(:,j) =VL(:,isort(j))
         enddo
      endif
      if(Present(VRx))then
         do j = 1,n
            VRx(:,j) =VR(:,isort(j))
         enddo
      endif
   End Function EigenVal
!-------------------------------------------------------------------------------
   Function EigenValGen(A,B,VRx,VLx) result(w)
      Real(float) :: A(:,:),B(:,:),w(size(A,1),2)
      Real(float), optional :: VLx(:,:),VRx(:,:)
      Real(float) :: VL(size(A,1),size(A,1)),VR(size(A,1),size(A,1))
      Real(float) :: wi(size(A,1))
      Real(float) :: Work(8*size(A,1))
      character :: CHR*1,CHL*1
      Integer :: n,Nwork,info
!-------------------------------------------------------------------------------
!  EigenGen computes the eigenvalues and optionally the left & right eigenvectors
!  for the generalized eigenvalue problem A - lambda*B)v.
!  w(:,2) - eigenvectors returned, real & imaginary parts
!  A  - matrix
!  B  - matrix
!  VR - right eigenvectors (optional)
!  VL - left eigenvectors (optional)
!-------------------------------------------------------------------------------
!  LaPack routine DGGEV computes for an N-by-N real nonsymmetric matrix A, the
!  eigenvalues and, optionally, the left and/or right eigenvectors.     
!                                                                       
!  The right eigenvector v(j) of A satisfies                            
!                   A * v(j) = lambda(j) * B * v(j)
!  where lambda(j) is its eigenvalue.                                   
!  The left eigenvector u(j) of A satisfies                             
!                u(j)**H * A = lambda(j) * B * u(j)**H
!  where u(j)**H denotes the conjugate transpose of u(j).               
!                                                                       
!  The computed eigenvectors are normalized to have Euclidean norm      
!  equal to 1 and largest component real.                               
!-------------------------------------------------------------------------------
      n = size(A,1)
      Nwork = 8*N;  CHR = 'N';  CHL = 'N'
      if(Present(VLx))CHL = 'V'
      if(Present(VRx))CHR = 'V'
      Call DGGEV(CHL,CHR,N,A,N,B,N,w(1,1),w(1,2),wi,VL,N,VR,N,WORK,Nwork,Info)
      w(:,1) = w(:,1)/wi(:)
      w(:,2) = w(:,2)/wi(:)
      if(CHL == 'V')VLx = VL
      if(CHR == 'V')VRx = VR
      if(Debug > 0)write(*,*)' EigenValGen: n, Info = ',n,Info
   End Function EigenValGen
!-------------------------------------------------------------------------------
   Function EigenValSym(A,B,v,typ) result(w)
!-------------------------------------------------------------------------------
      Real(float) :: A(:,:),B(:,:),w(size(A,1))
      Real(float), optional :: v(:,:)
      Integer, optional :: typ
      Real(float) :: Work(10*size(A,1))
      Integer :: n,Nwork,info,itype
!-------------------------------------------------------------------------------
!  EigenValSym computes the eigenvalues & eigenvectors of:
!    A*x = lambda*B*x
!  w  - eigenvalues returned
!  A  - matrix A (lower triangle) supplied, eigenvectors returned
!  B  - matrix B (lower triangle) supplied, overwritten
!-------------------------------------------------------------------------------
!  LaPack routine DSYGV computes all the eigenvalues, and optionally, the
!  eigenvectors of a real generalized symmetric-definite eigenproblem, of the
!  form (itype):
!     1. A*x=(lambda)*B*x,
!     2. A*Bx=(lambda)*x,
!     3. B*A*x=(lambda)*x.
!  Here A and B are assumed to be symmetric and B is also positive definite.
!-------------------------------------------------------------------------------
      n = size(A,1)
      Nwork = 10*N
      if(Present(typ))then
         itype = typ
      else
         itype = 1
      endif
      if(Present(v))then
         v = A
         Call DSYGV(itype,'V','L',N,v,N,B,N,W,Work,Nwork,Info)
      else
         Call DSYGV(itype,'V','L',N,A,N,B,N,W,Work,Nwork,Info)
      endif
      if(Debug > 0)write(*,*)' EigenValSym: n, Info = ',n,Info
   End Function EigenValSym
!-------------------------------------------------------------------------------
   Function EigenValTriSym(AD,AS) result(w)
      Real(float) :: AD(:),AS(:),w(size(AD))
      Real(float) :: zdum(1),zwrk(1)
      Integer :: n,info
      character :: CHV*1
!   Function EigenValTriSym(AD,AS,Vx) result(w)  <== to add option for eigenvalues
!      Real(float), optional :: Vx(:,:)
!-------------------------------------------------------------------------------
!  EigenValTriSym computes the eigenvalues and optionally the eigenvectors of a
!  symmetric tridiagonal matrix
!  w  - eigenvalues
!  AD - the matrix diagonal
!  AS - the matrix subdiagonal
!-------------------------------------------------------------------------------
!  LaPack routine DSTEQR computes for an N-by-N real nonsymmetric matrix A, the
      n = size(AD)
      CHV = 'N'
!     if(Present(Vx))CHV = 'V'
      call DSTEQR(CHV,n,AD,AS,zdum,1,zwrk,info)  ! in EigenLA
      w = AD
   End Function EigenValTriSym
   Function Sort_idx(arr)  result(idx)
      real(float), intent(in) :: arr(:)
      integer :: idx(size(arr))
      integer, parameter :: NN=15, NSTACK=50
      real(float) :: a
      integer :: n,k,i,j,idxt,jstack,l,r
      integer, dimension(nstack) :: istack
      n = size(arr)
      idx = (/(k,k=1,n)/)
      jstack=0
      l=1
      r=n
      do
         if (r-l < NN) then
            do j=l+1,r
               idxt=idx(j)
               a=arr(idxt)
               do i=j-1,l,-1
                  if (arr(idx(i)) <= a) exit
                  idx(i+1)=idx(i)
               end do
               idx(i+1)=idxt
            end do
            if (jstack == 0) RETURN
            r=istack(jstack)
            l=istack(jstack-1)
            jstack=jstack-2
         else
            k=(l+r)/2
            call swap(idx(k),idx(l+1))
            call icomp_xchg(idx(l),idx(r))
            call icomp_xchg(idx(l+1),idx(r))
            call icomp_xchg(idx(l),idx(l+1))
            i=l+1
            j=r
            idxt=idx(l+1)
            a=arr(idxt)
            do
               do
                  i=i+1
                  if (arr(idx(i)) >= a) exit
               end do
               do
                  j=j-1
                  if (arr(idx(j)) <= a) exit
               end do
               if (j < i) exit
               call swap(idx(i),idx(j))
            end do
            idx(l+1)=idx(j)
            idx(j)=idxt
            jstack=jstack+2
            if (jstack > NSTACK)Write(*,'(a)')'Sort_idx: NSTACK too small'
            if (r-i+1 >= j-l) then
               istack(jstack)=r
               istack(jstack-1)=i
               r=j-1
            else
               istack(jstack)=j-1
               istack(jstack-1)=l
               l=i
            end if
         end if
      end do
   contains
      subroutine icomp_xchg(i,j)
      integer, intent(inout) :: i,j
      integer :: swp
      if (arr(j) < arr(i)) then
         swp=i
         i=j
         j=swp
      end if
      end subroutine icomp_xchg
      subroutine swap(i,j)
         integer :: i,j,itmp
         itmp = i
         i = j
         j = itmp
      end subroutine swap
   end function sort_idx
End Module EigenValue
