Module BVPmod
!-------------------------------------------------------------------------------
!  Module BVPmod (just below) solves the problem
!  It solves a nonsymmetric reaction & diffusion problem using
!  Orthogonal Collocation with a choice of collocation points
!  and with Moments or Galerkin methods with various options
!  for integration of the rate terms.
!  Solves for Planar Geometry with no consideration of symmetry
!  For planar geometry the equations are:
!     d2y/dx2 + c(x)*(1-y) = 0
!  with y = 0 at x = 0,1
!  The effectiveness factor is calculated by three different
!  methods:
!     1. By integration to calculate the average reaction rate
!     2. Using the derivative of the solution at x = 0 & 1
!     3. Using the flux relation which corresponds to a natural
!        treatment of a flux boundary condition
!  c(x) = 4*q(x)*phi^2, phi is the Thiele Modulus (5 used in most examples)
!  q(x) is a cubic function with average value of 1, its value varies from
!  q(0) = Aparm, q(1) = 2 - Aparm, q(0.5) = 1
!  Aparm = .2 gives 0.2 at left, 1.8 at right (case in text)
!  Aparm = 1, gives no variation
! ----------------------------------------------------------------
   Use OrthogonalColloc, Only : ColDataType,Initialize,Xpoints,CollocCoef,MassMatrix,Lcoeff,
         Chebyshev,TxtType
   Use Array_Print, only : ArrayPrint, VectorPrint
   include 'defs.fi'
   Private
   Public SolveBVP
   Real(float), private, parameter :: Debug = 0
   integer, public, parameter :: nsolv = 4, nflux = 6
   integer :: ityp
   Real(float) :: Thparm,Thparm2             ! Thiele parameter & 4*Thiele^2 (constant in source)
   Real(float) :: Aparm, Aparm1              ! spatial variation parameters in source term
   Real(float) :: Biot(0:1),Biot2(0:1)       ! left & right side Biot numbers in flux condition
   Logical :: Dirichlet,ConstCoef,Symmetric  ! logicals Dirichlet B.C.?, Constant Coefficients?
   Type(ColDataType) :: OC,OX
   Integer, Parameter :: Nx = 20 ! too override these with Xex from Error routine
   real(float) :: xex(Nx)
!-------------------------------------------------------------------------------
   Contains
!-------------------------------------------------------------------------------
Subroutine SolveBVP(n,itypx,Thx,Apx,bcfx)
   Use LUSolvers, Only : LUSolveC, LUSolveSQ
   Integer, intent(in) :: n      ! interior points
   Integer :: itypx               ! Gauss = 1, Lobatto = 2
   real(float) :: Thx            ! average Thiele parameter
   Real(float) :: Apx            ! left side reactivity, right is 2 - Aparm
   Real(float) :: bcfx(2)        ! left and right side Biot number in flux condition
   Real(float) :: flux(0:nflux,nsolv)     ! flux
   Real(float) :: Ac(0:n+1,0:n+1),Bc(0:n+1,0:n+1),Cc(0:n+1,0:n+1),Dc(0:n+1,0:n+1)
   Real(float) :: xc(0:n+1),wc(0:n+1),yc(0:n+1,nsolv),rc(0:n+1),rs(0:n+1)
   Real(float) :: Ax(0:n+1,0:n+1),Af(0:n+1,0:n+1)
   Integer :: i,n1,i0,i1,k,mx
   character*(18) :: txtmass
   ! ---------------------------------------------------------------------------
   ! This program solves diffusion/conduction with linear source term.  The source
   ! term may have a variable coefficient as determined by parameter Apx.
   ! The problem is solved by 4 different methods:
   !  1. conventional orthogonal collocation (pseudospectral)
   !  2. equivalent formulation with symmetric matric, corrected B.C. treatement
   !  3. Moments (Gauss) or Galerkin (Lobatto) with interpolated souce terms
   !  4. Moments (Gauss) or Galerkin (Lobatto) with exact integration
   ! ---------------------------------------------------------------------------
   !  Problem parameters
   n1 = n+1
   ityp = itypx
   Thparm = Thx
   Thparm2 = (Thx*2.0_float)**2        ! 2.0 to account for full slab
   Aparm = Apx; Aparm1 = (1.0_float - Aparm)*4.0_float
   Aparm = Thparm2*Aparm;  Aparm1 = Thparm2*Aparm1
   Biot = bcfx                         ! left & right BC Biot (use 0 for Dirichlet)
   Biot2 = Biot*2.0_float              ! 2.0 to account for full slab
   ConstCoef = abs(Aparm1) < 1.0e-8    ! Constant coefficients
   Dirichlet = maxval(Biot) < 1.0e-8   ! Dirichlet or 3rd kind
   Symmetric = .not.(ityp == Chebyshev)
   if(Dirichlet)then
   ! for Dirichlet conditions, eqns are set up at 0 and n+1, but not used
      i0 = 1;  i1 = n   ! Dirichlet y(0) = y(n+1) = 0
   else
      i0 = 0;  i1 = n1  ! 3rd kind, solve for y(0) & y(n+1)
   endif
   !-------  Get the collocation data
   Call Initialize(OC,n,ityp)
   Call CollocCoef(OC,xc,wc,Ac,Bc,Cc)
   Call Initialize(OX,nx-2,ityp)
   xex = Xpoints(OX) ! for interpolated results
   rc = FuncRX(xc)
   Call ReportBasic(xc,wc,rc,Ac,Bc,Cc,ityp)
   ! ------- Create Solution Matrix & Solve, 4 different methods
   yc(0,:) = 0.0_float; yc(n1,:) = 0.0_float
   ! 1. ------ Conventional orthogonal collocation -----------------------------
   ! use first derivative -Ac(0)*y = Bi*y(0), etc
   do i=0,n+1
      Ax(i,:) = Bc(i,:)
      Ax(i,i) = Ax(i,i) - rc(i)
   enddo
   rs = -rc
   Ax(0,:) = -Ac(0,:);  Ax(0,0) = Ax(0,0) + Biot2(0);     rs(0) = 0
   Ax(n1,:) = Ac(n1,:); Ax(n1,n1) = Ax(n1,n1) + Biot2(1); rs(n1) = 0
   yc(i0:i1,1) = LUSolveC(Ax(i0:i1,i0:i1),rs(i0:i1))     ! Use Nonsymmetric Solver
   flux(:,1) = FluxCalc(rc,wc,Ac,Bc,Ax,rs,yc(:,1),1)
   call ArrayPrint("Matrix & RHS, conventional:",Ax(i0:i1,i0:i1),rs(i0:i1))
   if(Debug > 0)Call VectorPrint("Flux",flux(:,1))
   ! ------ Solve with 3 method differing only by mass matrix
   !  2. collocation, diagonal mass matrix
   !  3. interpolation of rates, full mass matrix
   !  4. integration by quadrature, m = n + mx quadrature points
   mx = 2   ! 2 gives exact, 1 approx., 0 collocation
   do k=2,4
      select case(k)
      case(2)
         txtmass = 'collocation (weak)'
         Dc = 0.0_float
         rs = rc*Wc
         forall(i=0:n+1)Dc(i,i) = rs(i)
      case(3)
         txtmass = 'interpolated rates'
         Dc = MassMatrix(OC,rc)  ! interpolated mass matrix
      case(4)
         txtmass = 'exact integration'
         Dc = MassMatrix(OC,FuncRX,mx) ! full mass matrix mx extra points
      end select
      if(Debug > 0)call ArrayPrint("Mass Matrix, "//txtmass,Dc)
      forall(i=0:n+1)rs(i) = sum(Dc(i,:)) ! = wc*rc except small n
      Ax = Cc + Dc
      Af = Ax
      Ax(0,0) = Ax(0,0) + Biot2(0)
      Ax(n1,n1) = Ax(n1,n1) + Biot2(1)
      yc(i0:i1,k) = LUSolveC(Ax(i0:i1,i0:i1),rs(i0:i1))
      flux(:,k) = FluxCalc(rc,wc,Ac,Bc,Af,rs,yc(:,k),k)
      call ArrayPrint("Matrix & RHS, "//txtmass,Ax(i0:i1,i0:i1),rs(i0:i1))
      if(Debug > 0)Call VectorPrint("Flux",flux(:,k))
      enddo
   Call Report(xc,yc,flux)
End Subroutine SolveBVP
! ----------------------------------------------------------------------
Function FluxCalc(rc,wc,Ac,Bc,Ax,rs,y,meth) result(f)
   Real(float) :: rc(0:),wc(0:),Ac(0:,0:),Bc(0:,0:),Ax(0:,0:),rs(0:),y(0:)
   Real(float) :: f(0:nflux),rx(0:size(y)-1)
   integer :: n,meth
   n =  size(y) - 2
   rx = rc*(1.0_float - y)
   f(0) = sum(wc*rx)       ! total flux, material balance
   f(4) = sum(Ac(0,:)*y)   ! first derivative, wrong for all but Gauss
   f(5) = -sum(Ac(n+1,:)*y)
   f(6) = f(4) + f(5)
   if(meth == 1)then
   ! correction for conventional collocation
   ! no correction for Gauss, since endpoint wc = 0
      f(1) = f(4) + wc(0)*(rx(0) + sum(Bc(0,:)*y))
      f(2) = f(5) + wc(n+1)*(rx(n+1) + sum(Bc(n+1,:)*y))
   else
   ! calculation from equation coefficients (all but conventional)
      f(1) = rs(0) - sum(Ax(0,:)*y)
      f(2) = rs(n+1) - sum(Ax(n+1,:)*y)
   endif
   f(3) = f(1) + f(2)   ! this total equals f(0)
End Function FluxCalc
! ----------------------------------------------------------------------
Function FuncRX(x)  result(f)
   real(float), intent(in) :: x(:)
   real(float) :: f(size(x))
   ! calculate cubic variation of rate coefficient
   f = Aparm + Aparm1*(1.5_float - x)*x*x
End Function
!-------------------------------------------------------------------------------
Subroutine ReportBasic(xc,wc,rc,Ac,Bc,Cc,ityp)
   real(float) :: xc(:),wc(:),rc(:),Ac(:,:),Bc(:,:),Cc(:,:)
   integer :: n,ityp
   n = size(xc) - 2
   write(iout,'(/,3a,i3,a)') &
      "Solution using ",trim(TxtType(ityp))," points with",n," interior points"
   write(iout,'(a,2f10.4)') &
      'For Thiele Modulus and variation parameters of:',Thparm,Aparm/Thparm2
   if(.not.Dirichlet)write(iout,'(A,2F8.2)')'Left and Right Biot Nos:',Biot
   Call ArrayPrint('Quadrature points & weights, & Reactivity:',xc,wc,rc)
   if(Debug > 0)then
      Call ArrayPrint('A - 1st Derivative Operator',Ac)
      Call ArrayPrint('B - 2nd Derivative Operator',Bc)
      Call ArrayPrint('C - 2nd Derivative stiffness',Cc)
   endif
End Subroutine ReportBasic
! -----------------------------------------------------------------------
Subroutine Report(xc,y,flux)
   real(float) :: xc(0:),y(0:,:)
   real(float) :: flux(0:,:)
   real(float) :: rqc(0:size(xc)-1)
   Real(float) :: yx(nx,nsolv),rqx(nx)
   integer :: n,i,j
   Real(float) :: Lc(nx,size(xc))
   Character (len=*), Parameter :: fmtv = '(25(f14.10,a1))'
   Character (len=*), Parameter :: fmtx = '(a,10(a1,f11.9))'
   Character (len=9), parameter :: titlex(4)= &
      (/"Collocat.","Symmetric","Interp. r","Gal./Mom."/)
   Character (len=*), Parameter :: title = tb//"   x    "//tb// &
      "   Collocation"//tb//"   Symmetric"//tb//" Interpolated r"//tb//"Exact Mass"
   Character (len=*), Parameter :: titlef = tb//tb//"   Integrated"//tb//tb//" Left"//tb//tb// &
      "Right"//tb//tb//"Total"//tb//tb//" Left"//tb//tb//"Right"//tb//tb//"Total"
   ! -----------------------------------
   n = size(xc) - 2
   flux = flux/Thparm2  ! normalize fluxes
   rqx = FuncRx(xex);  rqc = FuncRx(xc)
   ! ----- calculate interpolated solutions
   Lc = Lcoeff(xex,OC)
   do j=1,nsolv
      yx(:,j) = MatMul(Lc,y(:,j))
   enddo
   ! This output allows side by side comparison of 4 methods
   write(iout,'(/a)')"Results, four methods:"
   Call ArrayPrint(title,xc,y,fmtx=fmtv)
   write(iout,'(/a)')"Interpolated results:"
   Call ArrayPrint(title,xex,yx,fmtx=fmtv)
   write(iout,'(/a)')"flux calculations, integrated, correct & 1st derivative:"
   write(iout,'(A)')titlef
   do j=1,nsolv
      write(iout,fmtx)titlex(j),(tb,flux(i,j),i=0,nflux)
   enddo
End Subroutine Report
!-------------------------------------------------------------------------------
End Module BVPmod
!-------------------------------------------------------------------------------
Program Example01  ! Nonsymmetric BVP - particle problem
   Use OrthogonalColloc, Only : Gauss,Lobatto,Chebyshev,RadauR,RadauL
   Use BVPmod, Only : SolveBVP, nsolv
   Use Array_Print, Only : OpenFile
   include 'defs.fi'
   Integer :: ityp,n,Nmin,Nmax,Ndel   ! type, no., min, max, delta coll. points
   Character :: fname*(80),ch*1
   integer, parameter :: MaxNmax = 50
   real(float) :: Thparm,Aparm,BCflux(2)
   logical :: variable
   Character (len=*), Parameter :: fmtf ='(i3,12(a1,f16.12))'
   ! ----------------------------------------------------------------
   ityp = Gauss;  Nmin = 2;   Nmax = 2;   Ndel = 1
   ThParm = 10.0_float;   AParm = 1.0_float; BCflux = 0.0_float
   write(*,'(a)')' Enter Thiele Modulus and Biot no.>'
   read(*,*)Thparm, BCflux(1)
   write(*,'(a)')' Variable coefficient (T/F)? >'
   read(*,'(L5)')variable
   if(variable)Aparm = 0.2_float
   BCflux(2) = BCflux(1)  ! could be different
   write(*,'(a,i2,a,i2,a)')' Enter g/l/c/r/x for Gauss/Lobatto/Chebyshev/Radau(R/L) >'
   read(*,*)ch;  ityp = index('glcrx',ch);  ityp = Max(Gauss,Min(RadauL,ityp))
   write(*,'(a)')' Enter min, max & delta N>'
   read(*,*)Nmin,Nmax,Ndel
   Nmin = max(Nmin,1)
   Nmax = min(max(Nmin,Nmax),MaxNmax)
   write(*,'(A)')' Enter output file name>'
   read(*,*)fname
   Call OpenFile(fname)
   ! ---------------------------------------------------
   do n=Nmin,Nmax,Ndel
      Call SolveBVP(n,ityp,Thparm,Aparm,BCflux)
   enddo
End Program Example01

