Module SymBVP
   ! ---------------------------------------------------
   !  Solves symmetric Reaction-Diffusion Problem with nonlinear reaction
   !  Using Orthogonal Collocation with Gauss, Lobatto or Chebyshev points
   !  Solves for Planar, Cylindrical and Spherical Geometry
   !  For planar geometry the equations are:
   !     d(Ddy/dx)/dx + r(y) = 0
   !     Ddy/dx + Bi*y = 0  at x = 1
   !  or y = 0 at x = 1  (when Bi = 0)
   !     dy/dx = 0 at x = 0  (symmetry)
   !  the reaction rate is of the form:
   !     (thiele^2)*(1 - y)**k/(1 - Kd*y)**2
   !  Problem is solved by 2 different methods: (set in GetData)
   !     0.  Conventional Orthogonal Collocation, but with correct
   !         treatment for flux boundary condition. A simple code
   !         modification will give boundary collocation.
   !     1. A weak Galerkin type treatment which gives a symmetric
   !        solution matrix (except for Chebyshev).  It corresponds to
   !        multiplication of the conventional equations by -Wc.
   !  The effectiveness factor is calculated by 4 methods:
   !     1. By integration to calculate the average reaction rate
   !     2. Using derivative corrected by boundary residual
   !     3. Using Bi*y(n+1) boundary value
   !     4. First derivative
   !  Methods 1-3 are identical, all 4 are the same for Gauss points
   ! ---------------------------------------------------
   Use OrthogonalColloc, only : ColDataType,Initialize,DataDelete,CollocCoef,Xpoints,WeightQ,Interp, &
      TxtType, TxtGeom
   Use Array_Print, only : ArrayPrint, VectorPrint
   Private
   Public SolveBVP
   Public BVP_Data
   include 'defs.fi'
   Type BVP_Data
      integer     :: typ      ! Gauss, Lobatto, etc.
      integer     :: geom     ! geometric factor 0,1,2
      integer     :: Korder   ! reaction order (numerator)
      real(float) :: Kads     ! reaction, adsorption term
      real(float) :: Thparm   ! Thiele modulus
      real(float) :: Bi       ! constant in flux b.c.
      real(float) :: yg       ! initial guess value
   End Type BVP_Data
   Logical, Parameter :: Conventional = .false. ! conventional formulation?
   Logical, Parameter :: Bcolloc = .false.      ! use boundary colloc. with conventional?
   Logical, Parameter :: Generalized = .true.   ! to supply generalized Thiele values
   Integer :: typ, igeom   ! type points, geometry
   integer :: Korder       ! reaction rate order
   Real(float) :: Thparm   ! thiele parameter (squared)
   Real(float) :: Thfac    ! factor for genearalized theile parameter
   Real(float) :: Kd       ! reaction adsorption term
   Real(float) :: gfac     ! geometric factor
   integer, parameter :: maxiter = 60
   Integer, public, parameter :: Debug = 0
   Integer, Parameter :: Nv = 19          ! for interpolated solution
   Integer, Parameter :: igx = (Nv-1)/2  ! position of guess (this is bad coding, but expedient)
   Real(float) :: xv(Nv),wv(Nv),yv(Nv),rv(Nv),Resid(Nv),exff(2)
!-------------------------------------------------------------------------------
   Contains
!-------------------------------------------------------------------------------
   Subroutine SolveBVP(n1,d,y,eff)
      Use LUSolvers, Only : LUSolveC, LUSolveSym
      Type(BVP_Data) :: d
      integer :: n1
      Real(float) :: Bi,y(:),eff(:)
      Type(ColDataType) :: OC,OV
      Real(float) :: xc(n1)         ! collocation points
      Real(float) :: wc(n1)         ! quadrature weights
      Real(float) :: Ac(n1,n1)      ! symmetric derivative operator
      Real(float) :: Bc(n1,n1)      ! 2nd derivative (Laplacian) operator
      Real(float) :: Cc(n1,n1)      ! symmetric Laplacian operator
      Real(float) :: A(n1,n1)
      Real(float) :: wn1,An1n1
      Real(float) :: yn(n1),r(n1),dr(n1),b(n1),ygx(2)
      Real(float):: xgx
      Real(float) :: dymax, eps
      integer :: n,i,kiter,ns,itx
      Character (len=*), Parameter :: fmt17 = '(20(f21.17,a1))'
      Character (len=*), Parameter :: fmt22 = '(20(es22.15,a1))'
      ! ----------------------------------------------------------------
      typ = d%typ
      igeom = d%geom
      Bi = d%Bi
      Korder = d%Korder
      Kd = d%Kads
      if(Kd > 0.0 .and. Korder == 1)then
         Thfac = sqrt(-2.0_float*(Kd + log(1.0_float-Kd)))/Kd     ! 1st order, Kd > 0
      elseif(Kd == 0.0)then
         Thfac = 2.0_float/sqrt(Real(Korder+1,float)*2.0_float)   ! Kth order Kd = 0
      else
         Thfac = 1.0_float                                        ! others? check rate routine
      endif
      gfac = Real(igeom,float)
      Thfac = Thfac*gfac
      Thparm = (d%Thparm)**2        ! raw values supplied
      if(Generalized)Thparm = Thparm*Thfac**2
      n = n1 - 1
      eps = Epsilon(dymax)*1.0e5_float
      Call Initialize(OC,n,typ,igeom)
      Call CollocCoef(OC,xc,wc,Ac,Bc,Cc)
      Call Initialize(OV,Nv-1,typ,igeom)
      xv = Xpoints(OV)  ! xv for tabular output
      wv = WeightQ(OV)  ! wv for norms
      xgx = xv(igx)
      y = yguess(d%yg)  ! initial guess
      itx = 1           ! guess counter
      if(d%yg > 0.5_float)then ! guess after failure
         ygx = (/0.05,0.95/)  ! try small next
      else
         ygx = (/0.95,0.05/)  ! try large next
      endif
      ns = n
      if(Bi > 0.0_float)ns = n1
      if(Conventional)then ! conventional
         A(:n,:) = Bc(:n,:n1)
         wn1 = wc(n1)
         if(BColloc)wn1 = 0.0  ! mod for boundary collocation
         A(n1,:) = Ac(n1,:) - wn1*Bc(n1,:)
         An1n1 = A(n1,n1)
      else                 ! weak, symmetric
         A = Cc
      endif
      do kiter=1,maxiter
         r = RateRxn(y,dr)
         if(Conventional)then ! conventional formulation
            do i=1,n
               A(i,i) = Bc(i,i) + dr(i)
               b(i) = -r(i) + dr(i)*y(i)
            enddo
            if(Bi > 0.0_float)then
               A(n1,n1) = An1n1 - wn1*dr(n1) + Bi
               b(n1) = wn1*(r(n1) - dr(n1)*y(n1))
            endif
         else                 !  weak, symmetric formulation
            do i=1,n1
               A(i,i) = Cc(i,i) - wc(i)*dr(i)
               b(i) = wc(i)*(r(i) - dr(i)*y(i))
            enddo
            exff = 0
            if(Bi > 0.0_float)then
               A(n1,n1) = A(n1,n1) + Bi
            endif
         endif
         yn(:ns) = LUSolveC(A(:ns,:ns),b(:ns))   ! symmetric solve not exploited
         dymax = maxval(abs(yn(:ns) - y(:ns)))
         if(itx < 3 .and. (kiter > 10*itx .and. &  ! failed, try alternate guess
               (dymax > 0.01 .or. y(1) > 1.1 .or. y(1) < 0.01)))then
            yn = yguess(ygx(itx));  itx = itx + 1
            write(*,'(a,12f10.7)')'New Guess:',yn
         endif
         y(:ns) = yn(:ns)
         Call Debugx()
         if(dymax < eps)exit
      enddo
      r =RateRxn(y,dr)
      eff = FluxCalc(y,r,wc,Ac,Cc,Bi)
      Call ResidCalc(OC,xc,Bc,y)
      Call Report(typ,igeom,Bi,y,xc,r,eff)
      d%yg = yv(igx) ! save the value of the guess
      ! ------------------------------------
      contains
      ! ------------------------------------
      Function yguess(yg) result(y)
         real(float) :: yg,y(size(xc)),fac
         fac = 1.0_float/(1.0_float - xgx*xgx)
         y = yg*(1.0_float - xc**2)*fac
         y = min(y,0.999_float)
      End Function yguess
      ! ------------------------------------
      Subroutine Debugx
         if(kiter == 1)then
            if(Debug > 0)write(ilog,'(2a,i2,a,2f8.4,a)')trim(TxtType(typ)),' n =',n, &
               ' Thiele =',Sqrt(Thparm)/Thfac,Sqrt(Thparm),' (general, actual)'
            write(*,'(2a,i2,a,2f8.4,a)')trim(TxtType(typ)),' n =',n,' Thiele =', &
               Sqrt(Thparm)/Thfac,Sqrt(Thparm),' (general, actual)'
         endif
         if(Debug > 3 .and. kiter == 1)then
            Call Arrayprint('xc,wc',xc,wc,flog=.true.)
            Call Arrayprint('Ac',Ac,flog=.true.)
            Call Arrayprint('Bc',Bc,flog=.true.)
            Call Arrayprint('Cc',Cc,flog=.true.)
         endif
         if(Debug > 3) &
            call ArrayPrint('Solution Matrix, rhs, solution',A,b,y,flog=.true.)
         if(Debug > 0)write(ilog,'(a,i3,E11.2,f8.4)')'Iter, Max(Delta y):',kiter,dymax,y(1)
         write(*,'(a,i3,E11.2,f8.4)')'Iter, Max(Delta y):',kiter,dymax,y(1)
      End Subroutine Debugx
   End Subroutine SolveBVP
! -----------------------------------------------------------------------
   Function RateRxn(y,dr)       result(r)
! -----------------------------------------------------------------------
      real(float) :: y(:),dr(:),r(size(y)),d
      real(float), save :: rord, drext
      Logical, save :: First = .True.
      integer i
      if(First)then
         First = .False.  ! calculation performed only once
         rord = Real(Korder,float)
         drext = 0.0_float
         if(Korder == 1)drext = -1.0_float/(1.0_float - Kd)**2
      endif
      do i=1,size(y)
         if(y(i) < 1.0_float)then
            d = 1.0_float/(1.0_float - Kd*y(i))
            r(i) = (1.0_float - y(i))**Korder
            dr(i)= -rord*(1.0_float - y(i))**(Korder-1) + 2.0_float*r(i)*Kd*d
            r(i) = r(i)*(Thparm*d**2)
            dr(i) = dr(i)*(Thparm*d**2)
         else
            dr(i) = drext*Thparm
            r(i) = dr(i)*(y(i) - 1.0_float)
         endif
      enddo
      if(Debug > 1)then
         write(ilog,'(" y:",10F15.10)')y
         write(ilog,'(" r:",10F15.10)')r
         write(ilog,'("dr:",10G15.7)')dr
      endif
   End Function RateRxn
! -----------------------------------------------------------------------
   Function FluxCalc(y,r,wc,Ac,Cc,Bi)      result(eff)
! -----------------------------------------------------------------------
      real(float) :: y(:),r(:),wc(:),Ac(:,:),Cc(:,:),Bi
      real(float) :: eff(4)
      integer :: n1
      n1 = size(y)
      eff(1) = Sum(wc*r)         ! integrate
      eff(2) = wc(n1)*r(n1) - Sum(Cc(n1,:)*y)   ! using weak form
      eff(3) = Bi*y(n1)          ! use boundary value
      eff(4) = -Sum(Ac(n1,:)*y)  ! derivative
      eff = eff*gfac/Thparm
   End Function FluxCalc
! -----------------------------------------------------------------------
   Subroutine ResidCalc(OC,xc,Bc,y)
! -----------------------------------------------------------------------
      Type(ColDataType) :: OC
      real(float) :: xc(:),y(:),Bc(:,:),d2c(size(xc))
      ! This calculation only valid for constant diffusion
      yv = Interp(xv,OC,y)
      rv = RateRxn(yv,Resid)        ! rates
      d2c = MatMul(Bc,y)            ! 2nd derivative at points
      Resid = Interp(xv,OC,d2c)     ! Interpolate 2nd derivative
      Resid = Resid + rv            ! Residual
   End Subroutine ResidCalc
! -----------------------------------------------------------------------
   Subroutine Report(typ,igeom,Bi,y,xc,r,eff)
! -----------------------------------------------------------------------
      Integer :: typ,igeom
      Real(float) :: Bi
      Real(float) :: y(:),xc(:),r(:),eff(:)
      integer :: n,n1,ns
      character*(*), parameter :: fmtf='(2(f15.12,a1),4(g15.7,a1))'
      ! -------------------------------
      n1 = size(xc)
      n = n1 - 1
      write(iout,'(a,i2,a,L1,a,L1)')'Results for N =',N,' Conventional? ',Conventional, &
         ' Boundary Collocation? ',BColloc
      write(iout,'(3a)'  )'Solution with ',trim(TxtType(typ)),' Points'
      write(iout,'(2a)'  )'The Geometry is ',trim(TxtGeom(igeom))
      write(iout,'(a,f8.2)')'Flux Constant  :',Bi
      write(iout,'(a,2f9.5)')'Thiele Modulus(general,actual) :',Sqrt(Thparm)/Thfac,Sqrt(Thparm)
      write(iout,'(a,I8)'  )'Reaction Order :',Korder
      write(iout,'(a,f8.2)')'Adsorption Parm:',Kd
      Call ArrayPrint('Collocation Points, Solution, rates:',xc,y,r,fmtx=fmtf )
      Call ArrayPrint('Interpolated Solution, rate, residual:',xv,yv,rv,Resid,fmtx=fmtf)
      Call VectorPrint('Effectiveness Factor - Integrated, Corrected Derivative, Difference, Derivative:',eff)
      if(N > 6)Return
      ns = n
      if(Bi > 0.0_float)ns = n1
   End Subroutine Report
End Module SymBVP
! -----------------------------------------------------------------------
Program Example02
! -----------------------------------------------------------------------
   Use SymBVP, Only : SolveBVP, BVP_Data
   Use OrthogonalColloc, Only : Gauss, Lobatto, Planar, Cylindrical, Spherical
   include 'defs.fi'
   Integer, Parameter :: MaxNmax = 100, MaxTmax = 100
   Integer :: Nc(3),nnc,n,n1
   Integer :: nt,nthp
   integer :: i
   Real(float) :: Thp(3)
   Real(float) :: y(MaxNmax)           ! solution
   Real(float) :: effx(4)
   Type(BVP_Data) :: d
   Real(float), allocatable :: eff(:,:),Thpx(:)
   ! ---------------------------------------------------
   ! Main for Symmetric Reaction-Diffusion problem
   ! Note: can run using Ex02 < Ex02.dat if you put
   ! data in file Ex02.dat, otherwise will be interactive
   ! Code is set up to loop over either:
   !   N - collocation points to check accuracy
   !   Thiele No. - either forward or backward to track a solution
   ! ---------------------------------------------------
   call GetData(d,Nc,Thp)
   nthp = nint((Thp(2) - Thp(1))/Thp(3)) ! number of Thiele parameters
   nnc = (Nc(2) - Nc(1))/Nc(3) + 1                     ! no. differenet n
   allocate(eff(nnc,0:nthp))
   allocate(Thpx(0:nthp))
   ! ------------------
   Thpx = (/((Thp(1) + Thp(3)*i),i=0,nthp)/) ! thiele parameters solved
   do nt = 0,nthp
      d%Thparm = Thpx(nt)
      i = 0
      do n=Nc(1),Nc(2),Nc(3)
         i = i + 1
         n1 = n + 1
         Call SolveBVP(n1,d,y(:n1),effx)
         eff(i,nt)=effx(1)
      enddo
   enddo
   write(iout,'(a)')'Summary: Thparm  n   eff. factor'
   do nt = 0,nthp
      i = 0
      do n=Nc(1),Nc(2),Nc(3)
         i = i + 1
         write(iout,'(f8.3,a1,i2,a1,F18.14,a1,G12.5)')Thpx(nt),tb,n,tb,eff(i,nt)
      enddo
   enddo
End Program Example02
Subroutine GetData(d,Nc,Thp)
   Use SymBVP, Only : BVP_Data, Debug
   Use OrthogonalColloc, Only : Gauss, Lobatto, Chebyshev, Planar, Cylindrical, Spherical
   Use Array_Print, Only : OpenFile
   include 'defs.fi'
   Type(BVP_Data) :: d
   integer :: Nc(3)
   Real(float) :: Thp(3)
   Character :: fname*80,ch*1
   Integer, Parameter :: MaxNmax = 100
   ! ---------------------------
   ! typ    - 1 Gauss, 2 Lobatto, 3 Chebyshev
   ! geom   - geometric factor 0,1,2
   ! Korder - reaction order (numerator)
   ! Kd   - reaction, adsorption term
   ! Thparm - Thiele modulus
   ! Bi  - constant in flux b.c.
   ! yg     - initial guess
   ! reaction rate = Thparm**2*(1 - y)**Korder/(1 - Kd*y)**2
   ! ---------------------------
   ! default values
   Nc = 2
   Thp = 1.0_float
   d%typ = Lobatto
   d%geom = Planar
   d%Bi = 0.0_float
   d%Thparm = 1.0_float
   d%Korder = 1
   d%Kads = 0.0_float
   ! ---------------------------
   write(*,'(a,i2,a,i2,a)')' Enter g/l/c for Gauss/Lobatto/Chebyshev >'
   read(*,*)ch;  d%typ = index('glc',ch);  d%typ = Max(Gauss,Min(Chebyshev,d%typ))
   write(*,'(a)')' Enter n first, last and delta >'
   read(*,*)Nc
   Nc(1) = max(Nc(1),1)
   Nc(2) = min(max(Nc(1),Nc(2)),MaxNmax)
   write(*,'(a)')'Enter 1, 2 or 3 for Planar, Cylindrical or Spherical Geometry > '
   read(*,*)d%geom
   write(*,'(a)')' Enter Boundary Flux Constant (use 0 or / for specified value) > '
   read(*,*)d%Bi
   write(*,'(a)')' Enter reaction order, adsorption term > '
   read(*,*)d%Korder,d%Kads
   d%Kads = Max(0.0_float,Min(d%Kads,0.9999999_float))  ! 0 <= Kads < 1
   write(*,'(a)')' Thiele parameter first, last, delta >'
   read(*,*)Thp
   write(*,*)' Enter initial guess, recommend 0.3 lower, 0.8 upper'
   read(*,*)d%yg
   write(*,'(A)')' Enter root of output file name>'
   read(*,*)fname
   Call OpenFile(fname)
   if(Debug > 0)Call OpenFile(fname,'log')
End Subroutine GetData
