Module SymBVP
   ! ---------------------------------------------------
   !  Solves symmetric the Reaction-Diffusion Problem with nonlinear reaction.
   !  This is a more complex version of the code ex02s.f90. It can do Galerkin as well as
   !  collocation methods. It also calculates transforms of the residual (see section 3.1.6)
   !  Solves for Planar, Cylindrical and Spherical Geometry
   !  For planar geometry the equations are:
   !     d(Ddy/dx)/dx + r(y) = 0
   !     Ddy/dx + Bi*y = 0  at x = 1
   !  or y = 0 at x = 1  (when Bi = 0)
   !     dy/dx = 0 at x = 0  (symmetry)
   !  the reaction rate is of the form:
   !     (thiele^2)*(1 - y)**k/(1 - Kd*y)**2
   !  Problem is solved by 2 different methods: (set in GetData)
   !     0.  Conventional Orthogonal Collocation, but with correct
   !         treatment for flux boundary condition. A simple code
   !         modification will give boundary collocation.
   !     1. A weak Galerkin type treatment which gives a symmetric
   !        solution matrix (except for Chebyshev).  It corresponds to
   !        multiplication of the conventional equations by -Wc.
   !  The effectiveness factor is calculated by 4 methods:
   !     1. By integration to calculate the average reaction rate
   !     2. Using derivative corrected by boundary residual
   !     3. Using Bi*y(n+1) boundary value
   !     4. First derivative
   !  Methods 1-3 are identical, all 4 are the same for Gauss points
   ! ---------------------------------------------------
   Use OrthogonalColloc, only : ColDataType,Initialize,DataDelete,CollocCoef,Xpoints,WeightQ, &
      MassMatrix,Interp,Lcoeff,Gauss,Lobatto,Chebyshev,TxtType, TxtGeom
   Use Orth_Poly, only : Jacobi_Transform,Polyset,Jacobi_Scale_Sym
   Use Array_Print, only : ArrayPrint, VectorPrint
   Private
   Public SolveBVP
   Public ResFunc
   Public BVP_Data
   include 'defs.fi'
   Type BVP_Data
      integer     :: typ      ! Gauss, Lobatto, etc.
      integer     :: geom     ! geometric factor 0,1,2
      integer     :: Korder   ! reaction order (numerator)
      real(float) :: Kads     ! reaction, adsorption term
      real(float) :: Thparm   ! Thiele modulus
      real(float) :: Bi       ! constant in flux b.c.
      real(float) :: yg       ! initial guess value
   End Type BVP_Data
   Type(ColDataType) :: OC,OV
   Integer :: typ, igeom   ! type points, geometry
   integer :: Korder       ! reaction rate order
   Real(float) :: Thparm   ! thiele parameter (squared)
   Real(float) :: Thfac    ! factor for genearalized theile parameter
   Real(float) :: Kd       ! reaction adsorption term
   Real(float) :: Bi       ! Biot No.
   Real(float) :: gfac     ! geometric factor
   Logical, Parameter :: Generalized = .true.  ! to supply generalized Thiele values
   integer, Parameter :: method = 2    ! 0,1,2 conventional col., weak col., Galerkin
   integer, parameter :: maxiter = 50  ! max iterations
   Integer, parameter :: Debug = 1     ! increase for extra to .log file
   integer, parameter :: nxtra = 1   ! extra quadrature points in Galerkin method, > 0
   Integer, Parameter :: Nv = 19     ! for interpolated solution
   Real(float) :: xv(Nv),wv(Nv),yv(Nv),rv(Nv),wrv(Nv),Resid(Nv),apoly(Nv-1,2)
   Real(float) :: dymax, eps,ygx(2),exff(2),abeta(2)
   integer :: n,ns
   Integer, Parameter :: igx = (Nv-1)/2  ! position of guess (this is bad coding, but expedient)
   Real(float) :: xgx
!-------------------------------------------------------------------------------
   Contains
!-------------------------------------------------------------------------------
   Subroutine SolveBVP(n1,d,y,eff,Rerr)
      Use LUSolvers, Only : LUSolveC, LUSolveSym
      Type(BVP_Data) :: d
      integer :: n1
      Real(float) :: y(:),eff(:),Rerr(2)
      Real(float) :: yn(n1),r(n1),dr(n1)
      real(float) :: drq(n1+nxtra)
      Real(float) :: xc(n1)         ! collocation points
      Real(float) :: wc(n1)         ! quadrature weights
      Real(float) :: Ac(n1,n1)      ! symmetric derivative operator
      Real(float) :: Bc(n1,n1)      ! Laplacian
      Real(float) :: Cc(n1,n1)      ! symmetric Laplacian operator
      ! ----------------------------------------------------------------
      typ = d%typ
      if(Method==2)typ = Lobatto
      igeom = d%geom
      Bi = d%Bi
      Korder = d%Korder
      Kd = d%Kads
      if(Kd > 0.0 .and. Korder == 1)then
         Thfac = sqrt(-2.0_float*(Kd + log(1.0_float-Kd)))/Kd     ! 1st order, Kd > 0
      elseif(Kd == 0.0)then
         Thfac = 2.0_float/sqrt(Real(Korder+1,float)*2.0_float)   ! Kth order Kd = 0
      else
         Thfac = 1.0_float                                        ! others? check rate routine
      endif
      gfac = Real(igeom,float)
      Write(ilog,'(a,10g14.7)')'d%Thparm,Thfac,gfac:',d%Thparm,Thfac,gfac
      Thfac = Thfac*gfac
      Thparm = (d%Thparm)**2        ! raw values supplied
      if(Generalized)Thparm = Thparm*Thfac**2
      n = n1 - 1
      eps = Epsilon(dymax)*1.0e5_float
      Call Initialize(OC,n,typ,igeom)
      Call CollocCoef(OC,xc,wc,Ac,Bc,Cc)
      Call Initialize(OV,Nv-1,typ,igeom)
      xv = Xpoints(OV)
      wv = WeightQ(OV)
      abeta = Polyset(typ,igeom)
      xgx = xv(igx)
      yn = yguess(d%yg) ! initial guess
      if(d%yg > 0.5_float)then ! guess after failure
         ygx = (/0.05,0.95/)  ! try small next
      else
         ygx = (/0.95,0.05/)  ! try large next
      endif
      ns = n
      if(Bi > 0.0_float)ns = n1
      Select case(method)
      case default
         y = SolveColloc(yn)
      case(2)
         y = SolveGalerkin(yn)
      end select
      r =RateRxn(y,dr)
      eff = FluxCalc(y,r,wc,Ac,Cc,Bi)
      Rerr = ResidCalc(xc,Bc,y)
      Call Report(typ,igeom,Bi,y,xc,r,eff,Rerr)
      d%yg = yv(igx) ! save the value of the guess
      xgx = xv(igx)
      call DataDelete(OC)
      call DataDelete(OV)
   ! ------------------------------------
   contains
   ! ------------------------------------
   function SolveColloc(yn) result(y)
      real(float) :: y(n1),yn(n1)
      Real(float) :: A(n1,n1),b(n1),Bd(n),An1n1
      integer :: i,kiter,itx
      ! -----------------------------
      y = yn
      itx = 1  ! guess counter
      exff = 0.0_float
      select case(method)
      case(0)        ! conventional collocation
         ns = n
         if(Bi > 0.0_float)then
            An1n1 = 1.0_float/(Ac(n1,n1)+Bi)
            do i=1,n    !* eliminate y(n+1) from equation (boundary collocation)
               A(i,:) = Bc(i,:) + An1n1*Bc(i,n1)*Ac(n1,:)
            enddo
         else
            A = Bc
         endif
         do i=1,n
            Bd(i) = A(i,i) ! save diagonal
         enddo
      case default   ! weak, symmetric
         A = Cc
      end select
      do kiter=1,maxiter
         r = RateRxn(y,dr)
         select case(method)
         case(0)        ! conventional formulation
            b = MatMul(Bc,y)
            do i=1,n
               A(i,i) = Bd(i) + dr(i)
               b(i) = -r(i) - b(i)
            enddo
         case default   ! weak, symmetric collocation
            b = MatMul(Cc,y)
            do i=1,n1
               A(i,i) = Cc(i,i) - wc(i)*dr(i)
               b(i) = wc(i)*r(i) - b(i)
            enddo
            exff(2) = b(n1)  ! temporary for testing
            if(Bi > 0.0_float)then
               A(n1,n1) = A(n1,n1) + Bi
               b(n1) = b(n1) - Bi*y(n1)
            endif
         end select
         exff(1) = sum(wc*r)
         yn(:ns) = LUSolveC(A(:ns,:ns),b(:ns))   ! symmetric solve not exploited
         y(:ns) = y(:ns) + yn(:ns)
         dymax = maxval(abs(yn(:ns)))
         if(itx < 3 .and. (kiter > 10*itx .and. &  ! failed, try alternate guess
               (dymax > 0.01 .or. y(1) > 1.1 .or. y(1) < 0.01)))then
            y = yguess(ygx(itx));  itx = itx + 1
            write(*,'(a,12f10.7)')'New Guess:',y
         endif
         Call Debugx(kiter,y,yn,A,b,A,b)
         if(dymax < eps)exit
      enddo
   End function SolveColloc
   ! ------------------------------------
   function SolveGalerkin(yn)  result(y)
      real(float) :: y(n1),yn(n1)
      Type(ColDataType) :: OG
      Real(float) :: A(n1,n1),b(n1)
      Real(float) :: Dc(n1,n1),hc(n1)   ! mass matrix, load vector
      real(float) :: xq(n1+nxtra),wq(n1+nxtra),Lq(n1+nxtra,n1)
      real(float) :: yq(n1+nxtra),rq(n1+nxtra)
      integer :: kiter,itx
      ! -----------------------------
      Call Initialize(OG,n+nxtra,typ,igeom)
      xq = Xpoints(OG)
      wq = WeightQ(OG)
      Lq = Lcoeff(xq,OC)
      y = yn
      itx = 1  ! guess counter
      do kiter=1,maxiter ! iteration loop ------------
         yq = MatMul(Lq,y)
         rq = RateRxn(yq,drq)
         Dc = MassMatrix(OC,OG,RxnFunc)
         hc = MatMul(wq*rq,Lq)
         b = hc - MatMul(Cc,y)
         A = Cc + Dc
         exff(1) = sum(wq*rq)
         exff(2) = b(n1)  ! temporary for testing
         if(Bi > 0.0_float)then
            A(n1,n1) = A(n1,n1) + Bi
            b(n1) = b(n1) - Bi*y(n1)
         endif
         yn(:ns) = LUSolveC(A(:ns,:ns),b(:ns))   ! symmetric solve not exploited
         y(:ns) = yn(:ns) + y(:ns)
         dymax = maxval(abs(yn(:ns)))
         if(itx < 3 .and. (kiter > 10*itx .and. &  ! failed, try alternate guess
               (dymax > 0.01 .or. y(1) > 1.1 .or. y(1) < 0.01)))then
            y = yguess(ygx(itx));  itx = itx + 1
            write(*,'(a,12f10.7)')'New Guess:',y
         endif
         Call Debugx(kiter,y,yn,A,b,Dc,hc)
         if(dymax < eps)exit
      enddo ! end of iteration loop ------------------
      call DataDelete(OG)
   End Function SolveGalerkin
   ! ------------------------------------
      Function yguess(yg) result(y)
         real(float) :: yg,y(size(xc)),fac
         fac = 1.0_float/(1.0_float - xgx*xgx)
         y = yg*(1.0_float - xc**2)*fac
         y = min(y,0.999_float)
         if(Debug > 0)then
            write(ilog,'(a,2g14.7)')'Guess x, y =',xgx,yg
            call vectorprint('Guess',y,flog=.true.)
         endif
      End Function yguess
      ! ------------------------------------
      Function RxnFunc(x) result(r)
         include 'defs.fi'
         Real(float), intent(in) :: x(:)
         Real(float) :: r(size(x))
         !call arrayprint('RxnFunc',x,rq,drq)
         r = -drq
      end function RxnFunc
      ! ------------------------------------
      Subroutine Debugx(kiter,y,yn,A,b,Dc,hc)
         integer :: kiter
         real(float) :: y(:),yn(:),A(:,:),b(:),Dc(:,:),hc(:)
         if(kiter == 1)then
            if(Debug > 0)write(ilog,'(2a,i2,a,2f8.4,a)')trim(TxtType(typ)),' n =',n, &
               ' Thiele =',Sqrt(Thparm)/Thfac,Sqrt(Thparm),' (general, actual)'
            write(*,'(2a,i2,a,2f8.4,a)')trim(TxtType(typ)),' n =',n,' Thiele =', &
               Sqrt(Thparm)/Thfac,Sqrt(Thparm),' (general, actual)'
         endif
         if(Debug > 0)then
            write(ilog,'(a,i3,E11.2,2f12.8)')'Iter, Max(Delta y):',kiter,dymax,y(1)-yn(1),y(1)
            write(*,'(a,i3,E11.2,2f12.8)')'Iter, Max(Delta y):',kiter,dymax,y(1)-yn(1)
         endif
         if(Debug > 3)then
            Call ArrayPrint('Mass Matrix & Load Vector:',Dc,hc,flog=.true.)
            call ArrayPrint('Solution Matrix, rhs, solution',A,b,y,flog=.true.)
         endif
      End Subroutine Debugx
! -----------------------------------------------------------------------
   End Subroutine SolveBVP
! -----------------------------------------------------------------------
   Function RateRxn(y,dr)       result(r)
      real(float) :: y(:),dr(:),r(size(y)),d
      real(float), save :: rord, drext
      Logical, save :: First = .True.
      integer i
      if(First)then
         First = .False.  ! calculation performed only once
         rord = Real(Korder,float)
         drext = 0.0_float
         if(Korder == 1)drext = -1.0_float/(1.0_float - Kd)**2
      endif
      do i=1,size(y)
         if(y(i) < 1.0_float)then
            d = 1.0_float/(1.0_float - Kd*y(i))
            r(i) = (1.0_float - y(i))**Korder
            dr(i)= -rord*(1.0_float - y(i))**(Korder-1) + 2.0_float*r(i)*Kd*d
            r(i) = r(i)*(Thparm*d**2)
            dr(i) = dr(i)*(Thparm*d**2)
         else
            dr(i) = drext*Thparm
            r(i) = dr(i)*(y(i) - 1.0_float)
         endif
      enddo
      if(Debug > 1)then
         write(ilog,'(" y:",20F12.5)')y
         write(ilog,'(" r:",20F12.5)')r
         write(ilog,'("dr:",20F12.5)')dr
      endif
   End Function RateRxn
! -----------------------------------------------------------------------
   Function FluxCalc(y,r,wc,Ac,Cc,Bi)      result(eff)
! -----------------------------------------------------------------------
      real(float) :: y(:),r(:),wc(:),Ac(:,:),Cc(:,:),Bi
      real(float) :: eff(4)
      integer :: n1
      n1 = size(y)
      eff(1) = Sum(wc*r)         ! integrate
      eff(2) = wc(n1)*r(n1) - Sum(Cc(n1,:)*y)   ! using weak form
      eff(3) = Bi*y(n1)          ! use boundary value
      eff(4) = -Sum(Ac(n1,:)*y)  ! derivative
      eff = eff*gfac/Thparm
      exff = exff*gfac/Thparm
   End Function FluxCalc
! -----------------------------------------------------------------------
   Function ResidCalc(xc,Bc,y) result(Rerr)
! -----------------------------------------------------------------------
      real(float) :: xc(:),y(:),Bc(:,:),Rerr(2)
      real(float) :: d2c(size(xc)),Lc(Nv,size(xc))
      ! This calculation only valid for constant diffusion
      Lc = Lcoeff(xv,OC)
      yv = MatMul(Lc,y)          ! y at xv
      rv = RateRxn(yv,Resid)     ! rates at xv
      d2c = MatMul(Bc,y)         ! 2nd derivative at xc
      Resid = MatMul(Lc,d2c)     ! 2nd derivative at xv
      Resid = Resid + rv         ! Residual at xv
      apoly(:,1) = Jacobi_Transform(Nv-2,typ,igeom,func=ResFunc,Shift=.true.)
      apoly(:,2) = 1.0_float/Jacobi_Scale_Sym(Nv-2,abeta)
      where(abs(apoly(:,1))<1.0d-13)apoly(:,1)=0.0_float
      select case(typ)
      case(Lobatto)
         wrv = wv*(1.0_float - xv*xv)
      case(Chebyshev)
         wrv = wv*sqrt(1.0_float - xv*xv)
      case default
         wrv = wv
      end select
      Rerr(1) = Rnorm(wv*Resid**2,wv)     ! L2 norm
      Rerr(2) = Rnorm(wrv*Resid**2,wrv)   ! weighted norm
      contains
         Function Rnorm(w,w0) result(wr)
            real(float) :: w(:),w0(:),wr
            wr = sqrt(sum(w)/sum(w0))
         End Function Rnorm
   End Function ResidCalc
! -----------------------------------------------------------------------
   function ResFunc(x)  result(rx)
      real(float), intent(in) :: x(:)
      real(float) :: rx(size(x))
      rx = Resid(1:size(x))
   end function ResFunc
! -----------------------------------------------------------------------
   Subroutine Report(typ,igeom,Bi,y,xc,r,eff,Rerr)
! -----------------------------------------------------------------------
      Integer :: typ,igeom
      Real(float) :: Bi
      Real(float) :: y(:),xc(:),r(:),eff(:),Rerr(:)
      integer :: n,n1
      Character (len=*), Parameter :: TxtMethod(0:2) = (/'conventional collocation',&
              'weak collocation        ','Galerkin                '/)
      Character (len=*), Parameter :: fmtg = '(10(es9.2,a1))'
      Character (len=*), Parameter :: fmtf = '(10(f20.17,a1))'
      ! -------------------------------
      n1 = size(xc)
      n = n1 - 1
      write(iout,'(a,i3,a,L2)')'Results for N =',n    !,' using boundary collocation? ',BColloc
      write(iout,'(8a)')'Using ',trim(TxtType(typ)),' points solved by ', &
         trim(TxtMethod(method)),' method'
      if(method==2)write(iout,'(a,i2)')'Number extra quadrature points used =',nxtra
      write(iout,'(2a)'  )'The Geometry is ',trim(TxtGeom(igeom))
      write(iout,'(a,f8.2)')'Flux Constant  :',Bi
      write(iout,'(a,I2,a,f6.3)')'Reaction Order =',Korder,', with Ka =',Kd
      write(iout,'(a,2f9.5)')'Thiele Modulus(general,actual) :',Sqrt(Thparm)/Thfac,Sqrt(Thparm)
      Call ArrayPrint('Collocation Points, Solution, rates:',xc,y,r,fmtx='(2(f15.12,a1),2(g15.7,a1))')
      call arrayprint('Interpolated x W y rate residual:',reshape((/xv,wv,yv/),(/Nv,3/)), &
            rv,Resid,fmtx='(3(f15.12,a1),2(g15.7,a1))')
      Call VectorPrint('Effectiveness Factor - Integrated, Corrected Derivative, Difference, Derivative:',eff,fmtx=fmtf)
      write(iout,'(2(f20.17,a1),a)')exff(1),tb,exff(2),tb,'Alternate calculation'
      write(iout,'(a,2(a1,es11.4))')'Residual norms: L2 & Lw =',tb,Rerr(1),tb,Rerr(2)
      call arrayprint('Transform coefficients & scaling factors:',transpose(apoly))
   End Subroutine Report
End Module SymBVP
! -----------------------------------------------------------------------
Program Example02G
! -----------------------------------------------------------------------
   Use SymBVP, Only : SolveBVP, BVP_Data
!   Use OrthogonalColloc, Only : CollocSet, Gauss, Lobatto, Planar, Cylindrical, Spherical
   include 'defs.fi'
   Integer, Parameter :: MaxNmax = 100, MaxTmax = 100
   Integer :: Nc(3),nnc,n,n1
   Integer :: nt,nthp
   integer :: i
   Real(float) :: Thp(3)
   Real(float) :: y(MaxNmax)           ! solution
   Real(float) :: effx(4),Rerrx(2),errx
   Type(BVP_Data) :: d
   Real(float), allocatable :: eff(:,:),Thpx(:),Rerr(:,:)
   ! ---------------------------------------------------
   ! Main for Symmetric Reaction-Diffusion problem
   ! Note: can run using Ex02 < Ex02.dat if you put
   ! data in file Ex02.dat, otherwise will be interactive
   ! Code is set up to loop over either:
   !   N - collocation points to check accuracy
   !   Thiele No. - either forward or backward to track a solution
   ! ---------------------------------------------------
   call GetData(d,Nc,Thp)
   nthp = Int((Thp(2) - Thp(1) + Thp(3)*0.01_float)/Thp(3)) ! number of Thiele parameters
   nnc = (Nc(2) - Nc(1))/Nc(3) + 1                     ! no. differenet n
   allocate(eff(nnc+1,0:nthp))
   allocate(Thpx(0:nthp))
   allocate(Rerr(nnc,0:nthp))
   ! ------------------
   Thpx = (/((Thp(1) + Thp(3)*i),i=0,nthp)/) ! thiele parameters solved
   do nt = 0,nthp
      d%Thparm = Thpx(nt)
      i = 0
      do n=Nc(1),Nc(2),Nc(3)
         i = i + 1
         n1 = n + 1
         Call SolveBVP(n1,d,y(:n1),effx,Rerrx)
         eff(i,nt)=effx(1)
         Rerr(i,nt) = Rerrx(1)
      enddo
      eff(nnc+1,nt) = effx(1)
   enddo
   write(iout,'(a)')'Summary: Thparm  n   eff. factor   est. error   L2 error'
   do nt = 0,nthp
      i = 0
      do n=Nc(1),Nc(2),Nc(3)
         i = i + 1
         errx = abs(1.0_float - eff(i,nt)/eff(nnc+1,nt))
         write(iout,'(f8.3,a1,i2,a1,F18.14,2(a1,G12.5))') &
            Thpx(nt),tb,n,tb,eff(i,nt),tb,errx,tb,Rerr(i,nt)
      enddo
   enddo
End Program Example02G
Subroutine GetData(d,Nc,Thp)
   Use SymBVP, Only : BVP_Data
   Use OrthogonalColloc, Only : Gauss, Lobatto, Chebyshev, Planar, Cylindrical, Spherical
   Use Array_Print, Only : OpenFile
   include 'defs.fi'
   Type(BVP_Data) :: d
   integer :: Nc(3)
   Real(float) :: Thp(3)
   Character :: fname*80,ch*1
   Integer, Parameter :: MaxNmax = 100
   ! ---------------------------
   ! typ    - 1 Gauss, 2 Lobatto, 3 Chebyshev
   ! geom   - geometric factor 0,1,2
   ! Korder - reaction order (numerator)
   ! Kd   - reaction, adsorption term
   ! Thparm - Thiele modulus
   ! Bi  - constant in flux b.c.
   ! yg     - initial guess
   ! reaction rate = Thparm**2*(1 - y)**Korder/(1 - Kd*y)**2
   ! ---------------------------
   ! default values
   Nc = 2
   Thp = 1.0_float
   d%typ = Lobatto
   d%geom = Planar
   d%Bi = 0.0_float
   d%Thparm = 1.0_float
   d%Korder = 1
   d%Kads = 0.0_float
   ! ---------------------------
   write(*,'(a,i2,a,i2,a)')' Enter g/l/c for Gauss/Lobatto/Chebyshev >'
   read(*,*)ch;  d%typ = index('glc',ch);  d%typ = Max(Gauss,Min(Chebyshev,d%typ))
   write(*,'(a)')' Enter n first, last and delta >'
   read(*,*)Nc
   Nc(1) = max(Nc(1),1)
   Nc(2) = min(max(Nc(1),Nc(2)),MaxNmax)
   write(*,'(a)')'Enter 1, 2 or 3 for Planar, Cylindrical or Spherical Geometry > '
   read(*,*)d%geom
   write(*,'(a)')' Enter Boundary Flux Constant (use 0 or / for specified value) > '
   read(*,*)d%Bi
   write(*,'(a)')' Enter reaction order, adsorption term > '
   read(*,*)d%Korder,d%Kads
   d%Kads = Max(0.0_float,Min(d%Kads,0.9999999_float))  ! 0 <= Kads < 1
   write(*,'(a)')' Thiele parameter first, last, delta >'
   read(*,*)Thp
   write(*,*)' Enter initial guess, recommend 0.3 lower, 0.8 upper'
   read(*,*)d%yg
   write(*,'(A)')' Enter root of output file name>'
   read(*,*)fname
   Call OpenFile(fname)
   Call OpenFile(fname,'log')
End Subroutine GetData
