! =============================================================================
Module RateEqn
! =============================================================================
! This small module does rate calculations.
   Private
   Public RateParms  ! stores the reaction parameters
   Public RxnRate    ! calculates rates and derivatives
   include 'defs.fi'
   Real(float) :: Eact     ! activation energy
   Real(float) :: Rorder   ! reaction rate order
   Integer :: Korder       ! only 1st order implemented
   integer, parameter, public :: Debug = 0
   ! ---------------------------------------------------
   Contains
   ! ---------------------------------------------------
   Subroutine RateParms(Korderx,Eactx)
   ! ---------------------------------------------------
      Integer :: Korderx
      Real(float) :: Eactx
      Eact = Eactx
      Korder = Korderx
      Rorder = Real(Korder,float)
   End Subroutine RateParms
   ! ---------------------------------------------------
   Function RxnRate(y,dr) result(r)
   ! ---------------------------------------------------
      Real(float) :: y(:,:),dr(:,:),r(size(y,1))
      Real(float) :: Tinv(size(y,1))
      Tinv = 1.0_float/(y(:,1) + 1.0_float) ! only 1st order implemented
      dr(:,2) = -exp(Eact*(1.0_float - Tinv))
      r = -dr(:,2)*(1.0_float - y(:,2))
      dr(:,1) = Eact*Tinv*Tinv*r
      if(Debug > 3)then
         write(ilog,'(" r:",10F12.5)')r
         write(ilog,'("dr:",10F12.5)')dr(:,1)
         write(ilog,'("dr:",10F12.5)')dr(:,2)
      endif
   End Function RxnRate
End Module RateEqn
! =============================================================================
Module AxialDisp
! =============================================================================
   !  Nonsymmetric BVP for nonisothermal chemical reactor with axial dispersion and
   !  first order reaction. Solved using Gauss,Lobatto,Radau or Chebyshev points
   !     (1/Pe)*d2y/dz2 - dy/dz = r(y,t)
   !     (1/Pe)*d2t/dz2 - dt/dz - Bi*t = r(y,t)
   !  with:
   !     1/Pe*dy/dz - y = 0 at z = 0
   !     dy/dz = 0 at z = 1
   !  bc's on t or modification of above.
   !  Most of this code is written so that any number (Neq) balances can
   !  be solved. It would also be easy to modify for a variable number of
   !  rate equations. Currently, Neq = 2, i.e. y(1) = t and y(2) = y.
   !  The rate equation module can be replaced with any rate equation,
   !  currently:  r(y) = (1-y(2))*exp[e*(1-1/(y(1)+1))]
   ! ---------------------------------------------------
   Use RateEqn, Only : RxnRate, Debug
   Use OrthogonalColloc, Only : ColDataType,Initialize,CollocCoef,Lcoeff,TxtType
   Use LUSolvers, Only : LUSolveC, LUFactrC, LUSubstC ! "C" Crout (no pivot)
   Use Array_Print, Only : ArrayPrint
   Private
   Public SetParms
   Public Solve
   include 'defs.fi'
   Integer :: typ          ! 1 Gauss, 2 Lobatto, etc.
   Integer :: n,i          ! no. interior coll. pts.
   Integer :: Neq          ! set up for any number coupled equations
   Integer, Parameter :: MethodBC = 1  ! 0 - boundary coll., otherwise natural bc
   Integer :: MethSolve ! Solution method: 0 - NR reduced, 1 - NR full, 2 - Picard
   Real(float), Allocatable :: xc(:)
   Real(float), Allocatable :: wc(:)
   Real(float), Allocatable :: Ac(:,:)
   Real(float), Allocatable :: Aw(:,:)
   Real(float), Allocatable :: Bc(:,:)
   Real(float), Allocatable :: Cc(:,:)
   Integer, Parameter :: MaxEq = 2
   Real(float) :: Pe(MaxEq),Bi(Maxeq),Rx(MaxEq),Bc0(Maxeq),Bc1(Maxeq) ! data
   Real(float) :: Rbc(0:1,MaxEq)   ! boundary condition residuals
   integer, parameter :: maxiter =  100 ! 200
   Character (len=*), parameter :: fmtd = '(30(F12.8,a1))'
   Character (len=*), parameter :: fmtg = '(30(G12.5,a1))'
   Character :: cn(MaxEq) = (/(char(ichar('0')+i),i=1,MaxEq)/)
!-------------------------------------------------------------------------------
   Contains
! -----------------------------------------------------------------------
   Function Solve(OC,n1,typx,meth,xv,yv,rv)  result(y)
! -----------------------------------------------------------------------
      Type(ColDataType) :: OC
      integer :: n1,typx,meth
      Real(float) :: y(0:n1,Neq)
      Real(float) :: xv(:),yv(:,:),rv(:,:)
      Real(float) :: G(0:n1,0:n1,Neq),HP(0:n1,0:n1,Neq)
      Real(float) :: dy(0:n1,Neq),r(0:n1),dr(0:n1,Neq)
      Real(float) :: dymax(Neq), dymx, eps
      integer :: n2,j,kiter
      ! ----------------------------------------------------------------
      ! The coefficients are as follows:
      !  G(j,i,k) - matrix minus reaction terms for balance equation k, i.e. (T,y,etc)
      ! ----------------------------------------------------------------
      n = n1 - 1;   n2 = n + 2;  typ = typx;  MethSolve = meth
      y = 0.0_float;  eps = 1.0e-6_float;  ! Epsilon(1.0_float)*1.0e5_float
      Call GetCollocData(OC,n1) ! set up collocation data
      G = CalcCoef()         ! calculate constant coefficients
      do kiter=1,maxiter
         r = RxnRate(y,dr)
         select case(MethSolve)
         case(0)                             ! NR reduced to (n+2)x(n+2)
            dy = CalcReduce(G,y,r,dr,kiter)  ! implemented for Neq=2 only
         case(1)
            dy = CalcFull(G,y,r,dr)          ! NR [Neq(n+2)]x[Neq(n+2)]
         case(2)
            dy = CalcPicard(G,HP,y,r,kiter)  ! Picard Iteration
         end select
         do j=1,Neq
            dymax(j) = Maxval(abs(dy(:,j)))
            y(:,j) = y(:,j)+dy(:,j)
         enddo
         dymx = Maxval(dymax)
         Call DebugY(y)
         if(dymx < eps)exit
      enddo
      r = RxnRate(y,dr)
      Call ResultCalc(y,xv,yv,rv)
      Call Report(xc,y,r,xv,yv,rv)
      Call DeleteCollocData()
      ! ------------------------------------
      contains
      ! ------------------------------------
      Subroutine DebugY(y)
         Real(float) :: y(:,:)
         if(Debug > 0) &
            Call ArrayPrint('Max(Delta y), y:',dymax,Transpose(y),fmtx=fmtd,flog=.true.)
         if(Debug > 0)write(ilog,'(a,i3,e12.5)')'dymax:',kiter,dymx
         if(Debug > -1)write(*,'(a,i3,e12.5)')'dymax:',kiter,dymx
      End Subroutine DebugY
   End Function Solve
! -----------------------------------------------------------------------
   Function CalcCoef()  result(G)
! -----------------------------------------------------------------------
      Integer :: n1,i,j
      Real(float) :: G(0:n+1,0:n+1,Neq)   !),Gd(0:,:)
      n1 = n + 1
      ! Only the diagonal changes between iterations
      ! Calculate coefficients that do not change from iteration to iteration
      ! This layout makes it easy to compare the two representations of the method
      !----------------------------------
      do j=1,Neq
         G(:,:,j) = Pe(j)*Cc + Aw
         G(0,0,j) = G(0,0,j) + Bc0(j)
         G(n1,n1,j) = G(n1,n1,j) + Bc1(j)
         do i=0,n1
            G(i,i,j) = G(i,i,j) + wc(i)*Bi(j)
         enddo
         if(Debug > 4)Call ArrayPrint(  &
            'Constant Arrays, Eq = '//cn(j),G(:,:,j),fmtx=fmtd,flog=.true.)
      enddo
   End Function CalcCoef
! -----------------------------------------------------------------------
   Function CalcReduce(G,y,r,dr,iter)  result(dy)
! -----------------------------------------------------------------------
      Real(float) :: G(:,:,:),y(:,:),r(:),dr(:,:)
      Real(float) :: dy(size(y,1),size(y,2))
      Integer :: iter
      Real(float) :: H(size(y,1),size(y,1)),b(size(y,1))
      Integer :: n2,Neq,j
      ! ------------------------------------------------------------
      ! Reduce problem by elimination first unknown (temperature)
      ! This code assumes Neq = 2, but it could be generalized
      ! ------------------------------------------------------------
      n2 = size(y,1); Neq = size(y,2)
      if(iter == 1)then
         write(*,'(a,2f8.4)')'Reduced Newton-Raphson solution:'
         write(iout,'(a,2f8.4)')'#Reduced Newton-Raphson solution:'
         Call LUFactrC(G(:,:,1))
         G(:,:,1) = LUSubstC(G(:,:,1),G(:,:,2))
      endif
      b = Rx(2)*r*wc - MatMul(G(:,:,2),y(:,2))
      do j=1,n2
         H(:,j) = G(:,j,2) - Rx(1)*wc*dr(:,1)*G(:,j,1)
         H(j,j) = H(j,j) - Rx(2)*wc(j-1)*dr(j,2)
      enddo
      dy(:,2) = LUSolveC(H,b)
      dy(:,1) = (Rx(1)/Rx(2))*MatMul(G(:,:,1),dy(:,2));
      if(Debug > 4)then
         Call ArrayPrint('Reduced NR',H,b,fmtx=fmtd,flog=.true.)
         Call ArrayPrint('Solution dy  =',Transpose(dy),fmtx=fmtg,flog=.true.)
      endif
   End Function CalcReduce
! -----------------------------------------------------------------------
   Function CalcFull(G,y,r,dr)  result(dy)
! -----------------------------------------------------------------------
      Real(float) :: G(:,:,:),y(0:,:),r(0:),dr(0:,:)
      Real(float) :: dy(size(y,1),size(y,2))
      Real(float) :: GG(size(y,1)*size(y,2),size(y,1)*size(y,2)+1)
      Integer :: n2,Neq,ir,i,j,k,i1,i2,j1,j2
      ! ------------------------------------------------------------
      ! Assemble coefficients for Newton-Raphson block matrix
      ! G - Neq constant n2xn2 matrices
      ! GG - n2*Neq order matrix, n2xn2 blocks
      ! GG(:,ir) - rhs residual term
      ! ------------------------------------------------------------
      n2 = size(y,1);  Neq = size(y,2); ir=n2*Neq+1
      do i=1,Neq  ! load coefficients into full matrix
         i2 = i*n2;  i1 = i2-n2+1
         GG(i1:i2,ir) = Rx(i)*wc*r - MatMul(G(:,:,i),y(:,i))   ! residual
         do j=1,Neq
            j2 = j*n2;  j1 = j2-n2+1;
            if(i == j)then ! rate terms diagonals of diagonal blocks
               GG(i1:i2,j1:j2) = G(:,:,i)
               do k = 0,n2-1
                  GG(k+i1,k+i1) = GG(k+i1,k+i1) - Rx(i)*dr(k,j)*wc(k)
               enddo
            else           ! rate terms diagonals of off diagonal blocks
               GG(i1:i2,j1:j2) = 0.0_float
               do k = 0,n2-1
                  GG(k+i1,k+j1) = -Rx(i)*dr(k,j)*wc(k)
               enddo
            endif
         enddo
      enddo
      dy = Reshape(LUSolveC(GG,GG(:,ir)),shape(dy))
      if(Debug > 4)then
         Call ArrayPrint('Full NR Arrays',GG,fmtx=fmtd,flog=.true.)
         Call ArrayPrint('Solution dy  =',Transpose(dy),fmtx=fmtg,flog=.true.)
      endif
   End Function CalcFull
! -----------------------------------------------------------------------
   Function CalcPicard(G,H,y,r,iter)  result(dy)
! -----------------------------------------------------------------------
      Real(float) :: G(:,:,:),H(:,:,:),y(:,:),r(:)
      Integer :: iter,n2,Neq,j
      Real(float) :: dy(size(y,1),size(y,2))
      Real(float) :: b(size(y,1),size(y,2))
      Real(float), parameter :: Piter(2) = (/1.0_float,1.0_float/) ! iteration parameter
      ! ------------------------------------------------------------
      ! Solve with Picard iteration, fiddled with iteration parameter
      ! cannot get much improvement over values of 0. Convergence rate
      ! seems to be insensitive to values.
      ! ------------------------------------------------------------
      n2 = size(y,1); Neq = size(y,2)
      if(iter == 1)then
         write(*,'(a,2f8.4)')'Picard Iteration Parm =',Piter
         write(iout,'(a,2f8.4)')'Picard Iteration Parm =',Piter
         do j=1,Neq
            H(:,:,j) = G(:,:,j)
            do i=1,n2
               H(i,i,j) = H(i,i,j) + Rx(j)*wc(i-1)*Piter(min(j,2)) ! scaled iteration parms.
            enddo
            call LUFactrC(H(:,:,j))  ! factor Croute algorithm (no pivot)
         enddo
      endif
      do j=Neq,1,-1
         b(:,j) = Rx(j)*r*wc - MatMul(G(:,:,j),y(:,j))   ! residuals
         dy(:,j) = LUSubstC(H(:,:,j),b(:,j))             ! solve with factors
      enddo
      if(Debug > 1)then
         Call ArrayPrint('Residuals   =',Transpose(b),fmtx=fmtg,flog=.true.)
         Call ArrayPrint('Solution dy =',Transpose(dy),fmtx=fmtg,flog=.true.)
      endif
   End Function CalcPicard
! -----------------------------------------------------------------------
   Subroutine GetCollocData(OC,n1)
      Type(ColDataType) :: OC
      integer :: n1,i
      Allocate(xc(0:n1))
      Allocate(wc(0:n1))
      Allocate(Ac(0:n1,0:n1))
      Allocate(Aw(0:n1,0:n1))
      Allocate(Bc(0:n1,0:n1))
      Allocate(Cc(0:n1,0:n1))
      Call CollocCoef(OC,xc,wc,Ac,Bc,Cc)
      if(MethodBC == 0)then
      ! set up for boundary coll. (modify boundary appx.)
         wc(0) = 0.0_float;  wc(n1) = 0.0_float
         Cc(0,:) = -Ac(0,:); Cc(n1,:) = Ac(n1,:)
      endif
      do i=0,n1
         Aw(i,:) = wc(i)*Ac(i,:)
      enddo
   End Subroutine GetCollocData
! -----------------------------------------------------------------------
   Subroutine DeleteCollocData()
      DeAllocate(xc,wc,Ac,Aw,Bc,Cc)
   End Subroutine DeleteCollocData
! -----------------------------------------------------------------------
   Subroutine ResultCalc(y,xv,yv,rv)
! -----------------------------------------------------------------------
      Use RateEqn, Only : RxnRate
      real(float) :: y(0:,:),xv(0:),yv(0:,:),rv(0:,:)
      real(float) :: drv(size(yv,1),size(yv,2)),Li(size(xv),size(xc))
      real(float) :: Rs1(size(y,1)),Rs(0:1,Neq)
      integer :: j
      Li = Lcoeff(xv,xc)  ! interpolating polynomials at xv
      do j = 1,Neq
         yv(:,j) = Matmul(Li,y(:,j))
      enddo
      rv(:,1) = RxnRate(yv,drv)
      do j = 1,Neq
         Rs1(:) = MatMul(Pe(j)*Bc-Ac,y(:,j)) ! 1st & 2nd derivative terms
         rv(:,j+1) = MatMul(Li,Rs1) - Bi(j)*yv(:,j) + Rx(j)*rv(:,1)  ! residual
         Rbc(0,j) = Bc0(j)*y(0,j) - Pe(j)*sum(Ac(0,:)*y(:,j))        ! boundary resid.
         Rbc(1,j) = Bc1(j)*y(n+1,j) + Pe(j)*sum(Ac(n+1,:)*y(:,j))    ! boundary resid.
         Rs(0,j) = rv(0,j+1)*wc(0)              ! interior Resid at 0
         Rs(1,j) = rv(size(xv)-1,j+1)*wc(n+1)   ! interior Resid at 1
!         Call ArrayPrint('Residual Calc:',xc,Rs1,Bi(j)*y(:,j),Rx(j)*rv(1::4,1))
!         Call ArrayPrint('Residual Calc:',xv,MatMul(Li,Rs1),Bi(j)*yv(:,j),Rx(j)*rv(:,1))
      enddo
!      Call ArrayPrint('Boundary Residual:',Rbc,Rs)
   End Subroutine ResultCalc
! -----------------------------------------------------------------------
   Subroutine Report(xc,y,r,xv,yv,rv)
! -----------------------------------------------------------------------
      Real(float) :: xc(0:),y(0:,:),r(0:)
      real(float) :: xv(0:),yv(0:,:),rv(0:,:)
      integer :: j,ni
      Character :: cn*7
!     Character (len=*), parameter :: fmtt = '(20(D22.15,A1))'
      Character (len=*), parameter :: fmtt = '(20(F11.8,A1))'
      ! -------------------------------
      ni = size(xv) - 1; cn = ' n =xx"'; write(cn(5:6),'(I2)')n;
      write(iout,'(a,i2,a,i2)')'#Results for N =',N,' Method',MethodBC
      write(iout,'(3a)'  )'#Solution with ',Trim(TxtType(typ)),' Points'
      write(iout,'(a)')'#Coefficient - Energy/Mass:'
      write(iout,'(a,10f8.3)')'#Reaction Coeff.:',Rx(:Neq)
      write(iout,'(a,10f8.2)')'#Peclet Numbers :',1.0_float/Pe(:Neq)
      write(iout,'(a,10f8.3)')'#Biot Number    :',Bi(:Neq)
      write(iout,'(a,10f8.3)')'#Inlet BC Coeff.:',Bc0(:Neq)
      write(iout,'(a,10f8.3)')'#Oulet BC Coeff.:',Bc1(:Neq)
      write(iout,'(a,i3,10(a1,g12.5))')'#Inlet BC Resid.:',n, &
         (tb,abs(Rbc(0,j)),tb,abs(rv(0,j+1)),j=1,Neq)
      write(iout,'(a,i3,10(a1,g12.5))')'#Oulet BC Resid.:',n, &
         (tb,abs(Rbc(1,j)),tb,abs(rv(ni,j+1)),j=1,Neq)
      Call ArrayPrint('Collocation Points Solution, rates:',xc,y,r,fmtx=fmtt)
      Call ArrayPrint('Zone T="'//Trim(TxtType(typ))//cn,xv,yv,rv,fmtx=fmtt,flog=.false.)
   End Subroutine Report
!-------------------------------------------------------------------------------
   Subroutine SetParms(Pem,Pet,Bix,Rxm,Rxt)
! -----------------------------------------------------------------------
      Real(float) :: Pem,Pet,Bix,Rxm,Rxt
      Neq = 2
      Pe(1) = 1.0_float/Pet;   Pe(2) = 1.0_float/Pem
      Bi(1) = Bix;             Bi(2) = 0.0_float
      Rx(1) = Rxt;             Rx(2) = Rxm
      Bc0(1:Neq) = 1.0_float + Bi*Pe
      Bc1(1:Neq) = Bi*Pe
   End Subroutine SetParms
End Module AxialDisp
! =============================================================================
Program AxialDispRxn
! -----------------------------------------------------------------------
   Use RateEqn, Only : RateParms, Debug
   Use AxialDisp, Only : SetParms, Solve
   Use OrthogonalColloc, Only : ColDataType,Initialize,DataDelete,Xpoints,WeightQ,Interp, &
      TxtType,Gauss,Lobatto,RadauL
   include 'defs.fi'
   Integer :: Nmin, Nmax, Ndel
   Integer, Parameter :: MaxNmax = 101
   Integer :: n,n1,j
   Integer :: typ,MethSolve    ! point type, method of solution
   Integer, Parameter :: Neq = 2, Nsplit = 4
   Real(float) :: y(0:MaxNmax+1,Neq,0:MaxNmax)  ! solutions
   Real(float) :: ymid(2,Neq,MaxNmax)             ! midpoint and endpoint values
   Integer :: Nv
   Real(float), Allocatable :: xv(:),yv(:,:),rv(:,:)
   Type(ColDataType) :: OC,OX
   Integer, parameter :: Nx = 99
   Real(float) :: xx(0:Nx+1),wx(0:Nx+1),yx(0:Nx+1,Neq),erry(Neq,MaxNmax)
   ! ---------------------------------------------------
   ! Main for Reaction-Diffusion problem
   ! Note: can run using ex03ax < x3in.dat if you put
   ! data in file x3in.dat, otherwise will be interactive
   !  xx,yx - "exact" solution read from file
   !  xv,yv - is fine grid with solution values and points between
   ! ---------------------------------------------------
   Call Setup
   Call Initialize(OX,Nx,Lobatto)
   xx = Xpoints(OX)
   wx = WeightQ(OX)
   call DataDelete(OX)
   do n = Nmin,Nmax,Ndel
      write(*,'(a,i3,1x,2a)')'Solution for n =',N,Trim(TxtType(typ)),' Points'
      n1 = n + 1;  Nv = n1*Nsplit
      Allocate(xv(0:Nv),yv(0:Nv,Neq),rv(0:Nv,Neq+1))
      Call Initialize(OC,n,typ)
      xv = Fill(n)  ! xv - collocation pts & Nsplit - 1 in between
      y(:n1,:,n) = Solve(OC,n1,typ,MethSolve,xv,yv,rv)
      erry(:,n) = ErrorCalc(xv,yv,ymid(:,:,n))
      Deallocate(xv,yv,rv)
   enddo
   write(iout,'(a)')'Errors: L1-1, L1-2, y1mid y1end y2mid y2end'
   do n=Nmin,Nmax,Ndel  ! values at 0.5 and 1.0
      write(iout,'(i3,10(a1,g15.6))')n,(tb,erry(j,n),j=1,Neq), &
         (tb,ymid(1,j,n),tb,ymid(2,j,n),j=1,Neq)
   enddo
   ! ---------------------------------------
   Contains
   ! ---------------------------------------
   Subroutine Setup
      Use Array_Print, Only : OpenFile
      Real(float) :: Pem,Pet,Bix
      Real(float) :: Rxm,Rxt,Eac
      Integer :: Korder,i,ichk
      Character :: fname*80,ch*1,fexact*80
      ! default values
      typ = Gauss
      Nmin = 3 ; Nmax = 3 ; Ndel = 1
      Pem = 200.0_float ; Pet = 100.0_float
      Bix = 3.0_float
      Rxm = 0.5_float ; Rxt = 0.1_float  ! easier test problem
      Rxm = 0.5_float ; Rxt = 0.2_float  ! 1st example in thesis
 !    Rxm = 5.0_float ; Rxt = 0.0_float  ! isothermal
      Eac = 20.0_float
      Korder = 1
      ! ---------------------------------------------------
      write(*,'(a,i2,a,i2,a)')' Enter g/l/c/r/x for Gauss/Lobatto/Chebyshev/Radau(R/L) >'
      read(*,*)ch;  typ = index('glcrx',ch);  typ = Max(Gauss,Min(RadauL,typ))
      write(*,'(a)')' Enter min, max and delta N > '
      read(*,*)Nmin,Nmax,Ndel
      Nmin = max(Nmin,1)
      Nmax = min(max(Nmin,Nmax),MaxNmax)
      write(*,'(a)')' Solution method: 0 NR reduced, 1 NR full, 2 Picard >'
      read(*,*)MethSolve
      write(*,'(a)')' Enter Peclet No. (mass, heat) > '
      read(*,*)Pem,Pet
      write(*,'(a)')' Enter Heat Transfer Coef. > '
      read(*,*)Bix
      write(*,'(a)')' Enter Rate Coefs. (mass, heat) > '
      read(*,*)Rxm,Rxt
      write(*,'(a)')' Enter Activation Energy Paremeter > '
      read(*,*)Eac
      write(*,'(A)')' Enter output file name > '
      read(*,*)fname
      Call OpenFile(fname)
      if(Debug > 0)Call OpenFile(fname,ext='log')
      Call OpenFile(fname,append='2')
      Call RateParms(Korder,Eac)          ! rate parms to RateEqn module
      Call SetParms(Pem,Pet,Bix,Rxm,Rxt)  ! other parms to AxialDisp module
      xx = 0.0_float;  yx = 0.0_float;  wx = 0.0_float
      fexact = 'o/ex02axL99.dat'
      Open(11,File=trim(fexact),status='old',iostat=ichk)
      if(ichk .ne. 0)then
         write(*,*)'No file "'//trim(fexact)//'" with exact solution'
      else 
         do    ! lines before data
            read(11,*)fexact
            !write(*,*)fexact
            if(fexact(1:11) == 'Collocation')exit
         enddo
         do i = 0,Nx+1  ! data lines
            read(11,*)xx(i),yx(i,:)
         enddo
         close(11)
      endif
   End Subroutine Setup
   ! ---------------------------------------
   Function Fill(n)  result(xv)
      integer :: n,i
      Real(float) :: xv(0:(n+1)*Nsplit)
      Real(float) :: xc(0:n+1),wt,wtx
      ! create array xv which includes xc and Nsplit-1 points in between
      ! this is to improve accuracy of error norm calculation
      wtx = 1.0_float/Real(Nsplit,float);  wt = 0.0_float
      xc = Xpoints(OC);  xv(0:Nv:Nsplit) = xc
      do i = 1,Nsplit-1
         wt = wt + wtx
         xv(i:Nv+i-Nsplit:Nsplit) = xc(0:n)*(1.0_float - wt) + xc(1:n+1)*wt
      enddo
   End Function Fill
   ! ---------------------------------------
   Function ErrorCalc(xv,yv,ymid) result(erry)
      Real(float) :: xv(0:),yv(0:,:),ymid(:,:),erry(Neq)
      Real(float) :: xc(0:n1),yc(0:n1,Neq),yi(0:Nx+1,Neq),yxm(2,Neq)
      integer :: j,Nmid
      xc = xv(0::Nsplit);  yc = yv(0::Nsplit,:)  ! extract pt values
      yi = Interp(xx,OC,yc)  ! interpolate onto grid of "exact" solution
      ! calculate L1 error norm, assuming yx is exact
      do j=1,Neq
         erry(j) = Sum(wx*abs(yi(:,j)-yx(:,j)))
      enddo
      Nmid = (Nx+2)/2;  yxm(1,:) = yx(Nmid,:);   yxm(2,:) = yx(Nx+1,:)
      Nmid = (Nv+1)/2;  ymid(1,:) = yv(Nmid,:);  ymid(2,:) = yv(Nv,:)
      ymid = abs(ymid - yxm)
   End Function ErrorCalc
End Program AxialDispRxn
