Module FallingFilm
   ! ----------------------------------------------------------------
   ! This code solves mass transfer from falling film problem as described
   ! in B,S&L p. 538 and solved by Villadsen and Michelsen Ch. 4
   ! The problem is solved with the Galerkin method using eigenvalues
   ! and eigenvectors. The initial conditions are implemented in three ways.
   ! ----------------------------------------------------------------
   Use OrthogonalColloc, Only : ColDataType, Initialize, CollocCoef, MassMatrix, Lobatto
   Use Array_Print, only : ArrayPrint, VectorPrint
   include 'defs.fi'
   Public SolveFallingFilm
   Private
   Integer :: i, typ  ! 1 Gauss, 2 Lobatto points
   integer, public, parameter :: Debug = 2
   Integer, Parameter :: Nresultp = 6 ! ---- zp - profile calculation points to xxx2.dat
   Real(float), Parameter :: zp(Nresultp) = &
      (/0.0001_float,0.001_float,0.01_float,0.02_float,0.1_float,0.2_float/)
   ! "exact" values for yavg and flux at zv are stored in subroutine Calc
   Integer, Parameter :: Nresult = 41 ! ---- zv - y avg,  Sh and flux & errors calculations
   Real(float), Parameter :: zv(Nresult)=(/(10**(-4.0_float+0.1_float*Real(i,float)),i=0,Nresult-1)/)
   Integer, parameter :: Nfx = 3, Nmaxout = 10
   Real(Float) :: yv(Nresult,Nfx*2)  ! columns of output
!-------------------------------------------------------------------------------
   Contains
! ---------------------------------------------------------------------
   Subroutine SolveFallingFilm(n,cx)
      Use EigenValue, Only : EigenValSym, EigenVal
      Use LUSolvers, Only : LUSolve
      integer :: n,n1,i,ncx
      Type(ColDataType) :: OC
      Real(float) :: cx(:,:)
      Real(float) :: xc(0:n+1)       ! collocation points
      Real(float) :: wc(0:n+1)       ! quadrature weights
      Real(float) :: Ac(0:n+1,0:n+1) ! 1st derivattive
      Real(float) :: Cc(0:n+1,0:n+1) ! stiffness matrix
      Real(float) :: Dc(0:n+1,0:n+1)       ! mass matrix for z derivative
      Real(float) :: Dm(0:n+1,0:n+1)       ! mass matrix for initial conditions
      Real(float) :: y0(n+1)
      Real(float) :: wx(0:n+1),r0(0:n+1)  !wc*(1 - xc^2), rhs for i.c.
      Real(float) :: A(n+1,n+1),B(n+1,n+1)
      Real(float) :: xl(n+1),v(n+1,n+1),vf(n+1,n+1),c(n+1),yk(n+1,n+1),s(n+1)
      Real(float) :: yc(0:n+1,Nresultp)
      Integer, parameter :: ICmeth = 3  ! I.C. implementation method (see below)
      ! ----------------------------------------------------------------
      ! This code solves mass transfer from falling film problem
      ! ----------------------------------------------------------------
      n1 = n + 1;  typ = Lobatto
      call Initialize(OC,n,TType=typ)
      Call CollocCoef(OC,xc,wc,Ac,C=Cc)
      wx = wc*(1.0_float - xc**2)
      Dc = MassMatrix(OC,MassFunc,2) ! full mass matrix 1-x^2
      ! initial conditions, have tried:
      !  1. using constant value of 1 (same as collocation)
      !  2. using normal mass matrix
      !  3. using mass matrix with 1 - x^2 weighting
      Select Case(icmeth)
       case(1)
         y0 = 1.0_float    ! approximate by 1.0 (same as collocation)
       case(2)
         Dm = MassMatrix(OC)  ! use normal mass matrix
         y0 = LuSolve(Dm(1:n+1,1:n+1),wc(1:n+1))
       case(3)
         r0 = wx           ! weight by velocity profile (this is correct!)
         if(n == 1)r0(1:2) = sum(Dc(1:2,:))  ! quadrature exact except n = 1
         y0 = LuSolve(Dc(1:n+1,1:n+1),r0(1:n+1))
      End Select
      ! ---------------------------------------------------------------------
      !  Set up (A - lambda*B)x
      ! ---------------------------------------------------------------------
      A = Cc(1:,1:)
      B = Dc(1:,1:)
      Call PrintLog(1)
      xl = EigenValSym(A,B,v)    ! A & B symmetric
      xl = -xl;  vf = v
      c = LuSolve(vf,y0)
      do i=1,n1    ! coefficients for interior points
         yk(i,:) = v(i,:)*c
      enddo
      s = (1.5_float)*MatMul(wx(1:n),yk(1:n,:))
      ! -----------------
      ncx = min(n1,size(cx,1));  cx = 0.0_float
      cx(:ncx,1) = -xl(:ncx);  cx(:ncx,2) = s(:ncx)
      Call PrintLog(4)        ! .log file (activate with Debug in defs.fi)
      ! -------- integrate over time -----------
      Call Calc(xl,s,yk,yc)   ! calculate over z
      Call Results(xc,yc,yv)  ! .dat output
      ! ------------------------------------
      contains
      ! ------------------------------------
      Subroutine PrintLog(icase)
         integer :: icase
         real(float),parameter :: x1exact = -5.12166930737442_float
         real(float) :: xlz(n+1,4),errx,xmini,xmaxi,yinlet
         Character (len=*), parameter :: fmtv = '(1x,f5.1,20(a1,g22.15))'
         if(Debug < 2)return
         Select Case(icase)
         case(1)
            write(*,'(A,I3)')'Falling Film Eigenvalue Solution Galerkin Method, N =',n
            write(ilog,'(A,I3)')'SolveFallingFilm Eigenvalue Solution Galerkin Method, N =',n
            Call VectorPrint('Initial Values:',y0,flog=.true.)
            if(n <= Nmaxout) &
               Call ArrayPrint('Solution Matrix A & B diagonal for collocation',A,B,flog=.true.)
         case(4)
            xlz(:,1) = (/(Real(i,float),i=1,n+1)/);  xlz(:,2) = -xl;  xlz(:,3) = s
            yinlet = sum(s);  errx = abs(1.0_float - xl(1)/x1exact)
            xlz(:,4) = 0.0_float;
            xmini = Minval(xlz(:,4));  xmaxi = Maxval(xlz(:,4))
            write(ilog,'(A,A1,i3,A1,F18.15,A1,G15.8,A1,F12.8,A1,G15.8)')'Eigenvalues (min,max):',tb, &
               n,tb,minval(xlz(:,2)),tb,maxval(xlz(:,2)),tb,yinlet,tb,errx
!            Call VectorPrint('Boundary Flux Coefficients:',sf,flog=.true.)
            if(xmini /= xmaxi)then
               write(ilog,'(a,2g12.5)')' **** Warning eigenvalues imaginary:',xmini,xmaxi
               write(*,'(a,2g12.5)')' **** Warning eigenvalues imaginary:',xmini,xmaxi
            endif
            Call ArrayPrint('Eigenvalues & coefficients:',xlz,flog=.true.,fmtx=fmtv)
            if(n <= Nmaxout)then
               Call ArrayPrint('Eigenvalues & Eigenvectors',xl,v,flog=.true.)
               Call ArrayPrint('Coefficients of point values:',yk,flog=.true.)
            endif
         End Select
      End Subroutine PrintLog
      ! -------------------------------
      Function MassFunc(x) result(f)
         include 'defs.fi'
         Real(float), intent(in) :: x(:)
         Real(float) :: f(size(x))
           f = 1.0_float - x*x              ! falling film problem
      End Function MassFunc
   End Subroutine SolveFallingFilm
! ----------------------------------------------------------------------
   Subroutine Calc(xl,s,yk,yc)
! ----------------------------------------------------------------------
      real(float) :: xl(:),s(:),yk(:,:),yc(0:,:)     ! eigenvalues and coefficients
      real(float) :: expx(size(xl)),xsh
      integer :: j,n,n1
      ! values are given by: Sh, yavg, flux, errSh, erryavg, errflux
      integer, parameter :: ish=1,iy=2,idy=3,ise=4,iye=5,ide=6
      Real(float), parameter :: xyv(Nresult) = &  ! "exact" values of yavg at zv
      (/0.983074594615023D0,0.981009464783005D0,0.978692384823741D0,0.976092626130294D0,0.973175716400810D0,0.969902985218775D0, &
        0.966231055169997D0,0.962111272258640D0,0.957489068793326D0,0.952303251321919D0,0.946485205630360D0,0.939958010333820D0, &
        0.932635450249421D0,0.924420920657771D0,0.915206213899633D0,0.904870180760582D0,0.893277261142300D0,0.880275882167873D0, &
        0.865696727984171D0,0.849350895451222D0,0.831027965779397D0,0.810494047458406D0,0.787489886395725D0,0.761729205308586D0, &
        0.732897545556052D0,0.700652080291509D0,0.664623244999442D0,0.624419944341060D0,0.579643272204306D0,0.529924874923220D0, &
        0.475030072890847D0,0.415078515515748D0,0.350879529249625D0,0.284255713265780D0,0.218147686077766D0,0.156341797370461D0, &
        0.102789508910174D0,0.060627050504846D0,0.031190126112674D0,0.013509186015730D0,0.004711409556011D0 /)
      Real(float), parameter :: xdy(Nresult) = &  ! "exact" values of flux at zv
      (/84.6242054415916D0,75.4204256639378D0,67.2174290111613D0,59.9063666587493D0,53.3902251921072D0,47.5825392890876D0, &
        42.4062443677638D0,37.7926539718109D0,33.6805483225400D0,30.0153619406521D0,26.7484595543178D0,23.8364906801133D0, &
        21.2408143049916D0,18.9269860243048D0,16.8643008146857D0,15.0253853514262D0,13.3858344261979D0,11.9238865888791D0, &
        10.6201346306266D0,9.45726694400754D0,8.41983613330169D0,7.49405148485908D0,6.66759199847908D0,5.92943652026305D0, &
        5.26970683818502D0,4.67951758126704D0,4.15081921863026D0,3.67617990163086D0,3.24828923352892D0,2.85878058217229D0, &
        2.49668236829310D0,2.14869297218496D0,1.80334553847818D0,1.45709258371523D0,1.11743866250787D0,0.800742990114463D0,&
        0.526454339601708D0,0.310511711593497D0,0.159745511650109D0,0.0691895833845128D0,0.0241302817175155D0 /)
      ! ------------------------------
      n = size(xl);  n1 = n+1
      do j=1,Nresult
         expx = expfunc(xl*zv(j))
         yv(j,iy) = sum(s*expx)
         yv(j,idy) = -sum(s*xl*expx)
         yv(j,ish) = yv(j,idy)/(yv(j,iy)*1.5_float)
         yv(j,iye) = abs(1.0_float - yv(j,iy)/xyv(j))
         yv(j,ide) = abs(1.0_float - yv(j,idy)/xdy(j))
         xsh = xdy(j)/(xyv(j)*1.5_float)
         yv(j,ise) = abs(1.0_float - yv(j,ish)/xsh)
      enddo
      yc(0,:) = 0.0_float
      do j=1,Nresultp
         expx = expfunc(xl*zp(j))
         yc(1:,j) = Matmul(yk,expx)
      enddo
    contains
      function expfunc(x) result(ex)
         Real(float) :: x(:),ex(size(x))
         integer :: i
         do i = 1,size(x)
            ex(i) = 0.0_float
            if(x(i) > -600.0)ex(i) = exp(x(i))
         enddo
      end function expfunc
   End Subroutine Calc
! ----------------------------------------------------------------------
   Subroutine Results(xc,yc,yv)
! ----------------------------------------------------------------------
      Real(float) :: xc(0:),yc(0:,:),yv(:,:)
      Integer :: n,i
      Logical :: First = .true.
      Character(len=8), parameter :: Hdr1(2*Nfx+1) = &
         (/'  "z"   ','  "Sh"  ',' "yavg" ',' "Flux" ','"Err Sh"','"Err ya"','"Err fx"'/)
      Character (len=*), parameter :: titl = 'Title= "Falling Film Problem"'
      Character (len=*), parameter :: fmtv = '(1x,f10.7,3(a1,g22.15),4(a1,e11.4))'
      Character (len=*), parameter :: fmtz = '(20(f10.7,A1))'
      Character :: cn*7,cz(Nresultp)*10
      n = size(xc) - 2
      if(First)then
         write(iout,'(A)')titl
         write(iout,'(40A)')'Variables=',(tb,Hdr1(i),i=1,2*Nfx+1)
         write(iout2,'(A)')titl
         do i = 1,Nresultp
            cz(i) = '"z=x.xxxx"';  write(cz(i)(4:9),'(F6.4)')zp(i)
         enddo
         write(iout2,'(40A)')'Variables= "x"',(tb,cz(i),i=1,Nresultp)
         First = .false.
      endif
      cn = ' n =xx"'; write(cn(5:6),'(I2)')n;
      Call ArrayPrint('Zone T="Galerkin'//cn,zv,yv,fmtx=fmtv)
      Call ArrayPrint('Zone T="Galerkin'//cn,xc,yc,flog=.false.,fmtx=fmtz)
   End Subroutine Results
! ----------------------------------------------------------------------
End Module FallingFilm
! -----------------------------------------------------------------------
Program Example04FF
! -----------------------------------------------------------------------
   Use FallingFilm
   Use Array_Print, Only : OpenFile, ArrayPrint
   include 'defs.fi'
   Integer :: Nmin, Nmax, Ndel
   Integer, Parameter :: MaxNmax = 100, Nsave = 10
   Integer :: n,i
   Character :: fname*80
   Real(float) :: cx(Nsave,2),xn(MaxNmax,0:Nsave,4)
   Character (len=*), parameter :: fmtv = '(f5.0,21(a1,g22.15))'
   ! ---------------------------------------------------
   ! Main for Falling Film problem
   ! ---------------------------------------------------
   Nmin = 2 ; Nmax = 2 ; Ndel = 1
   ! ---------------------------------------------------
   write(*,'(a)')' Enter min, max and delta N > '
   read(*,*)Nmin,Nmax,Ndel
   Nmin = max(Nmin,1)
   Nmax = min(max(Nmin,Nmax),MaxNmax)
   write(*,'(A)')' Enter output file base name > '
   read(*,*)fname
   Call OpenFile(fname)
   if(Debug > 0)Call OpenFile(fname,ext='log')
   Call OpenFile(fname,append='2')
   ! ------------------------------------------------------------
   i = 0
   do n = Nmin,Nmax,Ndel
      i = i + 1
      Call SolveFallingFilm(n,cx)
      xn(i,0,:) = Real(n,float);     xn(i,1:,:2) = cx
      Call ErrorCalc(xn(i,1:,3:4))
   enddo
   Call ArrayPrint('Lower Eigenvalues',xn(:i,:,1),fmtx=fmtv)
   Call ArrayPrint('Lower Coefficients',xn(:i,:,2),fmtx=fmtv)
   Call ArrayPrint('Eigenvalue Errors',xn(:i,:,3),fmtx=fmtv)
   Call ArrayPrint('Coefficient Errors',xn(:i,:,4),fmtx=fmtv)
 Contains
   Subroutine ErrorCalc(errx)
      Real(float) :: errx(:,:)
      integer :: nk
      integer, parameter :: nxx = 10    ! best estimate of first 10 from spreadsheet
      Real(float), parameter :: xlxx(nxx)= (/ & ! eigenvalues
         5.12166930737427D+00,3.96608389141836D+01,1.06249232183627D+02,2.04856060486413D+02,3.35473196978814D+02, &
         4.98097083681151D+02,6.92725799446885D+02,9.19358173444519D+02,1.17799343263235D+03,1.46863103652777D+03/)
      Real(float), parameter :: cmxx(nxx)= (/ & ! coefficients
         7.89702616221030D-01,9.72551113742101D-02,3.60936164567818D-02,1.86863735945829D-02,1.14017596768274D-02, &
         7.67597018871592D-03,5.51794271090871D-03,4.15703395458558D-03,3.24397369495590D-03,2.60179461467994D-03/)
      nk = size(errx,1)
      errx(:,1) = abs(1.0_float - cx(:,1)/xlxx(1:nk))
      errx(:,2) = abs(1.0_float - cx(:,2)/cmxx(1:nk))
   End Subroutine ErrorCalc
End Program Example04FF

