Module FallingFilm
   ! ---------------------------------------------------
   Use OrthogonalColloc, Only : ColDataType, Initialize, CollocCoef, TxtType, Chebyshev
   Use LUSolvers, Only : LUFactrLDL, LUSubstLDL, LUFactrC, LUSubstC
   Use Array_Print, only : ArrayPrint, VectorPrint
   include 'defs.fi'
   Public SolveFallingFilm
   Private
   Integer :: n,n1,typ                       ! typ points - Gauss, Lobatto, etc.
   Integer, parameter :: Nmaxout = 10        ! max n for detailed output
   Logical, parameter :: BColloc = .false.   ! change manually and recompile
   Real(float), Allocatable :: C0(:),Cn1(:),wx(:)
   Real(float), Allocatable :: ys(:,:),yc(:,:)
!-------------------------------------------------------------------------------
   Contains
! ---------------------------------------------------------------------
   Subroutine SolveFallingFilm(nn,typx,zmethod,zmax,nstep,nmax)
      integer :: nn,typx,zmethod,nstep,nmax
      Real(float) :: dz,zmax
      Type(ColDataType) :: OC
      Real(float) :: xc(0:nn+1)        ! collocation points
      Real(float) :: wc(0:nn+1)        ! quadrature weights
      Real(float) :: Ac(0:nn+1,0:nn+1) ! 1st derivattive
      Real(float) :: Cc(0:nn+1,0:nn+1) ! stiffness matrix
      Real(float) :: wdz(nn)           ! wc*(1 - xc^2)
      Real(float) :: A(nn,nn),Aw(nn,nn)
      Real(float) :: z,dzi,y(nn),dy1(nn),dy0(nn),r(nn)
      Real(float) :: yx(3)
      integer :: j,k,nrk,mrk,nbk
      Integer, parameter :: mstart(2) = (/1,3/) ! Dirk method to start backward diff.
      logical :: Nonsym
      ! ----------------------------------------------------------------
      ! This code solves mass transfer from falling film problem as described
      ! in B,S&L p. 538 and solved by Villadsen and Michelsen [1978, Ch. 4]
      ! and Finlayson [1972, 2014, pp 41, 58]
      ! The code is set up for solution by collocation with
      !    Gauss, Lobatto, Chebyshev, Radau (L or R) points
      ! It can also treat the no flux boundary condition either naturally or
      ! by boundary collocation (set BColloc = .true. above)
      ! The system of linear algebraic equations is solved using various
      ! stepping methods
      ! ----------------------------------------------------------------
      n = nn; n1 = n + 1;  typ = typx
      Allocate(wx(n),C0(n),Cn1(n))
      Allocate(ys(0:nmax,3),yc(0:nmax,n+1))
      Nonsym = typ == Chebyshev .or. BColloc ! nonsymmetric matrix
      call Initialize(OC,n,TType=typ)
      Call CollocCoef(OC,xc,wc,Ac,C=Cc)
      if(BColloc)Cc(n1,:) = Ac(n1,:)   ! convert to boundary collocation
      Cn1 = -Cc(n1,1:n)/Cc(n1,n1)      ! coefficients of y(n+1)
      C0 = Cc(0,1:n) + Cc(0,n1)*Cn1    ! coefficients for flux
      do j = 1,n
         A(:,j) = Cc(1:n,j) + Cc(1:n,n1)*Cn1(j)
      end do
      wx = wc(1:n)*(1.0_float - xc(1:n)**2)
      Call PrintLog
      dz = zmax/Real(nmax,float);  dzi = (1.0_float)/dz;  wdz = -dz/wx
      z = 0.0_float;    y = 1.0_float;   ys = 0.0_float;   yc = 0.0_float
      yx = CalcFlux(y,0)
      Select Case(zmethod)
      case(1)  ! Forward Euler --------------------------------------------
         do k = 1,nmax
            y = y + MatMul(A,y)*wdz
            yx = CalcFlux(y,k)
            if(abs(yx(2)) > 10000.0_float)exit
         enddo
      case(2)     ! 2nd Order R-K, Heun, improved Euler ----------------------
         do k = 1,nmax
            dy0 = MatMul(A,y)*wdz
            y = y + dy0
            dy1 = MatMul(A,y)*wdz
            y = y + (dy1 - dy0)*0.5_float
            yx = CalcFlux(y,k)
            if(abs(yx(2)) > 10000.0_float)exit
         enddo
      case(3)     ! Trapezoidal Rule -----------------------------------------
         Aw = BuildF(A,wx*(2.0_float/dz)) ! factor matrix
         do k = 1,nmax
            r = (-2.0_float)*MatMul(A,y)
            y = y + LUSubstLDL(Aw,r)      ! forward & back solve on rhs
            yx = CalcFlux(y,k)
         enddo
      case(4:6)   ! 1st, 2nd, 3rd order backward diff. -----------------------
         nbk = zmethod - 4          ! no. points to look back
         if(nbk > 0)then
            mrk = mstart(nbk)       ! use DIRK method to start
            nrk = DirkStages(mrk)
            Call Dirk(A,nrk,mrk,dz,nbk)
         endif
         Call BackDiff(A,dz,nbk,nmax)
      case(7:)    ! DIRK methods  ---------------------------------------------
         mrk = zmethod - 6       ! r-k method
         nrk = DirkStages(mrk)   ! no. stages
         Call Dirk(A,nrk,mrk,dz,nmax)
      case default
         stop 'Bad Choice'
      End Select
      Call Report(nmax,nstep,dz)
      DeAllocate(wx,C0,Cn1,ys,yc)
      ! ------------------------------------
      contains
      ! ------------------------------------
      Subroutine PrintLog
         if(BColloc)then
            write(*,'(a)')'Using Bpundary Collocation'
            write(iout,'(a)')'# Using Bpundary Collocation'
         endif
         if(n <= Nmaxout)Call ArrayPrint(  &
            '#Solution Matrix A & B diagonal',A,wx,fmtx='("#",20(f14.9,a1))')
      End Subroutine PrintLog
   End Subroutine SolveFallingFilm
   ! ---------------------------------------------
   Subroutine BackDiff(A,dz,nbk,nmax)
      Real(float) :: A(:,:),Ax(n,n)
      Integer :: nbk,nmax,j,k
      Real(float) :: dz,ak(nbk+2)
      Real(float) :: wi(n),r(n),y(n),yx(3)
      Real(float), parameter :: bk0(4) = (/1.0_float,-1.0_float,0.0_float,0.0_float/)
      Real(float), parameter :: bk1(4) = (/1.5_float,-2.0_float,0.5_float,0.0_float/)
      Real(float), parameter :: bk2(4) = (/11.0_float,-18.0_float,9.0_float,-2.0_float/)/(6.0_float)
      Real(float), parameter :: bkx(4,0:2) = Reshape((/bk0,bk1,bk2/),(/4,3/))
      Character (len=*), parameter :: fmts = '(2a,i6,a1,f7.3,20(a1,g17.10))'
      ! ------------------------------------------
      ! Backward Difference: 0 - backward Euler, 1 - 2nd order, 2 - 3rd order
      ! ------------------------------------------
      ak = bkx(:nbk+2,nbk)/dz;  ak(3:) = ak(3:)/ak(2);   wi = -wx*ak(2)
      Call VectorPrint('#Backward Difference parameters:',bkx(:nbk+2,nbk),fmtx='("#",20(f10.6,a1))')
      Ax = BuildF(A,wx*ak(1))
      !call vectorprint('ak',ak)
      !call arrayprint('Ax:',Ax,fmtx='(25(f20.15,a1))')
      do k = 0,nbk
         write(iout,fmts)'# Step:',tb,k,tb,k*dz,(tb,yc(k,j),j=1,n)
      enddo
      do k = nbk+1,nmax
         if(nbk == 0)r = wi*yc(k-1,1:n)
         if(nbk == 1)r = wi*(yc(k-1,1:n) + ak(3)*yc(k-2,1:n))
         if(nbk == 2)r = wi*(yc(k-1,1:n) + ak(3)*yc(k-2,1:n) + ak(4)*yc(k-3,1:n))
         y = LUSubstLDL(Ax,r)
         !write(iout,fmts)'# Step r,y:',tb,k,tb,k*dz,(tb,r(j),j=1,n),(tb,y(j),j=1,n)
         yx = CalcFlux(y,k)
         !write(iout,fmts)'# Step:',tb,k,tb,k*dz,(tb,yc(k,j),j=1,n),(tb,yx(j),j=1,3)
      enddo
   End Subroutine BackDiff
   ! ------------------------------------
   Subroutine DIRK(A,ns,method,dz,nmax)
      Real(float) :: A(:,:),Ax(n,n)
      Integer :: ns,method,nmax,k,i
      Real(float) :: ar(ns,ns),cr(ns),br(ns)
      Real(float) :: dz,y(n),y0(n),f(n,ns)
      Real(float) :: wi(n),wa(n),yx(3)
      Real(float) :: ad,adi
      Character (len=*), parameter :: fmts = '(2a,i2,a1,i6,a1,f7.3,20(a1,g15.8))'
      Logical :: Explicit1
      ! ------------------------------------
      ! Diagonally Implicit Runge-Kutta - set up for a general method
      ! ns - number of stages
      ! method - method designator
      ! DirkParm - provides the parameters given the method
      ! DirkStages - returns the no. stages given the method
      ! Explicit1 - .true. for explicit 1st stage
      ! ------------------------------------
      Call DirkParm(ns,method,ar,br,cr)
      Explicit1 = MaxVal(ar(1,:)) == 0.0_float
      ar = ar*dz; br = br*dz; cr = cr*dz
      ad = ar(ns,ns); adi = 1.0_float/ad
      wa = wx/ad;  wi = (-1.0_float)/wx
      y = 1.0_float;
      yx = CalcFlux(y(:),0)
      Ax = BuildF(A,wa)
      do k = 1,nmax
         i = 0; y0 = y
         if(Explicit1)then
            i = 1; f(:,1) = wi*MatMul(A,y)
!           write(iout,fmts)' Step:',tb,i,tb,k,tb,k*dz,(tb,y(j),j=1,n),(tb,f(j,1),j=1,n)
         endif
         do i = i+1,ns ! loop over implicit stages
            y = y0 + MatMul(f(:,1:i-1),ar(i,1:i-1))
            f(:,i) = (LUSubstLDL(Ax,wa*y) - y)*adi
            !write(iout,fmts)' Step:',tb,i,tb,k,tb,k*dz,(tb,y(j),j=1,n),(tb,f(j,i),j=1,n)
         enddo
         y = y0 + MatMul(f,br)
         yx = CalcFlux(y,k)
        !write(iout,fmts)'Final:',tb,i,tb,k,tb,k*dz,(tb,y(j),j=1,n)
      enddo
   End Subroutine
   ! ------------------------------------
   Function DirkStages(method) result(ns)  ! returns no. stages
      Integer :: method,ns
      Integer, parameter :: NDIRKS = 5   ! number methods stored
      Integer, Parameter :: nstage(NDIRKS) = (/2,3,3,4,5/)
      ! General framework for DIRK methods
      !   DirkStages - returns the no. stages for given method
      !   DirkParm - returns the Butcher tableu
      !   methods:
      !     1. SDIRK - 2nd order, 2 stage, 2 implicit stages
      !     2. EDIRK - 2nd order, 3 stage, 2 implicit stages
      !     3. SDIRK - 3rd order, 3 stage, 3 implicit stages
      !     4. EDIRK - 3rd order, 4 stage, 3 implicit stages
      !     5. EDIRK - 3rd order, 5 stage, 4 implicit stages
      ! This framework could be expanded with its own module
      ! and additional functions, such as dense output,
      ! step size control, etc.
      ns = nstage(method)
   End Function DirkStages
   ! ------------------------------------
   Subroutine DirkParm(ns,method,a,b,c)   ! returns Butcher tableu
      Integer :: ns,method,i
      Real(float) :: a(ns,ns),b(ns),c(ns) ! ns stages & method
      Integer, parameter :: NDIRKS = 5   ! number methods stored
      Real(float), parameter :: sqr2 = sqrt(2.0_float)
      ! Parameters for ESDIRK2(1)3L[2]SA (2 implicit stages)
      Real(float), parameter :: E3g = 1.0_float - (1.0_float)/sqr2
      ! Parameters for ESDIRK3(2)4L[2]SA (3 implicit stages)
      Real(float), parameter :: E4g = (0.43586652150845899941601945_float)
      Real(float), parameter :: E4c3 = (0.6_float)
      Real(float), parameter :: E4xx = E4c3-2.0_float*E4g
      Real(float), parameter :: E4a32 = E4c3*E4xx/(4.0_float*E4g)
      Real(float), parameter :: E4b2 = (-2.0_float+3.0_float*E4c3+6.0_float*E4g*(1.0_float-E4c3)) &
                                       /(12.0_float*E4g*E4xx)
      Real(float), parameter :: E4b3 =(1.0_float-6.0_float*E4g+6.0_float*E4g**2)/(3.0_float*E4c3*E4xx)
      ! Parameters for ESDIRK3(2)5L[2]SA (4 implicit stages)
      Real(float), parameter :: E5g = (9.0_float)/(40.0_float)
      Real(float), parameter :: E5c(5) = (/0.0_float, (2.0_float*E5g), &
         (2.0_float + sqr2)*E5g, (3.0_float)/(5.0_float), 1.0_float /)
      Real(float), parameter :: E5a43 = (-7.0_float)/((1.0_float+sqr2)*40.0_float)
      Real(float), parameter :: E5a54 = (5827.0_float)/(7560.0_float)
      Real(float), parameter :: E5a53 =(-2374.0_float)*(1.0_float+2.0_float*sqr2) &
                                          /(2835.0_float*(5.0_float+3.0_float*sqr2))
      ! Method diagonals, must set a(1,1) back to zero for EDIRKs
      Real(float), parameter :: Diag(NDIRKS) = (/E3g,E3g,E4g,E4g,E5g/)
      Integer, parameter :: first(NDIRKS) = (/1,2,1,2,2/)
      ! ------------------------------------------------------------------------
      a = 0.0_float
      do i = first(method),ns
         a(i,i) = Diag(method)
      enddo
      Select Case(method)
      Case(1)  ! --- 2nd order 2 stage SDIRK (p. 72 Nasa review, Alexander 1977)
         a(2,1) = 1.0_float - E3g
      Case(2)  ! --- 2nd order 3 stage ESDIRK2(1)3L[2]SA (p. 73 Nasa review)
         a(3,1) = (0.25_float - 0.50_float*E3g)/E3g
         a(2,1) = E3g;  a(3,2) = a(3,1)
      Case(3)  ! --- 3rd order 3 stage SDIRK (p. 77 Nasa review, Alexander 1977)
         a(3,2) = 1.25_float + E4g*((1.5_float)*E4g - 5.0_float)
         a(3,1) = 1.0_float - a(3,2) - a(3,3);  a(2,1) = (1.0_float - E4g)*(0.5_float)
      Case(4)  ! --- 3rd order 4 stage ESDIRK3(2)4L[2]SA (p. 75 Nasa review)
         a(2,1) = E4g;  a(3,2) = E4a32; a(3,1) = E4c3 - sum(a(3,2:3))
         a(4,3) = E4b3; a(4,2) = E4b2;  a(4,1) = 1.0_float - sum(a(4,2:4))
      Case(5)  ! --- 3rd order 5 stage ESDIRK3(2)5L[2]SA (superconvergent?)
         a(2,1) = E5g; a(4,3) = E5a43; a(5,4) = E5a54;  a(5,3) = E5a53;
         do i=3,5
            a(i,1:2) = (E5c(i) - sum(a(i,3:i)))*0.5_float
         enddo
      End Select
      b = a(ns,:)  ! stiffly accurate (make exception if not true)
      do i = 1,ns
         c(i) = sum(a(i,:))   ! c = row sum(a)
      enddo
      !Call ArrayPrint('Implicit R-K coefficient array c(i), a(i,j):',c,a)
   End Subroutine DirkParm
   ! ------------------------------------
   Function CalcFlux(y,k) result(yx)
      Real(float) :: y(:), yx(3)
      Integer :: k, n
      n = size(y)
      yc(k,1:n) = y;                ! save values
      yc(k,n+1) = sum(Cn1*y)        ! y(n+1)
      yx(1) = (1.5_float)*sum(wx*y);! average y
      yx(2) = -sum(C0*y)            ! flux
      yx(3) = yx(2)/yx(1)           ! Sh no.
      ys(k,:) = yx;
   End Function CalcFlux
   ! ------------------------------------
   Function BuildF(A,w) result(Aw)
      Real(float) :: A(:,:),w(:)
      Real(float) :: Aw(size(A,1),size(A,2))
      Integer :: i
!      Call VectorPrint('BuildF w:',w)
!      Call ArrayPrint('BuildF A:',A)
      Aw = A
      do i=1,n
         Aw(i,i) = Aw(i,i) + w(i)
      enddo
      !Call ArrayPrint('BuildF A:',Aw,fmtx='(25(f20.15,a1))')
      Call LUFactrLDL(Aw)
   End Function BuildF
! ----------------------------------------------------------------------
   Subroutine Report(nmax,nstep,dz)
      Integer :: nmax,nstep,i,k
      Real(float) :: dz,errx(2)
      ! These are the exact average y and flux at z = 0.10, from continuous z solution
      ! for n = 1-10 for Lobatto & RadauL, other cases not loaded, so will be wrong
      Real(float), parameter :: ylobat(10) = (/  &
         0.474817419141886D0,0.479530842493556D0,0.474675015955082D0,0.475021278543443D0,0.475034509697763D0, &
         0.475029659864105D0,0.475030086575185D0,0.475030073809602D0,0.475030072721024D0,0.475030072903210D0 /)
      Real(float), parameter :: flobat(10) = (/  &
         1.447062610718130D0,1.711224323205850D0,1.664223256695420D0,1.663576790580970D0,1.664588476609990D0, &
         1.664446145498130D0,1.664454745061190D0,1.664454996659740D0,1.664454903464960D0,1.664454912691200D0 /)
      Real(float), parameter :: yradaul(10) = (/  &
         0.416860506786547D0,0.484061180960934D0,0.475491519630788D0,0.474878890444410D0,0.475042955492243D0, &
         0.475029940423042D0,0.475029990734273D0,0.475030082246454D0,0.475030072368794D0,0.475030072896541D0 /)
      Real(float), parameter :: fradaul(10) = (/  &
         1.125523368323670D0,1.673207309014650D0,1.683544986708120D0,1.661404146073370D0,1.664586327542870D0, &
         1.664474413408500D0,1.664450878272140D0,1.664455240619530D0,1.664454901766590D0,1.664454911376490D0 /)
      Real(float), parameter :: yexact(10,5) = Reshape((/ylobat,ylobat,ylobat,yradaul,yradaul/),(/10,5/))
      Real(float), parameter :: fexact(10,5) = Reshape((/flobat,flobat,flobat,fradaul,fradaul/),(/10,5/))
      Character (len=*), parameter :: fmts = '(i6,a1,f7.3,a1,f11.8,20(a1,g22.15))'
      Character (len=*), parameter :: fmt2 = '(f10.7,a1,f6.2,8(a1,g12.5))'
      do k=0,nmax
         if(ys(k,1) <= 0.0_float)exit
         write(iout,fmts)k,tb,k*dz,tb,dz,(tb,ys(k,i),i=1,3)   !,(tb,yc(k,i),i=1,n+1)
      enddo
      errx(1) = abs(1.0_float - ys(nstep,1)/yexact(n,typ))  ! yavg erro
      errx(2) = abs(1.0_float - ys(nstep,2)/fexact(n,typ))  ! flux error
      write(iout2,fmt2)dz,tb,dz*nstep,(tb,ys(nstep,i),i=1,3),tb,errx(1),tb,errx(2)
   End Subroutine Report
! ----------------------------------------------------------------------
End Module FallingFilm
! -----------------------------------------------------------------------
Program Example04FF
! -----------------------------------------------------------------------
   Use FallingFilm
   Use OrthogonalColloc, Only : Gauss,Lobatto,Chebyshev,RadauR,RadauL,TxtType
   Use Array_Print, Only : OpenFile
   include 'defs.fi'
   Integer :: n,i,typ,zmethod,nstep
   Integer, Parameter :: ndelz = 10
   Real(float) :: dz(ndelz),zmax,z01
   Character :: fname*80,ch*1
   ! ---------------------------------------------------
   ! Main for FallingFilm problem
   ! Note: can run using Ex04 < Ex.dat if you put
   ! data in file Ex.dat, otherwise will be interactive
   ! ---------------------------------------------------
   n = 2; typ = Lobatto; dz = -1.0_float
   ! ---------------------------------------------------
   write(*,'(a,i2,a,i2,a)')'Enter g/l/c/r/x for Gauss/Lobatto/Chebyshev/Radau(R/L) >'
   read(*,*)ch;  typ = index('glcrx',ch);  typ = Max(Gauss,Min(RadauL,typ))
   write(*,'(a)')' Enter n > '
   read(*,*)n
   Call PrintHdr(0) ! time integration methods
   read(*,*)zmethod
   write(*,'(a)')'Step Sizes (10 or less, end with "/") > '
   read(*,*)dz
   write(*,'(A)')' Enter output file base name > '
   read(*,*)fname
   Call OpenFile(fname)
   Call OpenFile(fname,append='2')
   Call PrintHdr(1)
   ! -----------------------------------
   z01 = 0.10_float
   do i=1,ndelz
      zmax = 4*z01
      if(dz(i) < 0.0)exit
      nstep = Int((z01 + dz(i)*0.10_float)/dz(i));  dz(i) = z01/Real(nstep,float)
      Call PrintHdr(2)
      Call SolveFallingFilm(n,typ,zmethod,zmax,nstep,4*nstep)
   enddo
   ! -----------------------------------
   Contains
   ! -----------------------------------
   Subroutine PrintHdr(ix)
      integer :: ix
      integer, parameter :: Nmethod = 11
      Character (len=14), parameter :: Method(Nmethod) = (/ 'Forward Euler ', &
         'Improved Euler','Trapezoid Rule','Backward Euler','2nd Order Bkwd', &
         '3rd Order Bkwd','SDIRK 2nd Ord.','EDIRK 2nd Ord.','SDIRK3 3 stage', &
         'EDIRK3 4 stage','EDIRK3 5 stage' /)
      Character(len=6), parameter :: Hdr1(8) = &
         (/'"Step"',' "z"  ','"delz"',' "y"  ','"Flux"',' "Sh" ','"erry"','"errf"'/)
      Select Case(ix)
      Case(0)
         write(*,'(a)')'Enter the Time Integration Method from:'
         do i=1,Nmethod
            write(*,'(I3,2A)')i,'. ',Method(i)
         enddo
      Case(1)
         write(iout,'(5A,I3,A1)')'Title = "Falling Film, ',Method(zmethod),' Method, ', &
            TxtType(typ),' Points, n =',n,'"'
         write(iout2,'(5A,I3,A1)')'Title = "Falling Film, ',Method(zmethod),' Method, ', &
            TxtType(typ),' Points, n =',n,'"'
         write(iout,'(20a)')'Variables = ',(Hdr1(i),tb,i=1,6)
         write(iout2,'(20a)')'Variables = ',Hdr1(3),tb,Hdr1(2),(tb,Hdr1(i),tb,i=4,8)
         write(iout2,'(3a,f7.4)')'Zone T="',Method(zmethod),'"'
      Case(2)
         write(iout,'(2a,f8.5,a)')'Zone T="',Method(zmethod),dz(i),'"'
         write(*,'(5A,I3,A,f9.6)')'Falling Film, ',Method(zmethod),' Method, ', &
            TxtType(typ),' Points, N =',n,' delz =',dz(i)
      End Select
   End Subroutine PrintHdr
End Program Example04FF

