Module LUSolvers
   Use Array_Print, only : ArrayPrint
   include 'defs.fi'
   Integer, Parameter :: DumpF = 0 ! = Debug
   Private
   ! Native code: compact routines with No Pivoting (not needed for symmetric)
   Public LUSolveC      ! solves Ax = b after factoring with Crout
   Public LUFactrC      ! factors A by Crout algorithm, unit upper
   Public LUSubstC      ! solves Ax = b given factored A from LUFactrC
   Public LUSolveSq     ! solves Ax = b when A is symmetric, positive definite
   Public LUFactrSq     ! symmetric factorization A = LL*, square root method
   Public LUSubstSq     ! symmetric solve given factored A = LL*
   Public LUSolveLDL    ! solves Ax = b when A is symmetric, positive definite
   Public LUFactrLDL    ! symmetric factorization into LDL* (no square roots)
   Public LUSubstLDL    ! symmetric solve given factored A = LDL*
   ! LApack routines called
   Public LUSolve       ! solves Ax = b after factoring A (LAPack)
   Public LUFactr       ! factors A (LAPack)
   Public LUSubst       ! solves Ax = b given factored A (LAPack)
   Public LUInvert      ! calculates inverse A (LAPack)
   Public LUSolveSym    ! solves Ax = b when A is symmetric, positive definite
   Public LUFactrSym    ! symmetric factorization A = LL*, square root method
   Public LUSubstSym    ! symmetric solve given factored A = LL*
   Public LUInvertSym   ! calculates inverse A (LAPack)
   Public LUSolveSB     ! Symmetric band solves Ax = b after factoring A (LAPack)
   Public LUFactrSB     ! Symmetric band factors A (LAPack)
   Public LUSubstSB     ! Symmetric band solves Ax = b given factored A (LAPack)
   Public LUInvertSB    ! Symmetric band calculates inverse A (LAPack)
   Interface LUSubstC    ! overload
      Module Procedure LUSubstC1, LUSubstCn ! for 1 or n right hand sides
   End Interface
   Interface LUSolveC    ! overload
      Module Procedure LUSolveC1, LUSolveCn ! for 1 or n right hand sides
   End Interface
Contains
! ------------------------------------------------------------------
   Function LUSolveC1(A,b)  Result(x)
      Real(float), intent(in) :: A(:,:),b(:)
      Real(float) :: x(size(b)),Ax(size(A,1),size(A,2))
      Ax = A;
      Call LuFactrC(Ax)       ! factor with Crout
      x = LUSubstC1(Ax,b)     ! forward and back single rhs
   End Function LUSolveC1
   ! ------------------------------------------------------------------
   Function LUSolveCn(A,b)  Result(x)
      Real(float), intent(in) :: A(:,:),b(:,:)
      real(float) :: x(size(b,1),size(b,2)),Ax(size(A,1),size(A,2))
      Ax = A;
      Call LuFactrC(Ax)
      x = LUSubstCn(Ax,b)     ! forward and back multiple rhs
   End Function LUSolveCn
   ! ------------------------------------------------------------------
   Subroutine LUFactrC(A)     ! Crout or Dolittle algorithm
      Real(float) :: A(:,:)   ! A must be well conditioned (no pivoting)
      Integer :: n,i,j        ! factors with unit upper diagonal, inverse on diagonal
      ! -----------------------------------------
      n = size(A,1)
      do i=1,n
         A(i,i) = (1.0_float)/(A(i,i) - Sum(A(i,1:i-1)*A(1:i-1,i)))
         do j=i+1,n
            A(i,j) = (A(i,j) - sum(A(i,1:i-1)*A(1:i-1,j)))*A(i,i)
            A(j,i) =  A(j,i) - sum(A(j,1:i-1)*A(1:i-1,i))
         enddo
      enddo
      if(DumpF > 1)write(ilog,*)'LUFactrC - Crout algorithm'
      if(DumpF > 3)Call LUTest(A,Option=2)
   End Subroutine LUFactrC
   ! ------------------------------------------------------------------
   Function LUSubstC1(A,b)  Result(x)
      real(float), intent(in) :: A(:,:), b(:)
      real(float) :: x(size(b))
      integer :: n,i
      ! Forward & back solve for LU with unit upper, i.e. U(k,k) = 1
      ! Inverse of lower diagonal stored. 1/L(k,k)
      ! -----------------------------------------
      x = b;   n = size(b)
      x(1) = x(1)*A(1,1)   ! forward sweep
      do i=2,n
         x(i)= (x(i) - sum(A(i,1:i-1)*x(1:i-1)))*A(i,i)
      enddo
      do i=n-1,1,-1        ! back substitution
         x(i)=x(i) - sum(A(i,i+1:n)*x(i+1:n))
      enddo
      if(DumpF > 1)write(ilog,*)'LUSubstC - Crout algorithm'
   End Function LUSubstC1
   ! ------------------------------------------------------------------
   Function LUSubstCn(A,b)  Result(x)
      real(float), intent(in) :: A(:,:), b(:,:)
      real(float) :: x(size(b,1),size(b,2))
      integer :: n(2),j
      ! Forward & back solve for multiple rhs
      ! LU with unit upper, i.e. U(k,k) = 1
      ! Inverse of lower diagonal stored. 1/L(k,k)
      ! -----------------------------------------
      n = shape(b)
      do j=1,n(2)
         x(:,j) = LUSubstC1(A,b(:,j))
      enddo
   End Function LUSubstCn
! ------------------------------------------------------------------
   Function LUSolveLDL(A,b)  Result(x)
      Real(float), intent(in) :: A(:,:),b(:)
      Real(float) :: x(size(b)),Ax(size(A,1),size(A,2))
      Ax = A;
      Call LuFactrLDL(Ax)
      x = LUSubstLDL(Ax,b)
   End Function LUSolveLDL
   ! ------------------------------------------------------------------
   Subroutine LUFactrLDL(A)      ! Cholesky factor into LDLt
      Real(float) :: A(:,:)
      Real(float) :: w(size(A,1)),t
      Integer :: n,i,j
      ! Factor symmetric matrix into LDLt, 1/D on diagonal
      ! Only reference to upper triangle is copy below.
      ! remove copy to develop compact storage versiou
      ! -----------------------------------------
      n = size(A,1)
      w(1)=1.0/A(1,1)
      do i=2,n
         do j=2,i-1
            A(i,j)=A(i,j) - sum(A(i,1:j-1)*A(j,1:j-1))
         enddo
         do j=1,i-1
            t = A(i,j);  A(i,j) = A(i,j)*w(j)
            A(i,i) = A(i,i) - t*A(i,j)
            A(j,i) = A(i,j)   ! only upper
         enddo
         w(i)=1.0/A(i,i)
      end do
      do i=1,n
         A(i,i)=w(i)
      end do
      if(DumpF > 1)write(ilog,*)'LUFactrLDL - LDLt Cholesky algorithm'
      if(DumpF > 3)Call LUTest(A,Option=4)
   End Subroutine LUFactrLDL
   ! ------------------------------------------------------------------
   Function LUSubstLDL(A,b)  Result(x)
      real(float), intent(in) :: A(:,:), b(:)
      real(float) :: x(size(b))
      integer :: n,i
      ! Forward & back solve for LDLt
      ! -----------------------------------------
      x = b;   n = size(b)
      do i=2,n
         x(i) = x(i) - sum(A(i,1:i-1)*x(1:i-1))
      end do
      do i=1,n
         x(i) = x(i)*A(i,i)
      end do
      do i=n-1,1,-1
         x(i) = x(i) - sum(A(i+1:n,i)*x(i+1:n))
      end do
      if(DumpF > 1)write(ilog,*)'LUSubstLDL - LDLt Cholesky algorithm'
!      call ArrayPrint("LUSubstLDL:",A,b,x) !,flog=.true.)
   End Function LUSubstLDL
! ------------------------------------------------------------------
   Function LUSolveSq(A,b)  Result(x)
      Real(float) :: A(:,:),b(:),x(size(b)),Ax(size(A,1),size(A,2))
      Ax = A;
      Call LuFactrSq(Ax)
      x = LUSubstSq(Ax,b)
   End Function LUSolveSq
! ------------------------------------------------------------------
   Subroutine LUFactrSq(A)
      Real(float) :: A(:,:)
      Integer :: n,i,j
      ! Factorization for Symmetric Positive Definite Matrix
      ! Using Square Root Method
      n = size(A,1)
      do i=1,n
         A(i,i) = Sqrt(A(i,i) - Sum(A(1:i-1,i)**2))
         do j=i+1,n
            A(i,j) = (A(i,j) - Sum(A(1:i-1,i)*A(1:i-1,j)))/A(i,i)
            A(j,i) = A(i,j)
         enddo
      enddo
      if(DumpF > 1)write(ilog,*)'LUFactrSq - square root method'
      if(DumpF > 3)Call LUTest(A,Option=0)
   End Subroutine LUFactrSq
! ------------------------------------------------------------------
   Function LUSubstSq(A,b)  Result(x)
      Real(float) :: A(:,:),b(:)
      Real(float) :: x(size(b))
      Integer :: n,i
      ! Solution using factors from LUFactrS and rhs b
      n = size(b);  x = b
      do i=1,n
         x(i) = (x(i) - Sum(A(1:i-1,i)*x(1:i-1)))/A(i,i)
      enddo
      do i=n,1,-1
         x(i) = (x(i) - Sum(A(i+1:n,i)*x(i+1:n)))/A(i,i)
      enddo
      if(DumpF > 1)write(ilog,*)'LUSubstSq - square root method'
   End Function LUSubstSq
   ! ------------------------------------------------------------------
   Function LUSolve(A,b,ipivot)  Result(x)
      Real(float), intent(in) :: A(:,:),b(:)
      Integer, Optional :: ipivot(size(b))
      Real(float) :: x(size(b)),Ax(size(A,1),size(A,2))
      Integer :: n,info,ipvt(size(b))
      Integer, parameter :: nrhs = 1
      ! Solve Ax = b using DGETRF & DGETRS (LAPack)
      ! Perform solution using factors & rhs b using dgesl
      n = size(b);  x = b;  Ax = A
      call DGETRF(n,n,Ax,n,ipvt,info)  ! factor matrix
      if(info == 0)then
         if(Present(ipivot))ipivot = ipvt
         call DGETRS('N',n,1,Ax,n,ipvt,x,n,info)   ! solve
      endif
      if(info /= 0 .and. DumpF > 2)write(ilog,*)'LUSolve: DGETRF/DGETRS info not 0'
      if(DumpF > 1)write(ilog,*)'LUSolve: DGETRF/DGETRS info =',info
   End Function LUSolve
   ! ------------------------------------------------------------------
   Function LUSubst(A,b,ipivot)  Result(x)
      Real(float) :: A(:,:),b(:),x(size(b))
      Integer :: n, info, ipivot(size(b))
      ! Solve using supplied factors & rhs b using DGETRS
      n = size(b);  x = b
      call DGETRS('N',n,1,A,n,ipivot,x,n,info)   ! solve
   End Function LUSubst
   ! ------------------------------------------------------------------
   Subroutine LUFactr(A,ipivot)
      Real(float) :: A(:,:)
      Integer :: n,info,ipivot(size(A,1))
      ! Perform LU Factorization using DGETRF
      n = size(A,1)
      call DGETRF(n,n,A,n,ipivot,info)   ! factor matrix
      if(info /= 0 .and. DumpF > 2)write(ilog,*)'LUFactr: DGETRF info not 0'
   End Subroutine LUFactr
   ! ------------------------------------------------------------------
   Subroutine LUInvert(A)
      Real(float) :: A(:,:)
      Real(float) :: Work(size(A))
      Integer :: n,lwork,info,ipivot(size(A,1))
      ! Invert A using DGETRF and DGETRI
      n = size(A,1);  lwork = size(A)
      call DGETRF(n,n,A,n,ipivot,info)   ! factor matrix
      if(info == 0)then
         call DGETRI(n,A,n,ipivot,Work,lwork,info) ! invert
      endif
      if(info /= 0 .and. DumpF > 2) &
         write(ilog,*)'LUInvert: DGETRF/DGETRI info not 0'
   End Subroutine LUInvert
   ! ------------------------------------------------------------------
   Function LUSolveSB(A,b)  Result(x)
      Real(float), intent(in) :: A(:,:),b(:)
      Real(float) :: x(size(b)),Ab(size(A,1),size(A,2))
      Integer :: n,kd,info,nd(2)
      Integer, parameter :: nrhs=1
      Character, parameter :: uplo = 'U'
      ! Solve symmetric band Ax = b using DPBTRF & DPBTRS (LAPack)
      nd = shape(A);  n = nd(2);  kd = nd(1)-1; x = b;  Ab = A
      call DPBTRF(uplo,n,kd,Ab,kd+1,info) ! factor
      if(info == 0)then
         call DPBTRS(uplo,n,kd,nrhs,Ab,kd+1,x,n,info)
      endif
      if(info /= 0 .and. DumpF > 2)write(ilog,*)'LUSolve: DPBTRF/DPBTRS info =',info
   End Function LUSolveSB
   ! ------------------------------------------------------------------
   Function LUSubstSB(Ab,b)  Result(x)
      Real(float), intent(in) :: Ab(:,:),b(:)
      Real(float) :: x(size(b))
      Integer :: nd(2),n,kd,info
      Integer, parameter :: nrhs=1
      Character, parameter :: uplo = 'U'
      ! Back substitution for factored symmetric band using LAPack
      nd = shape(Ab);  n = nd(2);  kd = nd(1)-1;  x = b
      call DPBTRS(uplo,n,kd,nrhs,Ab,kd+1,x,n,info)
      if(info /= 0 .and. DumpF > 1)write(ilog,*)'LUSubstSB: DPBTRS info =',info
   End Function LUSubstSB
   ! ------------------------------------------------------------------
   Subroutine LUFactrSB(Ab)
      Real(float) :: Ab(:,:)
      Integer :: n,kd,info,nd(2)
      Character, parameter :: uplo = 'U'
      ! Factor symmetric band matrix A DPBTRF (LAPack)
      nd = shape(Ab); n = nd(2); kd = nd(1)-1
      call DPBTRF(uplo,n,kd,Ab,kd+1,info) ! factor
      if(info /= 0 .and. DumpF > 1)write(ilog,*)'LUFactrSB: DPBTRF info =',info
   End Subroutine LUFactrSB
   ! ------------------------------------------------------------------
   Subroutine LUInvertSB(A)
      Real(float) :: A(:,:)
      Integer :: n,kd,info,nd(2)
      Integer, parameter :: nrhs=1
      Character, parameter :: uplo = 'U'
      ! Invert Symmetric Band matrix LAPack
      nd = shape(A); n = nd(2); kd = nd(1)-1
      call DPBTRF(uplo,n,kd,A,kd+1,info) ! factor
      if(info == 0)then
! is there a routine for this?  *** Warning this routine not finished ****
!         call DPBTRI(uplo,n,kd,nrhs,Ab,kd+1,x,n,info)
      write(*,*)'Stop: could not find an band inversion routine in LAPack'
      stop
      endif
      if(info /= 0 .and. DumpF > 2)write(ilog,*)'LUSolve: DPBTRF/DPBTRS info not 0'
      if(DumpF > 1)write(ilog,*)'LUSolveSB: DGETRF/DGETRS info =',info
 !     Call ArrayPrint('LUSolveSB Ab:',Ab)
      if(info /= 0 .and. DumpF > 2) &
         write(ilog,*)'LUInvert: DGETRF/DGETRI info not 0'
   End Subroutine LUInvertSB
   ! ------------------------------------------------------------------
   Function LUSolveSym(A,b)  Result(x)
      Real(float), intent(in) :: A(:,:),b(:)
      Real(float) :: x(size(b)),Ax(size(A,1),size(A,2))
      Integer :: n,info
      Integer, parameter :: nrhs = 1
      ! Perform LU Factorization using dgefa
      ! Perform solution using factors & rhs b using dgesl
      n = size(b);  x = b;  Ax = A
      call DPOTRF('L',n,Ax,n,info)  ! factor matrix
      if(info == 0)then
         call DPOTRS('L',n,nrhs,Ax,n,x,n,info)  ! solve
      endif
      if(info /= 0 .and. DumpF > 2)write(ilog,*) &
         'LUSolveSym: DPOTRF/DPOTRS info =',info
      if(DumpF > 1)write(ilog,*)'LUSolveSym: DPOTRF/DPOTRS info =',info
   End Function LUSolveSym
   ! ------------------------------------------------------------------
   Function LUSubstSym(A,b)  Result(x)
      Real(float) :: A(:,:),b(:),x(size(b))
      Integer :: n, info
      Integer, parameter :: nrhs = 1
      ! Solve using supplied factors & rhs b using DGETRS
      n = size(b);  x = b
      call DPOTRS('L',n,nrhs,A,n,x,n,info)  ! solve
   End Function LUSubstSym
   ! ------------------------------------------------------------------
   Subroutine LUFactrSym(A)
      Real(float) :: A(:,:)
      Integer :: n,info
      ! Perform LU Factorization using DPOTRF
      n = size(A,1)
      call DPOTRF('L',n,A,n,info)  ! factor matrix
      if(info /= 0 .and. DumpF > 2)write(ilog,*)'LUFactrSym: DPOTRF info not 0'
   End Subroutine LUFactrSym
   ! ------------------------------------------------------------------
   Subroutine LUInvertSym(A)
      Real(float) :: A(:,:)
      Integer :: n,info,i
      n = size(A,1)
      call DPOTRF('L',n,A,n,info)  ! factor matrix
      if(info == 0)then
         call DPOTRI('L',n,A,n,info)   ! invert
         do i=1,n
            A(i,i+1:n) = A(i+1:n,i)
         enddo
      endif
      if(info /= 0 .and. DumpF > 2) &
         write(ilog,*)'LUInvert: DGETRF or DGETRI info not 0'
   End Subroutine LUInvertSym
   ! ------------------------------------------------------------------
   Subroutine LUTest(A,Option)
      Real(float) :: A(:,:)
      Real(float) :: al(size(A,1),size(A,2)),au(size(A,1),size(A,2))
      Real(float) :: B(size(A,1),size(A,2)),d(size(A,1),size(A,2))
      integer :: n, i, j, Option
      ! -------------------------------------------
      ! option: -2 unit lower, 1/U(k,k) on diagonal
      !         -1 unit lower
      !          0  LLt
      !         +1 unit upper
      !         +2 unit upper, 1/L(k,k) on diagonal
      !          3 LDLt, D on diagonal
      !          4 LDLt, 1/D on diagonal
      ! option +1 for unit upper
      ! -------------------------------------------
      n = size(A,1)
      al = 0.0_float;  au = 0.0_float
      do i=1,n
         do j=i,n
            au(i,j) = a(i,j)
            al(j,i) = a(j,i)
         enddo
      enddo
      do i=1,n
         Select Case(Option)
         Case(-2,-1)
            al(i,i) = 1.0_float  ! unit lower
            if(Option == -2)au(i,i) = (1.0_float)/a(i,i)
         Case(1,2)
            au(i,i) = 1.0_float  ! unit upper
            if(Option == 2)al(i,i) = (1.0_float)/a(i,i)
         Case(3)
            au(i,i) = 1.0_float  ! unit upper
            al(i,i) = 1.0_float  ! unit lower
            d(i,i) = a(i,i)
         Case(4)
            au(i,i) = 1.0_float  ! unit upper
            al(i,i) = 1.0_float  ! unit lower
            d(i,i) = (1.0_float)/a(i,i)
         Case Default
         End Select
      enddo
      if(Option < 3)then
         B = MatMul(al,au)
      else
         B = MatMul(d,au)
         B = MatMul(al,B)
      endif
      call ArrayPrint("LUTest Factored:",A) !,flog=.true.)
      call ArrayPrint("LUTest Multiplied:",B) !,flog=.true.)
   End Subroutine LUTest
   ! ------------------------------------------------------------------
End Module LUSolvers
