! ---------------------------------------------------
Module OrthogonalColloc
! ---------------------------------------------------
   Use Array_Print, only : ArrayPrint, VectorPrint
   Use Orth_Poly, only : Jacobi_Quadrature,Polyset,Nasymptotic
   Use Orth_Cheb, only : Cheb_Quad
! This Module supports Gauss, Lobatto, Chebyshev & Radau base points
! Nonsymmetric problems in Planar geometry
! Symmetric problems in Planar, Cylindrical or Spherical geometry
!
! There are no hardwired dimensions, so any number of points
! may be specified.
!
! CollocSet must be called first to specify
!     Symmetry, Type (Lobatto, etc.) and Geometry
!     used to set the Jacobi weight factors:
!        ((1 - x)**alpha)*(x**beta)
!
! The arrays may be obtained individually using the array valued
! funcions: Xpoints, Weights, Deriv1, Deriv2, Stiffness or MassMat
! Or CollocCoef can be called to get X, W, A and optionally
! B and C all at once.
!
! The Jacobi & QWeights codes are general for other Jacobi
! polynomials weights.
!
! The precision can be changed by changing parameter "float" in defs.fi
! Intermediate calculation output is controlled by the value "Debug" in defs.fi
! ---------------------------------------------------
   include 'defs.fi'
   Private
   Public Initialize
   Public DataDelete ! deallocate data
   Public CollocCoef ! call following 5 functions to get all the arrays
   Public Xpoints    ! returns quadrature points
   Public WeightQ    ! returns quadrature weights
   Public WeightB    ! returns the barycentric weights
   Public Deriv1     ! returns first derivative matrix
   Public Deriv2     ! returns second derivative matrix
   Public Stiffness  ! returns W*At*A for Lobatto -wB for Gauss
   Public MassMatrix ! Galerkin or Moments mass matrix (overloaded)
   Public DerivOdd    ! matrix for first derivative of odd function
   Public LCoeff     ! Lagrange polynomials (overloaded), 1 or n points
   Public Lpoly      ! interpolating coefficients as monomials
   Public Interp     ! 1 time interpolation (overloaded)
   Public CollocSym  ! returns value of Sym
   ! public parameters  ---------------------------------------------
   Integer, Parameter, Public :: General=0,Gauss=1,Lobatto=2,Chebyshev=3,RadauR=4,RadauL=5,Chebyshev1=6
   Integer, parameter, public :: Nonsymmetric=0,Planar=1,Cylindrical=2,Spherical=3,Symmetric = 1
   ! public text strings --------------------------------------------
   Character (len=*), Public, Parameter :: TxtType(0:6) = &
      (/'Generic  ','Gauss    ','Lobatto  ','Chebyshev','Radau Rgt','Radau Lft','Cheby 1  '/)
   Character (len=*), Public, Parameter :: TxtGeom(0:3) = (/'No symmetry','Planar     ','Cylindrical','Spherical  '/)
   Character (len=*), Public, Parameter :: TxtSym(0:1) = (/'Nonsymmetric','Symmetric   '/)
   ! private module data --------------------------------------------
   Integer :: Sym = NonSymmetric    ! default value Nonsymmetric (0)
   Integer :: Typ = Lobatto         ! default value Lobatto (2)
   Integer :: Geom = Nonsymmetric   ! default value Nonsymmetric (0)
   logical :: shft = .true.         ! default shifted, i.e. (0,1)
   Real(float), parameter :: Debug = 0
   Real(floatq), Parameter :: piq = 3.14159265358979323846264338327950288419716939937510_floatq
   Real(float), public, parameter :: pi = real(piq,float)
   ! structure to hold basic data ---------------------------------
   Type, Public :: ColDataType
      Private     ! Private type with private data
      integer :: n,nt
      Integer :: sym = NonSymmetric    ! default value Nonsymmetric (0)
      Integer :: typ = Lobatto         ! default value Lobatto (2)
      Integer :: geo = Nonsymmetric    ! default geometry none
      logical :: sft = .true.
      integer :: nneg = 0
      real(float) :: fwb
      Real(float), allocatable :: x(:)    ! points (roots) (new)
      Real(float), allocatable :: th(:)   ! theta points (roots) (new)
      Real(float), allocatable :: wb(:)   ! barycentric weight from product
      Real(float), allocatable :: w(:)    ! quadrature weight  (new)
      Real(float), allocatable :: dx(:,:) ! x(i) - x(j) array
      Type(ColDataType), pointer :: next => null()
   End Type ColDataType
   ! overloaded functions -------------------------------------------
   Interface Initialize
      Module Procedure InitializeN, InitializeX
   End Interface
   Interface Lcoeff
      Module Procedure LCoef1, LCoefN, LCoef1x, LCoefNx
   End Interface
   Interface Interp     ! overload to accept x as scalar or array
      Module Procedure Interp1, InterpN
   End Interface
   Interface MassMatrix ! overload MassMatrix(x,[nx,func]) optional function
      Module Procedure MassMat00, MassMatFn, MassMatFx
   End Interface
   Character (len=*), Parameter :: fmf17 = '(25(f20.16,a1))'
   Character (len=*), Parameter :: fme17 = '(25(es24.17,a1))'
!-------------------------------------------------------------------------------
   Contains
!-------------------------------------------------------------------------------
   Subroutine InitializeN(d,n,TType,Geometry,Shift)
      Type(ColDataType) :: d
      Integer :: n
      Integer, optional :: TType     ! default: Lobatto=2, optional: 1 to 5
      Integer, optional :: Geometry  ! default: none=0 optional: Planar=1, Cylindrical=2, Spherical=3
      logical, optional :: Shift     ! default: .true. shifted (symmetric always shifted)
      Integer :: nt
      !---------------------------------------
      ! set up one of the built in types Gauss, Lobatto, Chebyshev, RadauR, RadauL
      !---------------------------------------
      call DataDelete(d)
      Call CollocSet(TType,Geometry,Shift=Shift)
      nt = n + 2 - sym
      if(Debug > 0)then
         write(*,'(a,i5,1x,a)')'InitializeN:',n,txttype(typ)
         write(iout,'(a,i6,3i3,L2)')'Initialize: n,t,g,s,shft',n,typ,Geom,Sym,shft
         write(*,'(a,i5,1x,a)')'InitializeN: allocate',nt*(4+nt)
      endif
      Allocate(d%x(nt),d%th(nt),d%wb(nt),d%w(nt),d%dx(nt,nt)) ! x,theta,wb,w,dx
      d%n = n;       d%nt = nt;     d%sft = shft
      d%sym = Sym;   d%typ = typ;   d%geo = Geom
      call CalcData(n,d%x,d%th,d%w,d%wb,d%fwb,d%dx,d%nneg)
   End Subroutine InitializeN
!-------------------------------------------------------------------------------
   Subroutine InitializeX(d,x,Geometry,Shift)
      Type(ColDataType) :: d
      real(float) :: x(:)
      Integer, optional :: Geometry  ! default: none=0 optional: Planar=1, Cylindrical=2, Spherical=3
      logical, optional :: Shift     ! default: .true. shifted (symmetric always shifted)
      Integer :: n,nt
      !---------------------------------------
      ! set up a generic type with provided points
      !---------------------------------------
      call DataDelete(d)
      Call CollocSet(General,Geometry,Shift=Shift)
      nt = size(x)
      n = nt - 2 + sym
      if(Debug > 0)then
         write(*,'(a,i5,1x,a)')'InitializeX:',n,txttype(typ)
         write(iout,'(a,i6,3i3,L2)')'InitializeX: n,t,g,s,shft',n,typ,Geom,Sym,shft
         write(*,'(a,i5,1x,a)')'InitializeX: allocate',nt*(4+nt)
      endif
      Allocate(d%x(nt),d%th(nt),d%wb(nt),d%w(nt),d%dx(nt,nt)) ! x,theta,wb,w,dx
      d%x = x
      d%n = n;       d%nt = nt;     d%sft = shft
      d%sym = Sym;   d%typ = typ;   d%geo = Geom
      call CalcData(n,d%x,d%th,d%w,d%wb,d%fwb,d%dx,d%nneg)
   End Subroutine InitializeX
!-------------------------------------------------------------------------------
   Subroutine CollocSet(TType,Geometry,Symmetry,Shift)
      Integer, optional :: TType     ! default: Lobatto=2, optional: 1 to 5
      Integer, optional :: Geometry  ! default: Planar=0, optional: Cylindrical=1, Spherical=2
      Logical, optional :: Symmetry  ! default: .false => Symmetry =NonSymmetric=0
      logical, optional :: Shift     ! default: .true. shifted (symmetric always shifted)
      !---------------------------------------
      if(present(Geometry))then
         geom = max(Nonsymmetric,min(Geometry,Spherical))
         sym = max(Nonsymmetric,min(geom,Symmetric))
      elseif(present(Symmetry))then
         geom = Nonsymmetric
         if(Symmetry)geom = Planar
      endif
      if(present(TType))then
         typ = Max(General,Min(RadauL,TType))
         if(sym == Symmetric)typ = Min(typ,Chebyshev)
      endif
      if(present(Shift))shft = Shift
      if(sym == Symmetric)shft = .true.
      if(Debug > 0)then
         write(*,'(a,3i3,5L2)')'CollocSet: 1 typ,sym,geom,shift =',typ,sym,geom,shft
         write(iout,'(a,3i3,5L2)')'CollocSet: 2 typ,sym,geom,shift =',typ,sym,geom,shft, &
            present(TType),present(Geometry),present(Symmetry),present(Shift)
      endif
   End Subroutine CollocSet
!-------------------------------------------------------------------------------
   Subroutine DataDelete(d)
      Type(ColDataType) :: d
      ! Delete data from structure
      if(allocated(d%x))then
         if(Debug > 0)then
            Write(*,'(a,i5,1x,a)')'DataDelete: ',d%n,txttype(d%typ)
            Write(iout,'(a,i5,1x,a)')'DataDelete: ',d%n,txttype(d%typ)
            if(Debug > 1)then
               Write(*,'(a,L2,i5)')'DataDelete: x  =',allocated(d%x),size(d%x)
               Write(*,'(a,L2,i5)')'DataDelete: th =',allocated(d%th),size(d%th)
               Write(*,'(a,L2,i5)')'DataDelete: w  =',allocated(d%w),size(d%w)
               Write(*,'(a,L2,i5)')'DataDelete: wb =',allocated(d%wb),size(d%wb)
               Write(*,'(a,L2,4i5)')'DataDelete: dx =',allocated(d%dx),size(d%dx),shape(d%dx)
            endif
         endif
         Deallocate(d%x,d%th,d%wb,d%w,d%dx)
      endif
   End Subroutine DataDelete
!-------------------------------------------------------------------------------
   Subroutine CalcData(n,x,th,w,wb,fwb,dx,nneg)
      integer, intent(in) :: n
      Real(float), intent(out) :: x(:),th(:),w(:),wb(:),fwb,dx(:,:)
      integer, intent(out) :: nneg
      Real(float) :: xw(sym:n+1,4)
      Real(float) :: ab(2)
      logical :: usex
      ! ----------------------------------
      ! Gets and stores basic data: roots & weights
      ! ----------------------------------
      usex = .true.
      select case(typ)
      case(Gauss,Lobatto,RadauR,RadauL)
         ab = Polyset(typ,geom)
         xw = Jacobi_Quadrature(n,ab,geom,fwb,nneg)
         call Xfix(xw,fwb,nneg)
         usex = n <= Nasymptotic  ! x(i) - x(j) from theta, large n
      case(Chebyshev,Chebyshev1)
         ab = Polyset(typ,geom)
         xw = Cheb_Quad(n,geom,sym,fwb,Shift=shft)
!        if(geom > Planar)xw(:,4) = QW_general(xw(:,1)) ! less accurate
         if(geom > Planar)xw(:,4) = Quad_general(xw(:,1),xw(:,3)*fwb,sym)
      case default   ! generic
         xw(:,1) = x
         xw(:,2) = acos(x)
         if(sym == Symmetric)x = x*x
         xw(:,3) = WBcalc(x,fwb)
         xw(:,4) = Quad_general(xw(:,1),xw(:,3)*fwb,sym)
      end select
      x  = xw(:,1);  th = xw(:,2)
      wb = xw(:,3);  w  = xw(:,4)
      dx = DelX(x,th,nneg,shft,usex)
      if(Debug > 0)then
         write(iout,'(a,2i6,3i2,L2,2f5.1,g15.8)')'Initialize: n,nneg,t,g,s,shift', &
            n,nneg,typ,geom,sym,shft,ab,fwb
         Call Arrayprint('CalcData: x,t,wb,wq:',xw,fmtx=fmf17)
         Call Arrayprint('CalcData: x(i) - x(j):',dx,fmtx=fmf17)
      endif
   End Subroutine CalcData
   !----------------------------------------------------------------------------
   Subroutine Xfix(xw,fwb,nneg)
      real(float), intent(inout) :: xw(sym:,:),fwb
      integer, intent(inout) :: nneg
      integer :: n
      ! fix negative values of x
      ! modifications for shifted case
      select case(geom)
      case(:0)
         n = size(xw,1) - 2
         xw(:nneg-1,1) = -xw(:nneg-1,1)
         if(shft)then   ! convert to shifted
            !write(iout,'(a,2g14.7,3i7)')'Wb-factor:',fwb,fwb*2**n,n,2**n,sym
            fwb = fwb*2**n
            xw(:,1) = (xw(:,1)+1.0_float)*0.5_float
            xw(:,3) = xw(:,3)*2.0_float
            xw(:,4) = xw(:,4)*0.5_float
         endif
      end select
   End Subroutine Xfix
!-------------------------------------------------------------------------------
   Function DelX(x,t,nneg,shft,usex)  result(dx)
      Real(float), intent(in) :: t(:),x(:)
      integer, intent(in) :: nneg
      logical, intent(in) :: shft,usex
      Real(float) :: dx(size(t),size(t))
      Real(float) :: tx(size(t)),factr
      integer :: n,i
      ! -------------------------------------------
      !  x - points/roots
      !  t = angles (x = cos(t) normal, x^2 = [cos(t)-1]/2  symmetric
      !  x(i) - x(j) calculated & returned
      !  nneg - no. negative roots, t = arccos(-x) = pi - arccos(x)
      !  shft - true if points are shifted
      !  usex - true, use x(i), false us t = arccos(x)
      ! -------------------------------------------
      n = size(t)
      factr = -2.0_float
      if(shft)factr = -1.0_float
      !write(iout,'(a,2L2,i5)')'CalcData/Delx: Use x?, shift, nneg =',usex,shft,nneg
      if(usex)then   ! calculation using x
         if(Sym == Symmetric)then
            tx = x*x
         else
            tx = x
         endif
         do i=1,n
            dx(i,:) = tx(i) - tx
            dx(i,i) = 1.0_float
         enddo
      else     ! calculation using theta
         tx(:nneg) = pi - t(:nneg)
         tx(nneg+1:) = t(nneg+1:)
         do i=1,n
            dx(i,:) = sin((tx(i)+tx(:))*0.5_float)*sin((tx(i)-tx(:))*0.5_float)*factr
            dx(i,i) = 1.0_float
         enddo
      endif
   End Function DelX
!-------------------------------------------------------------------------------
   Integer Function CollocSym()
      if(Debug > 0)write(iout,'(a,3i3,L2)')'CollocSym: typ,sym,geom,shift =',typ,sym,geom,shft
      CollocSym = sym
   End Function CollocSym
!-------------------------------------------------------------------------------
   Subroutine CollocCoef(d,x,w,A,B,C)
      Type(ColDataType), intent(in) :: d
      Real(float),intent(out) :: A(:,:),w(:),x(:)
      Real(float), optional :: B(:,:),C(:,:)
      ! Coefficients run 0:n+1 or 1:n+2 if nonsymmetric, 1:n+1 if symmetric
      x = d%x
      w = d%w
      A = Deriv1(d)
      if(present(B))B = Deriv2(d)
      if(present(C))C = Stiffness(d)
   End Subroutine CollocCoef
!-------------------------------------------------------------------------------
   Function Xpoints(d) result(x)
      Type(ColDataType), intent(in) :: d
      Real(float) :: x(d%nt)
      x = d%x  ! return point value
   End Function Xpoints
!-------------------------------------------------------------------------------
   Function WeightQ(d) result(w)
      Type(ColDataType), intent(in) :: d
      Real(float) :: w(d%nt)
      w = d%w  ! return Quadrature Weights
   End Function WeightQ
!-------------------------------------------------------------------------------
   Function WeightB(d,fscale) result(w)
      Type(ColDataType), intent(in) :: d
      Real(float) :: w(d%nt),fscale
      w = d%wb    ! return barycentric weights
      fscale = d%fwb
   End Function WeightB
!-------------------------------------------------------------------------------
   Function Deriv1(d)  result(A)
      Type(ColDataType), intent(in) :: d
      Real(float) :: A(d%nt,d%nt)
!      real(float) :: da(d%nt,d%nt),px(d%nt),x(d%nt)
      integer :: i
      ! First Derivative matrix
      if(Debug > 0)write(*,'(a,3i3)')' Acoef:',d%n
!      if(Sym == Symmetric)then
!         x = (d%x)*(d%x)
!      else
!         x = d%x
!      endif
      do i=1,d%nt
!         px = x(i) - x;  px(i) = 1.0_float
!         da(i,:) = px
!         A(i,:) = d%wb/(px*d%wb(i))
         A(i,:) = d%wb/(d%wb(i)*d%dx(i,:))
         A(i,i) = 1.0_float - Sum(A(i,:))
      enddo
      !call arrayprint('x(i) - x(j) original',da)
      !call arrayprint('x(i) - x(j) new',d%dx)
      if(Sym == Symmetric)then
         do i =1,d%nt
            A(i,:) = 2.0_float*d%x(i)*A(i,:)
         enddo
      elseif(d%typ == Lobatto)then
         do i =2,d%n+1  ! diagonals zero except A(0,0) & A(n+1,n+1)
            A(i,i) = 0.0_float
         enddo
      endif
   End Function Deriv1
!-------------------------------------------------------------------------------
   Function DerivOdd(d) result(An)
      Type(ColDataType), intent(in) :: d
      Real(float) :: An(d%nt,d%nt)
      Real(float) :: gx, A(d%nt,d%nt)
      Integer :: i
      ! This function gives matrix to approximate derivative of
      ! an odd function, e.g. the derivative of an even function
      ! Note:  An*A = B,  problem is symmetric
      if(Debug > 0)write(*,'(a,3i3)')' DerivOdd:', d%n,d%geo
      A = Deriv1(d)
      if(Sym == NonSymmetric)then
         An = A  ! in case of call for nonsymmetric
      else
         gx = Real(d%geo,float)
         do i=1,d%nt
            An(i,:) = d%x(i)*A(i,:)/d%x(:)
            An(i,i) = An(i,i) + gx/d%x(i)
         enddo
      endif
   End Function DerivOdd
!-------------------------------------------------------------------------------
   Function Deriv2(d)  result(B)
      Type(ColDataType), intent(in) :: d
      Real(float) :: B(d%nt,d%nt)
      ! Laplacian Derivative Matrix
      if(Debug > 0)write(*,'(a,i5,L2)')' Bcoef:',d%n  !,present(a)  ! < causes failure?
      if(Sym == Symmetric)then
         Call Deriv2S(d%x,d%dx,Deriv1(d),B,d%geo)
      else
         Call Deriv2N(d%x,d%dx,Deriv1(d),B)
      endif
   End Function Deriv2
!-------------------------------------------------------------------------------
   Subroutine Deriv2N(x,dx,A,B)
      Real(float), intent(in) :: x(:),dx(:,:),A(:,:)
      Real(float), intent(out) :: B(:,:)
      Integer :: i
      ! Laplacian derivative, nonsymmetric
      do i = 1,size(x)
         B(i,:) = 2.0_float*A(i,:)*(A(i,i) - 1.0_float/dx(i,:))
         B(i,i) = 0.0_float;   B(i,i) = -sum(B(i,:))
      enddo
   End Subroutine Deriv2N
!-------------------------------------------------------------------------------
   Subroutine Deriv2S(x,dx,A,B,g)
      Real(float), intent(in) :: x(:),dx(:,:),A(:,:)
      integer, intent(in) :: g
      Real(float), intent(out) :: B(:,:)
      Real(float) :: gx
      Integer :: i
      ! Laplacian derivative, symmetric
      gx = 0.5_float*Real(g,float)
      do i = 1,size(x)
         B(i,:) = 2.0_float*A(i,:)*(A(i,i) + gx/x(i) - 2.0_float*x(i)/dx(i,:))
         B(i,i) = 0.0_float;   B(i,i) = -sum(B(i,:))
      enddo
   End Subroutine Deriv2S
!-------------------------------------------------------------------------------
   Function Stiffness(d) result(C)
      Type(ColDataType), intent(in) :: d
      Real(float) :: A(d%nt,d%nt),C(d%nt,d%nt)
      integer :: i,nt
      ! Stiffness Matrix
      if(Debug > 0)write(*,'(a,3i3)')' Ccoef:',d%n
      nt = d%nt
      A = Deriv1(d)
      if(Sym == Symmetric)then
         Call Deriv2S(d%x,d%dx,A,C,d%geo)
      else
         Call Deriv2N(d%x,d%dx,A,C)
      endif
      do i=1,nt
         C(i,:) = -d%w(i)*C(i,:)
      enddo
      C(nt,:) = C(nt,:) + A(nt,:)
      if(Sym == NonSymmetric)then
         C(1,:) = C(1,:) - A(1,:)
      endif
   End Function Stiffness
!-------------------------------------------------------------------------------
   Function MassMat00(d,fx) result(Dm)        ! version of MassMat with one vector argument
      Type(ColDataType), intent(in) :: d
      Real(float) :: Dm(d%nt,d%nt)
      Real(float), optional :: fx(:)
      integer :: i,icase
      ! -------------------------------------------------------
      !  This is 1 of 2 overloaded MassMat functions. External calls
      !  normally have just one argument and assumes no function, i.e. fx = 1.
      !  The optional fx argument is primarily in case the other MassMat option
      !  (MassMatFn) is called with Chebyshev points
      ! -------------------------------------------------------
      if(Debug > 0)write(*,'(a,3i3)')' MassMat00:',d%nt
      icase = typ + 10*sym
      select case(icase)
      case(Gauss)
         Call MassMatG(d,Dm)     ! for Gauss nonsymmetric
      case(Lobatto)
         Call MassMatL(d,Dm)     ! for Lobatto nonsymmetric
      case default
         Dm = 0.0_float
         do i=1,d%nt
            Dm(i,i) = d%w(i)
         enddo
      end select
      if(present(fx))then  ! interpolated coefficient
         do i=1,d%nt
            Dm(i,:) = Dm(i,:)*fx
         enddo
      endif
   End Function MassMat00
!-------------------------------------------------------------------------------
   Subroutine MassMatG(d,Dm)
      Type(ColDataType), intent(in) :: d
      Real(float), intent(out) :: Dm(0:,0:)
      Real(float) :: Cn,Cd,wx(0:d%n+1)
      integer :: i,n
      ! -------------------------------------------------------
      !  Calculate moments/Gauss mass matrix from formula, f(x) = 1
      ! -------------------------------------------------------
      n = d%n
      if(Debug > 0)write(*,'(a,3i3)')' MassMatG:',n
      Cn = (2.0_float)/real(2*n+1,float)
      if(shft)Cn = Cn*0.5_float
      Cd = real(-2*n,float)
      wx = Cn*(d%dx(1,:))*(d%dx(d%nt,:))*(d%wb)   ! Cn*(1-x^2)*Wb
      Dm(0,:) = 0.0_float; Dm(n+1,:) = 0.0_float
      do i = 1,n
         Dm(i,:) = wx(i)*d%wb
         Dm(i,i) = Cd*Dm(i,i)
      enddo
   End Subroutine MassMatG
!-------------------------------------------------------------------------------
   Subroutine MassMatL(d,Dm)
      Type(ColDataType), intent(in) :: d
      Real(float), intent(out) :: Dm(0:,0:)
      Real(float) :: Cn,Cd
      integer :: i,n
      ! -------------------------------------------------------
      !  Calculate Galerkin/Lobatto mass matrix from formula, f(x) = 1
      ! -------------------------------------------------------
      n = d%n
      Cn = real(-8*(n+1),float)/real((2*n+3)*(n+2),float)
      if(shft)Cn = Cn*0.125_float
      Cd = real(-2*(n+1),float)
      do i = 0,n+1
         Dm(i,:) = Cn*d%wb(i+1)*d%wb
         Dm(i,i) = Cd*Dm(i,i)
      enddo
      if(Debug > 0)write(*,'(a,3i3)')' MassMatL:',d%n,Cn,Cd
   End Subroutine MassMatL
!-------------------------------------------------------------------------------
   Function MassMatFx(d,func,nqx) result(Dm)   ! called when func supplied
      Type(ColDataType), intent(in) :: d
      integer, optional :: nqx
      Real(float) :: Dm(d%nt,d%nt)
      integer :: nx,nq
      Type(ColDataType) :: dx
      Interface
         Function func(x)  result(f)
            include 'defs.fi'
            Real(float), intent(in) :: x(:)
            Real(float) :: f(size(x))
         End Function func
      End Interface
      ! -------------------------------------------------------
      nx = 1
      if(Present(nqx))nx = nqx
      nq = d%n + nx
      call Initialize(dx,nq,d%typ,d%geo,d%sft)
      Dm =  MassMatFn(d,dx,func)
      call DataDelete(dx)     ! delete internally used data
   End Function MassMatFx
!-------------------------------------------------------------------------------
   Function MassMatFn(d,dq,func) result(Dm)   ! called when func supplied
      Type(ColDataType), intent(in) :: d,dq
      Real(float) :: Dm(d%nt,d%nt)
      integer :: nt
      Real(float), Allocatable :: fq(:)
      Interface
         Function func(x)  result(f)
            include 'defs.fi'
            Real(float), intent(in) :: x(:)
            Real(float) :: f(size(x))
         End Function func
      End Interface
      ! -------------------------------------------------------
      ! This function calculates the Galerkin or Moments mass matrix
      !  x - nodes in Lagrange trial functions
      !  func - f(x) in mass matrix integral
      !  nqx - number extra quadrature points (optional)
      !  nt, n - total/interior points,  nq - interior quadrature
      ! -------------------------------------------------------
      nt = d%nt
      if(Debug > 0)write(iout,'(a,5i3,L2)')' MassMatFn:',d%n,dq%n,d%typ,d%geo,d%sym,d%sft
      Allocate(fq(dq%nt))
      if(d%typ > Lobatto)then          ! if not Gauss or Lobatto
         fq(1:nt) = func(d%x)          ! use diagonal approximation
         Dm = MassMat00(d,fq(1:nt))
      else
         fq = (dq%w)*func(dq%x)        ! W*f(x)
         Call MassMatX(d,dq,dq%x,fq,Dm)
      endif
      Deallocate(fq)
   End Function MassMatFn
!-------------------------------------------------------------------------------
   Subroutine MassMatX(d,dq,xq,wq,Dm)
      Type(ColDataType), intent(in) :: d,dq
      Real(float), intent(in) :: xq(sym:)
      Real(float),intent(inout) :: wq(d%sym:)
      Real(float),intent(out) :: Dm(d%sym:,d%sym:)
      Real(float) :: Lq(size(xq)+d%sym-2,d%sym:d%n+1)
      Integer :: n,nq,i,j
      ! -------------------------------------------------------
      ! Internal procedure to calculate the Galerkin or Moments mass matrix
      ! for a general function f(x)
      ! d,dq - structure for base data & quadrature data with larger n
      ! Dm - mass matrix calculated
      ! xq - quadrature base points
      ! wq - w*f(x), quadrature weights time function
      ! n - interpolation pts., nq - quad. pts.
      ! see documentation for definition of mass matrix, section 3.1.2
      ! for Galerkin method:
      !  Int f(x)*Lj(x)*Li(x) ~ sum{Wk*f(xk)*Lj(xk)*Li(xk)} (endpoints simplify)
      ! for Moments (when not shifted):
      !  Int Lj(x)*Li(x)*f(x)*(1-xi^2)/(1-x^2)
      ! d%dx & dq%dx contain x(i) - x(j) arrays
      ! note: d%sym = 0 or 1, so dimension are 0/1 to n+1
      ! -------------------------------------------------------
      n = d%n; nq = size(xq)-2+d%sym
      Lq = LCoeff(xq(1:nq),d) ! evaluate interp. poly. at interior quad. pts
      if(d%typ==Gauss .and. d%sym==Symmetric)then
         wq = wq/(dq%dx(:,nq+1))   ! moments, wq = wk*f(xk)/(1-xk^2)
      elseif(d%typ==Gauss)then
         wq = wq/((dq%dx(:,1))*(dq%dx(:,nq+2)))
      endif
      do i=d%sym,n+1   ! normal integration (symmetric)
         do j=i,n+1
            Dm(i,j) = Sum(wq(1:nq)*Lq(:,i)*Lq(:,j))
            Dm(j,i) = Dm(i,j)
         enddo
      enddo
      if(d%typ==Gauss .and. d%sym==Symmetric)then
         Dm(n+1,:) = 0.0_float
         do i=1,n+1
            Dm(1:n,i) = Dm(1:n,i)*d%dx(1:n,n+1)
         enddo
      elseif(d%typ==Gauss)then
         Dm(0,:) = 0.0_float;  Dm(n+1,:) = 0.0_float
         do i=0,n+1
            Dm(1:n,i) = Dm(1:n,i)*d%dx(2:n+1,1)*d%dx(2:n+1,n+2)
         enddo
      elseif(d%typ==Lobatto .and. d%sym==Nonsymmetric)then
         Dm(0,0) = Dm(0,0) + wq(0)          ! boundary wt. contribution
         Dm(n+1,n+1) = Dm(n+1,n+1) + wq(nq+1)
      elseif(d%typ==Lobatto)then
         Dm(n+1,n+1) = Dm(n+1,n+1) + wq(nq+1)
      endif
   End Subroutine MassMatX
!-------------------------------------------------------------------------------
   Function LCoef1(x,d) Result(L)
      Type(ColDataType), intent(in) :: d
      Real(float), intent(in) :: x
      Real(float) :: L(d%nt)
      Real(float) :: L1(1,d%nt),x1(1)
      ! -------------------------------------------
      !  Calculate Lagrange interpolating polynomials at a single point
      ! -------------------------------------------
      x1(1) = x
      L1 = LCoefN(x1,d)
      L = L1(1,:)
   End Function LCoef1
!-------------------------------------------------------------------------------
   Function LCoef1x(x,xc) Result(L)
      Real(float), intent(in) :: x,xc(:)
      Real(float) :: L(size(xc))
      Real(float) :: L1(1,size(xc)),x1(1)
      ! -------------------------------------------
      !  Calculate Lagrange interpolating polynomials at a single point
      ! -------------------------------------------
      x1(1) = x
      L1 = LCoefNx(x1,xc)
      L = L1(1,:)
   End Function LCoef1x
!-------------------------------------------------------------------------------
   Function LCoefN(x,d) Result(L)
      Type(ColDataType), intent(in) :: d
      Real(float), intent(in) :: x(:)
      Real(float) :: L(size(x),d%nt)   ! interpolating polynomial at xx
      ! -------------------------------------------
      !  Calculate values of the Lagrange interpolating polynomials:
      !     x  - points where values are desired (evaluation points)
      !     d  - structure containing interpolation points & wb
      !     L(i,j) - value jth interpolating polynomial at xx(i)
      ! -------------------------------------------
      if(Debug > 0)write(*,'(a,5i5)')' LcoefN:',shape(L)
      if(d%sym == Symmetric)then
         call Lcoefx(x*x,(d%x*d%x),(d%wb*d%fwb),L)
      else
         call Lcoefx(x,d%x,(d%wb*d%fwb),L)
      endif
   End Function LCoefN
!-------------------------------------------------------------------------------
   Function LCoefNx(x,xc) Result(L)
      Real(float), intent(in) :: x(:),xc(:)
      Real(float) :: L(size(x),size(xc))   ! interpolating polynomial at xx
      ! -------------------------------------------
      !  Calculate values of the Lagrange interpolating polynomials:
      !     x  - points where values are desired (evaluation points)
      !     xc - the n interpolation points
      !     L(i,j) - value jth interpolating polynomial at x(i)
      ! -------------------------------------------
      if(Debug > 0)write(*,'(a,5i5)')' LcoefNx:',shape(L)
      if(sym == Symmetric)then
         call LCoefx(x*x,xc*xc,WBcalc(xc*xc),L)
      else
         call LCoefx(x,xc,WBcalc(xc),L)
      endif
   End Function LCoefNx
!-------------------------------------------------------------------------------
   Subroutine LCoefx(x,xc,wb,L)
      Real(float), intent(in) :: x(:),xc(:),wb(:)
      Real(float), intent(out) :: L(:,:)
      Real(float) :: px(size(xc)),pp(size(xc))
      Integer :: i,k
      ! -------------------------------------------
      !  Calculate values of the Lagrange interpolating polynomials:
      !     x  - points where values are desired (evaluation points)
      !     xc - the n interpolation points (collocation points)
      !     wb - barycentric weights
      !     L(i,j) - value jth interpolating polynomial at xx(i)
      !  example usage: given values yc at the n interpolation points
      !     y(x) = MatMul(L,yc) is the interpolant at x(:)
      ! -------------------------------------------!
!      call Arrayprint('LCoefx',xc,wb)
      do k=1,size(x)
         if(minval(abs(x(k)-xc(:))) > 0.0001_float)then
            L(k,:) = product(x(k)-xc)*wb/(x(k)-xc)
         else
            do i=1,size(xc)  ! use long calculation if points too close
               px(:) = x(k)-xc(:); px(i) = 1.0_float; pp(i) = product(px)
            enddo
            L(k,:) = pp*wb
         endif
      enddo
      !call Arrayprint('LCoefx',xc,wb)
      !call Arrayprint('LCoefx',x,L)
   End Subroutine LCoefx
!-------------------------------------------------------------------------------
   Function Lpoly(xx) Result(q)
      Real(float) :: xx(:),q(size(xx),size(xx))
      Real(float) :: x(size(xx)),xj(size(xx),size(xx)-1)
      integer :: i,k,n,kk
! ----------------------------------------------
!  Calculated expanded interpolating polynomial
!     xx = vector of points
!     L(i) = sum(q(i,j)*x^(j-1)) expansion of interpolating polynomial
!  ** Warning - monomial calculations subject to rounding errors
! ----------------------------------------------
      n = size(xx)
      if(sym == Symmetric)then
         x = xx*xx
      else
         x = xx
      endif
      ! construct x array with i = j dropped out, then denominator
      do i=1,n
         xj(i,:i-1) = x(:i-1)
         xj(i,i:) = x(i+1:)
         q(i,1) = (1.0_float)/(product(x(i)-xj(i,:)))
      enddo
      do k=2,n
         q(:,k) = q(:,k-1);
         do kk=k-1,2,-1
           q(:,kk) = q(:,kk-1) - xj(:,k-1)*q(:,kk);
         enddo
         q(:,1) = -xj(:,k-1)*q(:,1);
      enddo
   End Function Lpoly
!-------------------------------------------------------------------------------
   Function InterpN(x,d,yc) Result(y)
      Type(ColDataType), intent(in) :: d
      Real(float), intent(in) :: x(:)        ! x values for which y's are calculated
      Real(float), intent(in) :: yc(:,:)     ! known y values (at colloc. points)
      Real(float) :: y(size(x),size(yc,2))   ! approx. y at x
      Real(float) :: Lc(size(x),d%nt)
      integer :: k
      ! interpolate multiple sets of y values
      Lc = Lcoeff(x,d) ! interpolating polynomials
      do k = 1,size(yc,2)  !* eliminate this loop?
         y(:,k) = Matmul(Lc,yc(:,k))
      enddo
   End Function InterpN
!-------------------------------------------------------------------------------
   Function Interp1(x,d,yc) Result(y)
      Type(ColDataType), intent(in) :: d
      Real(float), intent(in) :: x(:)  ! x values for which y's are calculated
      Real(float), intent(in) :: yc(:) ! known y values (at colloc. points)
      Real(float) :: y(size(x))        ! approx. y at x
      Real(float) :: Lc(size(x),d%nt)
      ! interpolate one set of y values
      Lc = Lcoeff(x,d)  ! interpolating polynomials
      y = Matmul(Lc,yc)
   End Function Interp1
!-------------------------------------------------------------------------------
!   Function QW_general(x) result(w)
!      Real(float) :: x(:),w(size(x))
!      Real(float) :: q(size(x),size(x)),v(size(x))
!      integer :: k,m
!   Alternative code to Quad_general uses monomial tranform >> roundoff issues
!      m = size(x)
!      q = Lpoly(x)                  ! interpolating polynomial as monomials
!      v = Real((/(k,k=0,m-1)/),float)
!      if(geom > Nonsymmetric)v = v*2.0_float + Real(geom,float)
!      v = 1.0_float/v
!      w = MatMul(q,v)
! !!     call Arrayprint('QW_general',q,v,w)
!   End Function QW_general
!-------------------------------------------------------------------------------
   Function Quad_general(x,wb,s) result(w)
      integer :: s
      Real(float), intent(in) :: x(s:),wb(s:)
      Real(float) :: w(s:size(x)+s-1)
      Real(float) :: ab(2),fscale,fac
      Real(float) :: xq(s:size(x)/2+1,4),Lq(size(x)/2,size(x))
      Integer :: n,nq
      ! ---------------------------------------
      ! Brute force calculation of quadrature weights
      ! x,w - arbitrary points & weights
      ! wb - barycetric weights
      ! s = 0/1 symmetry
      ! xq,wq - Lobatto points & weights used for integration
      ! Lq - Lagrange interpolating polynomials evalated at xq < 1.0
      ! calculate w by direct integration of interpolating polynomial
      ! ---------------------------------------
      n = size(x)-2+s;  nq = size(x)/2
      fac = 1.0_float
      ab = Polyset(Lobatto,geom)
      xq = Jacobi_Quadrature(nq,ab,geom,fscale)
      if(sym==Symmetric)then
         call LCoefx(xq(1:nq,1)**2,x*x,wb,Lq)
         w = Matmul(xq(1:nq,4),Lq)        ! integrate L
         w(n+1) = w(n+1) + xq(nq+1,4)     ! last point
      else
         if(shft)then
            xq(:,1) = (xq(:,1)+1.0_float)*0.5_float
            fac = 0.5_float
         endif
         call LCoefx(xq(1:nq,1),x,wb,Lq)
         w = Matmul(xq(1:nq,4),Lq)        ! integrate L
         w(0) = w(0) + xq(0,4)            ! 1st point
         w(n+1) = w(n+1) + xq(nq+1,4)     ! last point
         w = w*fac
      endif
      if(Debug > 2)then
         write(iout,'(a,10i4)')'Quad_general: n,nq,geom =',n,nq,geom,shape(xq),shape(Lq)
         call Arrayprint('Quad_general xq,Lq',xq(1:nq,1),Lq)
         call Arrayprint('Quad_general x,w,wb',x,w,wb)
      endif
   End Function Quad_general
!-------------------------------------------------------------------------------
   Function WBcalc(x,fwb) result(w)
      Real(float), intent(in) :: x(:)
      Real(float), optional :: fwb
      Real(float) :: w(size(x)),px(size(x))
      integer :: i
      ! ----------------------------------------------
      ! Calculate barycentric weights via product
      ! fwb - optional scaling factor
      ! ----------------------------------------------
      do i=1,size(x)
         px = x(i) - x(:);  px(i) = 1.0_float
         w(i) = (1.0_float)/(product(px))
      enddo
      if(present(fwb))then
         fwb = sum(abs(w))/real(size(x),float)
         w = w/fwb
      endif
   End Function WBcalc
!-------------------------------------------------------------------------------
End Module OrthogonalColloc
