Module Orth_Cheb  ! fundamental calculations with Chebyshev polynomials
   Use Array_Print, only : ArrayPrint, VectorPrint
   Use Orth_Poly, only : Polyset, Jacobi_All, Jacobi_Lead, Jacobi_Ends
   include 'defs.fi'
   Private
   Public Cheb_Quad
   Public Cheb_Scale
   Integer, parameter, private :: Debug = 0
   Integer, parameter, private :: Nonsymmetric=0, Planar=1, Cylindrical=2, Spherical=3, Symmetric = 1
   Real(floatq), Parameter, Private :: qpi= 3.14159265358979323846264338327950288419716939937510_floatq
   Real(float), Parameter, Private :: pi = real(qpi,float)
   Integer, Parameter :: Chebyshev=3, Chebyshev1=6
   Character (len=*), Parameter, private :: fmt17 = '(20(f20.17,a1))'
   contains
   !----------------------------------------------------------------------------
   Function Cheb_Quad(n,g,s,fs,Shift)  result(xw)
      integer, intent(in) :: n,g,s
      real(float), intent(out) :: fs
      logical, optional :: Shift
      real(float) :: xw(s:n+1,4),p0(2)
      integer :: nj
      logical :: shft,shftx
      real(float), parameter :: abeta(2) = (/0.5_float,0.5_float/)
      ! --------------------------------------------------
      shft = .false.;   if(present(Shift))shft = Shift
      shftx = shft .and. .not.(s==Symmetric)
      nj = n*(s+1) ! double value for symmetric cases
      call XCheb(n,s,xw)                  ! x & theta
      if(g <= Planar)then
         call QW_Cheb(s,xw(:,1),xw(:,4))   ! w - quadrature weights
      endif
      fs = Jacobi_Lead(nj,abeta,Shift=shftx)
      p0 = Jacobi_Ends(nj,abeta,Monic=.false.,Shift=shftx)
      call WBcalc(xw(:,1),xw(:,3),s,shft) ! wb - barycentric weights
      fs = fs/p0(2)
      if(Debug > 1)then
         write(iout,*)'fs,p0 =',fs,fs*p0(2),p0(2)
         call arrayprint('Cheb_Quad: x,theta,wb,wq',xw,fmtx=fmt17)
      endif
      if(shftx)then
         xw(:,1) = (xw(:,1)+1.0_float)*0.5_float
      elseif(s==Nonsymmetric)then
         xw(:,4) = xw(:,4)*2.0_float
      endif
   End Function Cheb_Quad
!-------------------------------------------------------------------------------
   Subroutine WBcalc(x,wb,s,shft)
      integer, intent(in) :: s
      real(float), intent(in) :: x(s:)
      real(float), intent(out) :: wb(s:)
      logical, intent(in) :: shft
      real(float) :: f0
      integer :: n,k
      ! -----------------------------------
      ! Calculate barycentric weights for chebyshev
      ! -----------------------------------
      n = size(x) - 2 + s
      f0 = 1.0_float;  if(shft)f0 = 2.0_float
      wb(n+1) = f0*0.5_float
      do k=n-1,s,-2
         wb(k) = f0
      enddo
      f0 = -f0
      do k=n,s,-2
         wb(k) = f0
      enddo
      if(s==0)then
         wb(s) = wb(s)*0.5_float
      else
         wb = x*wb
      endif
   End Subroutine WBcalc
!-------------------------------------------------------------------------------
   Subroutine XCheb(n,s,x)
      integer, intent(in) :: n,s
      Real(float), intent(out) :: x(s:n+1,2)
      Real(float) :: fac
      Integer :: i   !, ix(s:n+1)
      !--------------------------------------------
      if(s == Symmetric)then
         fac = pi/Real(2*n+1,float) ! gives shifted values
      else
         fac = pi/Real(n+1,float)   ! unshifted values
      endif
      x(:,2) = fac*Real((/((i),i=n+1-s,0,-1)/),float)
      x(:,1) = cos(x(:,2))
   End Subroutine XCheb
!-------------------------------------------------------------------------------
   Subroutine QW_Cheb(s,xq,wq)
      integer, intent(in) :: s
      Real(float), intent(in)  :: xq(:) ! base points
      Real(float), intent(out) :: wq(:) ! weights
      integer :: nc,m
      if(Debug > 0)write(iout,'(a,5i5)')'QW_Cheb 1',s
      nc = size(xq) - 2 + s
      m  = size(xq)*(s + 1)
      wq = QW_ClenCurt(nc,m,s)
   End Subroutine QW_Cheb
!-------------------------------------------------------------------------------
   Function QW_ClenCurt(nc,n,s) result(w)
      integer, intent(in) :: nc,n,s
      integer :: nn,nodd,k,i
      Real(float) :: w(s:nc+1),ang(n/2+1:n-1),v(n/2+1:n-1)
      ! Clenshaw-Curtis quadrature weights
      nn = n - 1;  nodd = mod(n,2); i = n/2 + 1
      ang = (/(Real(k,float),k=i-1,n-2)/)*(Pi/Real(n-1,float))
      v = 1.0_float;  w(nc+1) = 0.5_float/Real(nn*nn-nodd,float)
      do k=1,n/2-1
         v = v - cos(Real(2*k,float)*ang)*(2.0_float)/Real(4*k*k-1,float)
      enddo
      if(nodd == 1)then  ! odd number interior points
         v = v - cos(Real(nn,float)*ang)*w(nc+1)*2.0_float
      endif
      v = v/Real(n-1,float)
      if(s == NonSymmetric)then
         w(i-1:nc) = v;  w(0:i-2) = w(nc+1:n+1-i:-1)
      else
         w(:nc) = v;  w = w*2.0_float
      endif
   End Function QW_ClenCurt
!-------------------------------------------------------------------------------
   function Cheb_Scale(n,abeta)  result(sf)
      integer :: n,k,ia
      real(float) :: abeta(2),sf(0:n),zn(n)
      ! calculate scale factor for Jacobi polynomial scaling
      !write(iout,'(a,2i4)')'Cheb_Tscale:',n,typ
      ia = nint(abeta(1)*2.0_float)
      zn = (/(real(k,float),k=1,n)/)
      sf(0) = 1.0_float
      do k=1,n
         sf(k) = sf(k-1)*(1.0_float + abeta(1)/zn(k))
      enddo
      if(ia == 1)sf(1:) = sf(1:)/(zn + 1.0_float)
   end function Cheb_Scale
!-------------------------------------------------------------------------------
!   Function QW_ChebSym(x) result(w)
!      Real(float) :: x(:),w(size(x))
!      Real(float) :: xq((size(x)+2)/2),wq((size(x)+2)/2),Lq(size(x)/2,size(x))
!      Integer :: n1,nq
!      ! x,w - Chebyshev points & weights
!      ! xq,wq - Lobatto points & weights used for integration
!      ! Lq - Lagrange interpolating polynomials evalated at xq < 1.0
!      ! calculate w by direct integration of interpolating polynomial
!      n1 = size(x);  nq = size(xq) - 1
!      if(Debug>0)write(iout,'(a,4i3)')'QW_ChebSym: n,nq =',n1-1,nq,shape(Lq)
!      Call CollocSet(TType=Lobatto)    ! set to Lobatto
!      xq = Xpoints(nq)                 ! get base points
!      d => DataFind(xq)                ! d should be correct from Xpoints
!      Call QWeights(xq,d%wd,wq)        ! get weights
!      Lq = LCoeff(xq(1:nq),x)          ! L(xq) (all but xq = 1
!      w = Matmul(wq(1:nq),Lq)          ! integrate L
!      w(n1) = w(n1) + wq(nq+1)         ! last point
!      Call CollocSet(TType=Chebyshev)
!      call DataDelete(xq)              ! delete internally used data
!   End Function QW_ChebSym
   !----------------------------------------------------------------------------
End Module Orth_Cheb

