Program Jacobi_test
   Use Orth_Poly, Only : Legendre,Legendre_Transform,Jacobi_Series,Legendre_Deriv,Jacobi_Deriv
   Use OrthogonalColloc, Only : ColDataType,Initialize,Xpoints,WeightQ,DataDelete,Lobatto,Gauss,TxtType
   Use LUSolvers, Only : LUSolve
   Use Array_Print, only : ArrayPrint, VectorPrint, OpenFile
   include 'defs.fi'
   integer :: n
   integer, parameter :: ity(2) = (/Gauss,Lobatto/)
   logical :: ExtraOutput = .false.
   Character (len=*), Parameter :: fmt6 = '(20(f12.7,a1))'
   Real(float), parameter :: thx = (5.0_float)**2  ! thiele parameter
   ! --------------------------------------------------------
   ! This code solves linear source problem with Legendre polynomial
   ! trial functions. It does not attempt to be efficient, just uses quadrature
   ! for integration even when analytical integrals are possible. This is more
   ! for checking calculations reported in the monograph
   !     MWRTest - solves the differential equation
   !     DerivTest - displays the modal differentiation matrices
   ! --------------------------------------------------------
   write(*,'(a)')' Value of n>'
   read(*,'(i5)')n      ! n is equivalent interior points
   n = n - 1            ! 1 less
   Call OpenFile('polybvp')
   Call MWRTest(n)      ! solutions by MWR w/Legendre trial functions
   if(ExtraOutput)then
      Call DerivTest(n) ! derivatives of Legendre polynomials
   endif
!----------------------------------------------------------
   contains
!----------------------------------------------------------
   Subroutine MWRTest(n)
      Integer :: n,nc,i,k
      Integer, parameter :: ns=3
      Real(float) :: a(0:n,ns),as(0:n/2),yc(0:n+2,ns),ap(0:n+2,ns),am(0:n+2,ns)
      Real(float) :: xc(0:n+2),wc(0:n+2)
      Real(float) :: Q(0:n+2,0:n+2)
      Type(ColDataType) :: OC
      Character (len=*), Parameter :: TxtMeth(1:2) = (/'Moments ','Galerkin'/)
      ! ---------------------------------
      ! solves the catalyst particle problem:
      !   the symmetric problem with constant coefficients
      !   the nonsymmetric problem with cubic coefficient
      ! 0:n - n+1 basis functions after b.c.'s imposed
      ! n+2 - highest polynomial degree
      ! nc = n+1 - equivalent number of interior pts
      ! Solve the following problems stored in a(:,1:6)
      !  - Symmetric problem solution Moments (same as Gauss Fig. 3.1)
      !  - Nonsymmetric collocation at Gauss pts. Fig. 3.14
      !  - Nonsymmetric Moments, Fig. 3.15
      !  - Symmetric solution Galerkin (same as Lobatto Fig. 3.1)
      !  - Nonsymmetric collocation at Lobatto pts. Fig. 3.14
      !  - Nonsymmetric Galerkin, Fig. 3.16
      ! The code also shows how to calculate nodal values and how to
      ! calculate modal coefficients from nodal values
      ! ---------------------------------
      nc = n + 1;  a = 0.0_float  ! note: n = nc-1
      do i= 1,2   ! Gauss/moments 1st, Lobatto/Galerkin 2nd
         write(iout,'(/,5a,2i3)')'Three solutions using ',trim(TxtMeth(i)),' & collocation at ', &
            trim(TxtType(ity(i))),' points: n, nc =',n,nc
         write(iout,'(10a)')'    (1) Symmetric (2) Nonsymmetric Collocation, (3) Nonsymmetric ',TxtMeth(i)
         call Initialize(OC,nc,TType=ity(i),Shift=.false.)
         xc = Xpoints(OC);   wc = WeightQ(OC)
         as = MWRTestSym(n,n+2,ity(i),TxtMeth(ity(i)))      ! symmetric const. coeff
         a(0:n:2,1) = as   ! odd coefficients zero
         a(:,2) = MWRTestNon(n,n+1,ity(i),TxtMeth(ity(i)))  ! collocation, variable coef.
         a(:,3) = MWRTestNon(n,n+3,ity(i),TxtMeth(ity(i)))  ! Galerkin/Moments, variable coef
         am(0:1,:) = -a(0:1,:)      ! a are coefficients of p_k+2 - p_k
         am(n+1:n+2,:) = a(n-1:n,:) ! am are coefficients of p_k
         do k = 2,n
            am(k,:) = a(k-2,:) - a(k,:)
         enddo
         do k=1,3    ! calculate nodal values from Legendre series
            yc(:,k) = Jacobi_Series(am(:,k),xc)
         enddo
         Q = Legendre_Transform(xc,wc,ity(i))   ! ok Gauss or Lobatto
         do k=1,3    ! tranform from nodal values back to Legendre coefficients
            ap(:,k) = MatMul(Q,yc(:,k))
         enddo
         write(iout,'(a)')'Result Order: symmetric, collocation, full method'
         if(ExtraOutput)then
            Call ArrayPrint('Transformation Matrix '//TxtType(ity(i)),Q,fmtx=fmt6)
         endif
         Call ArrayPrint('Modal Solutions, symmetric/nonsymmetric:',Transpose(a),fmtx=fmt6)
         Call ArrayPrint('Solutions Legendre coefficients:',Transpose(am),fmtx=fmt6)
         call ArrayPrint('Nodal '//trim(TxtType(ity(i))) &
            //': x, symmetric, collocation, full method',xc,yc,fmtx=fmt6)
         Call ArrayPrint('Transform back to modal solutions ',Transpose(ap),fmtx=fmt6)
      enddo
   End Subroutine MWRTest
!----------------------------------------------------------
   Function MWRTestSym(n,nc,ityp,txt) result(as)
      integer :: n,nc,i,j,ns,ityp
      Real(float) :: as(0:n/2)
      real(float) :: xc(0:nc+1),wc(0:nc+1),pcn(0:nc+1,0:n+2,0:2)
      Real(float) :: ds(0:n/2,0:n/2),bs(0:n/2,0:n/2),qs(0:n/2)
      Real(float) :: dz(0:n/2+1,0:n/2),bz(0:n/2+1,0:n/2),qz(0:(n+1)/2)
      Type(ColDataType) :: OC
      Character :: txt*(*)
      ! ----------------------------------------------------
      !  calculate modal Legendre polynomial solution to the symmetric problem
      !  Coefficients could be constructed analytically, but quadrature is
      !  used here. Shortcut polynomials could also be used to reduce the
      !  calculations
      ! ----------------------------------------------------
      write(*,'(a,2i3)')trim(txt)//' Method Symmetric, n,nc =',n+1,nc
      write(iout,'(a,2i3)')trim(txt)//' Symmetric Solution: n, nc =',n,nc
      call Initialize(OC,nc,TType=ityp,Shift=.false.)
      xc = Xpoints(OC)  ! quadrature nodes & weights
      wc = WeightQ(OC);
      pcn = Legendre(xc,n+2,2)   ! Legendre polynomials & 2 derivatives
      if(ExtraOutput)then
         Call ArrayPrint('Legendre Polynomial at Quadrature Points:',xc,pcn(:,:,0),fmtx=fmt6)
         Call ArrayPrint('Legendre Polynomial 1st Derivatives:',xc,pcn(:,:,1),fmtx=fmt6)
         Call ArrayPrint('Legendre Polynomial 2nd Derivatives:',xc,pcn(:,:,2),fmtx=fmt6)
      endif
      ns = n/2
      do i = 0,ns+1  ! numerical integration
         qz(i) = sum(wc(:)*(pcn(:,i*2,0)))
         do j = 0,ns
            dz(i,j) = sum(wc(:)*(pcn(:,j*2+2,0)-pcn(:,j*2,0))*(pcn(:,i*2,0)))
            bz(i,j) = sum(wc(:)*(pcn(:,j*2+2,2)-pcn(:,j*2,2))*(pcn(:,i*2,0)))
         enddo
      enddo
      if(ityp == Lobatto)then    ! Galerkin
         qs = qz(1:ns+1) - qz(:ns)
         ds = dz(1:ns+1,:) - dz(:ns,:)
         bs = bz(1:ns+1,:) - bz(:ns,:)
      else
         qs = qz(:ns)   ! moments
         ds = dz(:ns,:)
         bs = bz(:ns,:)
      endif
      if(ExtraOutput)then
         call ArrayPrint('Mass Matrix, load vector:',ds,qs,fmtx=fmt6)
         call ArrayPrint('Stiffness Matrix Symmetric:',bs,fmtx=fmt6)
      endif
      bs = bs - thx*ds  ! build final matrix problem, then solve
      qs = -thx*qs
      as = Lusolve(bs,qs)
      call ArrayPrint('Symmetric Solve: a, matrix, rhs',as,bs,qs,fmtx=fmt6)
      Call DataDelete(OC)
   End Function MWRTestSym
!----------------------------------------------------------
   Function MWRTestNon(n,nc,ityp,txt) result(a)
      integer :: n,nc,i,j,ityp
      Real(float) :: a(0:n)
      real(float) :: xc(0:nc+1),wc(0:nc+1),fx(0:nc+1),pcn(0:nc+1,0:n+2,0:2)
      Real(float) :: dx(0:n+2,0:n),bx(0:n+2,0:n),qx(0:n+2)
      Real(float) :: d(0:n,0:n),b(0:n,0:n),q(0:n)
      Real(float) :: ars(0:nc+1,0:n),arm(0:nc+1,0:n),arr(0:nc+1,0:n+1),ac(0:n)
      Type(ColDataType), save :: OC ! save over multiple calls (explicit deletion not required)
      Character :: txt*(*)
      ! ----------------------------------------------------
      !  Calculates modal solution for nonsymmetric variable coefficient
      !  problem. Implementation uses numerical integration.
      ! ----------------------------------------------------
      write(*,'(a,2i3)')trim(txt)//' Method Nonsymmetric, n,nc =',n+1,nc
      write(iout,'(a,2i3)')trim(txt)//' Nonsymmetric Solution: n, nc =',n,nc
      call Initialize(OC,nc,TType=ityp,Shift=.false.)
      xc = Xpoints(OC)  ! quadrature nodes & weights
      wc = Weightq(OC)
      pcn = Legendre(xc,n+2,2)   ! polynomials & derivatives
      if(ExtraOutput)then
         Call ArrayPrint('Legendre Polynomials at Quadrature Points:',xc,pcn(:,:,0),fmtx=fmt6)
         Call ArrayPrint('Legendre Polynomial 1st Derivatives:',xc,pcn(:,:,1),fmtx=fmt6)
         Call ArrayPrint('Legendre Polynomial 2nd Derivatives:',xc,pcn(:,:,2),fmtx=fmt6)
      endif
      fx = MassFunc(xc)  ! variable coefficients of mass matrix
      if(nc == n+1)then  ! demonstrate collocation is equivalent for this case
         ars = 0.0_float;  arm = 0.0_float
         do j=0,n
            arm (:,j) = fx(:)*(pcn(:,j+2,0)-pcn(:,j,0))
            ars (:,j) = pcn(:,j+2,2) - pcn(:,j,2)
         enddo
         arr(:,0:n) = ars - thx*arm
         arr(:,n+1) = -fx*thx
         ac = Lusolve(arr(1:nc,:),arr(1:nc,n+1))
         if(ExtraOutput)then
            Call ArrayPrint('Residual (mass):',xc,arm,fmtx=fmt6)
            Call ArrayPrint('Residual (stiffness):',xc,ars,fmtx=fmt6)
         endif
         Call ArrayPrint('Collocation: a, matrix rhs:',ac,arr(1:nc,:),fmtx=fmt6)
      endif
      do i = 0,n+2   ! numerical integration
         qx(i) = sum(wc(:)*fx*(pcn(:,i,0)))
         do j = 0,n
            dx(i,j) = sum(wc(:)*fx*(pcn(:,j+2,0)-pcn(:,j,0))*(pcn(:,i,0)))  !
            bx(i,j) = sum(wc(:)*(pcn(:,j+2,2)-pcn(:,j,2))*(pcn(:,i,0)))
         enddo
      enddo
      if(ityp == Lobatto)then
         q = qx(2:n+2) - qx(:n)     ! Galerkin
         d = dx(2:n+2,:) - dx(:n,:)
         b = bx(2:n+2,:) - bx(:n,:)
      else
         q = qx(:n)                 ! moments
         d = dx(:n,:)
         b = bx(:n,:)
      endif
      if(ExtraOutput)then
         call VectorPrint('Variable Coefficients at quadrature points',fx,fmtx=fmt6)
         call ArrayPrint('Mass Matrix, load vector:',d,q,fmtx=fmt6)
         call ArrayPrint('Stiffness Matrix (b):',b,fmtx=fmt6)
      endif
      b = b - thx*d  ! assemble final matrix and solve
      q = -thx*q
      a = Lusolve(b,q)
      call ArrayPrint('Full method: a, matrix rhs',a,b,q,fmtx=fmt6)
      Call DataDelete(OC)
   End Function MWRTestNon
  !----------------------------------------------------------
   Function MassFunc(x) result(f)  ! cubic variation for mass matrix
      Real(float), intent(in) :: x(:)
      Real(float) :: f(size(x)),xx(size(x))
      Real(float), parameter :: a0 = 0.20_float, a1 = (1.0_float - a0)*4.0_float
      xx = (x + 1.0_float)*0.5_float      ! convert x to 0,1
      f = a0 + a1*(1.5_float - xx)*xx*xx  ! BVP with variable coef.
   End Function MassFunc
   !----------------------------------------------------------
   Subroutine DerivTest(n)
      integer :: n,i
      integer :: d(0:n+2,1:5),ds(0:n+2,0:n),d2(0:n+2,0:n),dj(0:n+1,0:n+1)
      real(float) :: abeta(2)
      ! --- Legendre Derivative coefficients for display
      if(n > 20)return
      d = 0;  ds = 0
      d(:,1) = Legendre_Deriv(n+2,1)   ! P'_n+2
      d(:,2) = Legendre_Deriv(n+2,2)   ! P"_n+2
      d(:n,3) = Legendre_Deriv(n,2)    ! P"_n
      d(:n,4) = Legendre_Deriv(n,3)    ! P"_n+2 - P"_n
      d(:,5) = d(:,2) - d(:,3)         ! P"_n+2 - P"_n
      Call ArrayPrint('Derivatives: 1st, 2nd: n, n-2, diff:',transpose(d))
      abeta = 0.0_float
      dj = Jacobi_Deriv(n+2,abeta)   ! Jacobi(1,1) 1st derivative (same as d(:,1) when a = 0)
      Call ArrayPrint('Derivatives: 1st Jacobi(0,0):',dj)
      abeta = 1.0_float
      dj = Jacobi_Deriv(n+2,abeta)   ! Jacobi(1,1) 1st derivative (same as d(:,1) when a = 0)
      Call ArrayPrint('Derivatives: 1st Jacobi(1,1):',dj)
      do i=0,n
         d2(0:i+2,i) = Legendre_Deriv(i+2,2)  ! coefficients of Legendre derivatives
         ds(0:i+2,i) = Legendre_Deriv(i+2,3)  ! coefficients of Legendre derivatives
      enddo
      Call ArrayPrint('Derivatives: 2nd derivatives all:',d2)
      Call ArrayPrint('Derivatives: 2nd difference all:',ds)
      Call ArrayPrint('Derivatives: 2nd all symmetric:',ds(0::2,0::2))
   End Subroutine DerivTest
   !----------------------------------------------------------
End Program Jacobi_test

